// ---------------------------------------------------------------------------
// LTLChecker.java
// An LTL (Linear Temporal Logic) model checker.
// $Id: LTLChecker.java,v 1.20 2023/05/09 09:23:03 schreine Exp $
//
// Authors: Ágoston Sütő <agoston.suetoe@jku.at>
//      and Wolfgang Schreiner <Wolfgang.Schreiner@risc.jku.at>
// Copyright (C) 2021-, Research Institute for Symbolic Computation (RISC)
// Johannes Kepler University, Linz, Austria, http://www.risc.jku.at
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program.  If not, see <http://www.gnu.org/licenses/>.
// ----------------------------------------------------------------------------
package riscal.ltl;

import java.io.*;
import java.util.*;

import riscal.*;
import riscal.ltl.automaton.AutomatonCreator;
import riscal.ltl.automaton.AutomatonNode;
import riscal.ltl.automaton.LTLPreprocessor;
import riscal.ltl.automaton.NNFConverter;
import riscal.ltl.search.CheckingState;
import riscal.ltl.search.CounterExample;
import riscal.ltl.search.EmptinessChecker;
import riscal.syntax.AST;
import riscal.syntax.AST.*;
import riscal.types.Environment;
import riscal.types.Environment.Symbol.*;
import riscal.semantics.*;
import riscal.semantics.Systems.*;
import riscal.semantics.Types.*;
import riscal.swt.MainSWT;
import riscal.util.Seq;

public class LTLChecker {

    // if set to true, in the RISCAL GUI the "Tasks" folder of a system
    // lists a task for every LTL property of the system
    public final static boolean ENABLED = true;

    // the translator of syntax to semantics
    public static final Translator translator = new Translator(Main.getOutput(), Main.getNondeterministic());

    /****************************************************************************
     * Check whether "system" satisfies the LTL formula "ltl".
     * @param system the system expected to satisfy the LTL formula.
     * @param fairness null or a list of (possibly universally quantified)
     *        "weak fairness" and/or "strong fairness" conditions.
     * @param ltl a formula in linear temporal logic (LTL).
     * Returns normally if and only if the system satisfies the formula under
     *         the specified fairness conditions.
     * also prints a success message (otherwise an error message with
     * a counterexample trace is printed and a RuntimeError is thrown).
     ***************************************************************************/
    public static void check(SystemSymbol system, List<LTL> fairness, LTL ltl) {
        PrintWriter out = Main.getOutput();
        String ltls = ltl.toString();
        if (ltls.length() > 50) {
            ltls = ltls.substring(0, 50);
        }
        out.println("Checking LTL formula " + ltls + "...");

        try {
            long t = System.currentTimeMillis();
            CounterExample result = new LTLChecker(system).check(ltl, fairness);
            t = System.currentTimeMillis() - t;
            if (result == null) {
                out.println("LTL formula is satisfied (model checking time: " + t + " ms).");
                if (Main.getCheckDepth() != null)
                  out.println("WARNING: counterexample search was limited to depth " + Main.getCheckDepth() + 
                      "; clear the \"Depth\" field to be sure that there is indeed no counterexample."); 
            } else {
                throw new Translator.RuntimeError(ltl,
                        "LTL formula is NOT satisfied (model checking time: " + t + " ms).\n" +
                                "Counterexample execution:\n" + result.toString(system.getNames()));
            }
        } catch (LTLError e) {
            reportError(e);
            throw new Translator.RuntimeError(ltl, "internal error in model checking.");
        }
    }

    /****************************************************************************
     * Report LTL error.
     * @param e the error.
     ***************************************************************************/
    private static void reportError(LTLError e) {
        PrintWriter out = Main.getOutput();
        if (e.tree == null || e.tree.getPosition() == null) {
            out.println(e.getMessage());
            out.println("ERROR: error in LTL model checking.");
            return;
        }
        SourcePosition position = e.tree.getPosition();
        int line = position.lineFrom;
        int character = position.charFrom;
        AST.Source source = position.source;
        out.println(source.file.getName() + ":" + (line + 1) + ": " + e.getMessage());
        out.println(source.lines.get(line));
        for (int i = 0; i < character; i++) {
            out.print(" ");
        }
        out.println("^");
        out.println("ERROR: error in LTL model checking.");
        MainSWT.reportError(e.getMessage(), position);
    }

    /****************************************************************************
     * Construct from given context the list of contexts that arises from the
     * evaluation of a quantified variable.
     * @param context the given context.
     * @param qvar the quantified variable.
     * @return the list of contexts.
     ***************************************************************************/
    public static List<Types.Context> getContexts(Types.Context context, AST.QuantifiedVariable qvar) {
        // compute context list arising from quantification
        Types.QvarSem qvsem = translator.process(qvar);
        if (!(qvsem instanceof Types.QvarSem.Single)) {
            throw new LTLError(qvar, "nondeterministic semantics not supported in LTL quantifier");
        }

        Environment.Symbol.ValueSymbol[] symbols = Translator.getSymbols(qvar);
        int[] slots = Translator.toSlots(symbols);
        String[] names = Translator.toNames(symbols);
        Seq<Value[]> seq = ((Types.QvarSem.Single) qvsem).apply(context);
        List<Types.Context> result = new ArrayList<>();
        while (true) {
            Seq.Next<Types.Value[]> next = seq.get();
            if (next == null) {
                break;
            }
            Types.Value[] values = next.head;
            Types.Context context0 = new Types.Context(context);
            context0.set(slots, values, names);
            result.add(context0);
            seq = next.tail;
        }
        return result;
    }

    // the current system
    private final SystemSymbol system;
    // the number of slots required to represent its state
    private final int stateSize;
    // the system to be checked as a labelled transition system
    private final SysSem lts;

    public final Map<StateClass, List<StateClass>> successorsCache;

    // this map is used to "intern" CheckingStates (think String::intern),
    // that is to make sure only one exists for a given equivalence class
    // using a Map<T, T> for this purpose is a standard Java hack
    // since Set<T> does not allow access to the underlying object
    public final Map<CheckingState, CheckingState> checkingStateCache;

    private boolean deadlockWarningEmitted;

    /****************************************************************************
     * Construct checker for given system.
     * @param system the system for which an LTL formula is to be checked.
     ***************************************************************************/
    private LTLChecker(SystemSymbol system) {
        // set up global variables
        this.system = system;
        ComponentSymbol[] components = system.getComponents();
        stateSize = components == null ? system.getVars().length : components.length;
        successorsCache = new HashMap<>();
        checkingStateCache = new HashMap<>();
        lts = system.getValue();
    }

    public List<StateClass> getSuccessors(StateClass state) {
        if (state == null) {
            return lts.getInitial().list();
        } else {
            return successorsCache.computeIfAbsent(state, (s) -> lts.getNext(s).list());
        }
    }

    // --------------------------------------------------------------------------
    //
    // The following methods represent the interface to system states/actions.
    //
    // These methods may raise an exception that has to be caught and handled
    // by the checker (by printing the exception message and a counterexample
    // trace which leads to the error state; see Systems.Visitor.visit() for
    // inspiration on how to deal with errors and error traces).
    //
    // --------------------------------------------------------------------------

    /****************************************************************************
     * Returns true if "state" satisfies state formula with semantics "sem"
     * defined in "context".
     * @param state a system state.
     * @param context the values of the free (logical) variables in the formula.
     * @param sem the semantics of an LTL state formula.
     * @return true if and only if the state satisfies the formula;
     ***************************************************************************/
    boolean checkStateFormula(StateClass state, Context context,
                              ExpSem.Single sem) {
        Context context0 = new Context(state.context);
        context0.setFrom(context, stateSize);
        Value.Bool result = (Value.Bool) sem.apply(context0);
        return result.getValue();
    }

    /****************************************************************************
     * Returns true if "state" satisfies action formula with semantics determined
     * by "kind", "action", "instance", "args" defined in "context".
     * @param state a system state.
     * @param kind the kind of the action formula.
     * @param action the action referenced in the action formula.
     * @param instance the instance number of the action (may be null).
     * @param args the argument values of the action (may be null).
     * @return true if and only if the state satisfies the formula.
     ***************************************************************************/
    boolean checkActionFormula(StateClass state, ActionLTLTag.Kind kind, ActionSymbol action,
                               Integer instance, Value[] args) {
        // LTL formula "Executed action"
        if (kind == ActionLTLTag.Kind.executed) {
            return match(state.label, action, instance, args);
        }

        // LTL formula "Enabled action" (ltl.tag.kind == ActionLTLTag.Kind.enabled)
        // here we actually compute all successors of the given state to determine
        // whether any of these states has the denoted action label; while this
        // is a costly "brute-force" approach, it is by far the simplest one
        // to determine whether the action is enabled in the given state (mainly due
        // to the complex semantics of component actions in distributed systems)
        for (StateClass successor : getSuccessors(state)) {
            if (match(successor.label, action, instance, args)) {
                return true;
            }
        }
        return false;
    }

    /****************************************************************************
     * Returns true if action label matches the denoted action
     * (optionally extended by instance number and argument values).
     * @param label an action label.
     * @param action an action.
     * @param instance an instance number (may be null).
     * @param args an argument sequence (may be null).
     * @return true if and only if the matching succeeds.
     ***************************************************************************/
    private static boolean match(Label label, ActionSymbol action, Integer instance, Value[] args) {
        if (label == null || action != label.action
                || instance != null && instance != label.instance) {
            return false;
        }
        if (args != null) {
            int n = args.length;
            for (int i = 0; i < n; i++) {
                if (!args[i].equals(label.args[i])) {
                    return false;
                }
            }
        }
        return true;
    }

    // --------------------------------------------------------------------------
    //
    // the LTL model checker
    //
    // --------------------------------------------------------------------------

    /****************************************************************************
     * Check whether current system satisfies the LTL formula "ltl", return
     * null if yes, CounterExample object otherwise.
     ***************************************************************************/
    private CounterExample check(LTL ltl, List<LTL> fairness) {
        // the initial context for the translation of LTL formula to automaton
        Context context = new Context(system.getSlots());

        LTLFormula preprocessed = LTLPreprocessor.preprocess(ltl, context);
        LTLFormula negated = new LTLFormula.NegationNode(preprocessed);
        LTLFormula negationNormalForm = NNFConverter.convertToNNF(negated);

        AutomatonNode automaton = AutomatonCreator.createAutomaton(negationNormalForm);

        CounterExample counterExample = new EmptinessChecker(this, automaton, fairness).findCounterExample();

        PrintWriter out = Main.getOutput();
        out.println(successorsCache.size() + " system states and "
                + checkingStateCache.size() + " product automaton states investigated.");

        return counterExample;
    }

    public void emitDeadlockWarning(StateClass state) {
        if (!deadlockWarningEmitted) {
            PrintWriter out = Main.getOutput();
            out.println("Deadlock encountered for state: " + state.toString(system.getNames()));
            deadlockWarningEmitted = true;
        }
    }
}
// ----------------------------------------------------------------------------
// end of file
// ----------------------------------------------------------------------------