// ---------------------------------------------------------------------------
// LTLFormula.java
// Formulas internally used by the LTL model checker.
// $Id: LTLFormula.java,v 1.1 2022/08/24 12:47:14 schreine Exp $
//
// Author: Ágoston Sütő <agoston.suetoe@jku.at>
// Copyright (C) 2021-, Research Institute for Symbolic Computation (RISC)
// Johannes Kepler University, Linz, Austria, http://www.risc.jku.at
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program.  If not, see <http://www.gnu.org/licenses/>.
// ----------------------------------------------------------------------------
package riscal.ltl;

import riscal.semantics.Systems;
import riscal.semantics.Types;
import riscal.syntax.AST;
import riscal.types.Environment;

import java.util.Arrays;
import java.util.stream.Collectors;

public abstract class LTLFormula {

	public enum LTLOperator {
		AND("∧"),
		OR("∨"),
		UNTIL("U"),
		RELEASE("R");

		private final String val;

		LTLOperator(String val) {
			this.val = val;
		}

		@Override
		public String toString() {
			return val;
		}
	}

	public abstract <T, Q> T process(LTLProcessor<T, Q> processor, Q context);

	public abstract boolean isCompatibleWith(Systems.StateClass systemState, LTLChecker ltlChecker);

	public static class BooleanNode extends LTLFormula {
		public static final BooleanNode TRUE = new BooleanNode(true);
		public static final BooleanNode FALSE = new BooleanNode(false);

		public final boolean value;

		private BooleanNode(boolean value) {
			this.value = value;
		}

		@Override
		public String toString() {
			return value ? "T" : "F";
		}

		@Override
		public <T, Q> T process(LTLProcessor<T, Q> processor, Q context) {
			return processor.process(this, context);
		}

		@Override
		public boolean isCompatibleWith(Systems.StateClass systemState, LTLChecker ltlChecker) {
			return value;
		}
	}

	public static class NegationNode extends LTLFormula {
		public final LTLFormula child;

		public NegationNode(LTLFormula child) {
			this.child = child;
		}

		@Override
		public String toString() {
			return "¬(" + child.toString() + ")";
		}

		@Override
		public <T, Q> T process(LTLProcessor<T, Q> processor, Q context) {
			return processor.process(this, context);
		}

		@Override
		public boolean isCompatibleWith(Systems.StateClass systemState, LTLChecker ltlChecker) {
			return !child.isCompatibleWith(systemState, ltlChecker);
		}
	}

	public static class NextNode extends LTLFormula {
		public final LTLFormula child;

		public NextNode(LTLFormula child) {
			this.child = child;
		}

		@Override
		public String toString() {
			return "X(" + child.toString() + ")";
		}

		@Override
		public <T, Q> T process(LTLProcessor<T, Q> processor, Q context) {
			return processor.process(this, context);
		}

		@Override
		public boolean isCompatibleWith(Systems.StateClass systemState, LTLChecker ltlChecker) {
			return false;
		}
	}

	public static class OperatorNode extends LTLFormula {
		public final LTLOperator operator;
		public final LTLFormula left;
		public final LTLFormula right;

		public OperatorNode(LTLOperator operator, LTLFormula left, LTLFormula right) {
			this.operator =  operator;
			this.left = left;
			this.right = right;
		}

		@Override
		public String toString() {
			return "(" + left.toString() + " " + operator.toString() + " " + right.toString() + ")";
		}

		@Override
		public <T, Q> T process(LTLProcessor<T, Q> processor, Q context) {
			return processor.process(this, context);
		}

		@Override
		public boolean isCompatibleWith(Systems.StateClass systemState, LTLChecker ltlChecker) {
			return false;
		}
	}

	public static class StateNode extends LTLFormula {
		// only for serialization
		private final AST.LTL.StateLTL original;

		public final Types.ExpSem.Single sem;
		public final Types.Context context;

		public StateNode(AST.LTL.StateLTL original, Types.ExpSem.Single sem, Types.Context context) {
			this.original = original;
			this.sem = sem;
			this.context = context;
		}

		@Override
		public String toString() {
			return original.toString();
		}

		@Override
		public <T, Q> T process(LTLProcessor<T, Q> processor, Q context) {
			return processor.process(this, context);
		}

		@Override
		public boolean isCompatibleWith(Systems.StateClass systemState, LTLChecker ltlChecker) {
			return ltlChecker.checkStateFormula(systemState, context, sem);
		}
	}

	public static class ActionNode extends LTLFormula {
		// only for serialization
		public final AST.ActionLTLTag.Kind kind;
		public final Environment.Symbol.ActionSymbol action;
		public final Integer instance;
		public final Types.Value[] args;

		public ActionNode(AST.ActionLTLTag.Kind kind, Environment.Symbol.ActionSymbol action,
				  Integer instance, Types.Value[] args) {
			this.kind = kind;
			this.action = action;
			this.instance = instance;
			this.args = args;
		}

		@Override
		public String toString() {
			String argString = args != null ?
					"(" + Arrays.stream(args).map(Object::toString).collect(Collectors.joining(", ")) + ")" : "";
			return kind + " " + action.ident.string + argString;
		}

		@Override
		public <T, Q> T process(LTLProcessor<T, Q> processor, Q context) {
			return processor.process(this, context);
		}

		@Override
		public boolean isCompatibleWith(Systems.StateClass systemState, LTLChecker ltlChecker) {
			return ltlChecker.checkActionFormula(systemState, kind, action, instance, args);
		}
	}
}
// ----------------------------------------------------------------------------
// end of file
// ----------------------------------------------------------------------------
