// ---------------------------------------------------------------------------
// AutomatonCreator.java
// Generation of automata from LTL formulas.
// $Id: AutomatonCreator.java,v 1.2 2023/05/09 09:23:03 schreine Exp $
//
// Author: Ágoston Sütő <agoston.suetoe@jku.at>
// Copyright (C) 2021-, Research Institute for Symbolic Computation (RISC)
// Johannes Kepler University, Linz, Austria, http://www.risc.jku.at
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program.  If not, see <http://www.gnu.org/licenses/>.
// ----------------------------------------------------------------------------
package riscal.ltl.automaton;

import riscal.Main;
import riscal.ltl.LTLError;
import riscal.ltl.LTLFormula;
import riscal.ltl.LTLProcessor;

import java.util.*;
import java.util.stream.Collectors;

import static riscal.ltl.LTLFormula.LTLOperator.*;

/**
 * Generates a labelled generalized Büchi automaton A from the input
 * formula f such that every word in the language of A corresponds
 * to a model of f and vice-versa.
 * The algorithm is described in the paper
 * "Simple On-the-fly Automatic Verification of Linear Temporal Logic"
 * by R. Gerth, D. Peled, M. Y. Vardi, and P. Wolper
 */
public class AutomatonCreator {

	private static final FormulaProcessor processor = new FormulaProcessor();

	// dummy initial node for the tableau
	private static final TableauNode INIT = new TableauNode(null, null, null, null);

	public static AutomatonNode createAutomaton(LTLFormula root) {
		Set<TableauNode> graph = createTableau(root);
		Main.getOutput().println("Formula automaton with " + graph.size() + " states generated.");

		Map<TableauNode, AutomatonNode> tableauToAutomaton = new LinkedHashMap<>();

		// create final automaton nodes by copying only the relevant state formulas
		AutomatonNode initialNode = new AutomatonNode();
		tableauToAutomaton.put(INIT, initialNode);

		for (TableauNode tableauNode : graph) {
			AutomatonNode node = new AutomatonNode();
			node.formulas = tableauNode.oldFormulas.stream().filter(
					formula -> formula instanceof LTLFormula.StateNode
							|| formula instanceof LTLFormula.ActionNode
							|| formula instanceof LTLFormula.NegationNode
			).collect(Collectors.toList());

			tableauToAutomaton.put(tableauNode, node);
		}

		// finding the right children for each formula
		for (Map.Entry<TableauNode, AutomatonNode> entry : tableauToAutomaton.entrySet()) {
			entry.getValue().children = graph.stream()
					.filter(tableauNode -> tableauNode.incoming.contains(entry.getKey()))
					.map(tableauToAutomaton::get)
					.collect(Collectors.toList());
		}

		// set the right final states for each node
		List<LTLFormula.OperatorNode> untilSubformulas = UntilSubformulaFinder.findUntilSubformulas(root);
		if (untilSubformulas.size() >= Long.SIZE) {
			throw new LTLError(null, "LTL formula too complex");
		}
		initialNode.finalStates = untilSubformulas.size(); // mark the number of final state sets

		for (int i = 0; i < untilSubformulas.size(); i++) {
			for (Map.Entry<TableauNode, AutomatonNode> entry : tableauToAutomaton.entrySet()) {
				TableauNode tableauNode = entry.getKey();
				if (tableauNode.oldFormulas != null && (!tableauNode.oldFormulas.contains(untilSubformulas.get(i))
						|| tableauNode.oldFormulas.contains(untilSubformulas.get(i).right))) {
					entry.getValue().finalStates |= 1L << i;
				}
			}
		}

		return initialNode;
	}

	/**
	 * First step of the automaton creation algorithm, the creation
	 * of the tableau. An initial node is initialized with the root
	 * formula and INIT as a parent, and then expanded until it is
	 * fully processed
	 */
	private static Set<TableauNode> createTableau(LTLFormula root) {
		TableauNode start = new TableauNode(
				new HashSet<>(Collections.singleton(INIT)),
				new HashSet<>(Collections.singleton(root)),
				new HashSet<>(),
				new HashSet<>()
		);

		return expand(start, new LinkedHashSet<>());
	}

	private static Set<TableauNode> expand(TableauNode node, Set<TableauNode> nodesSet) {
		if (node.newFormulas.isEmpty()) {
			Optional<TableauNode> equivalentNode = nodesSet.stream()
					.filter(n -> node.oldFormulas.equals(n.oldFormulas) && node.nextFormulas.equals(n.nextFormulas))
					.findAny();

			if (equivalentNode.isPresent()) {
				equivalentNode.get().incoming.addAll(node.incoming);
				return nodesSet;
			} else {
				nodesSet.add(node);
				TableauNode newNode = new TableauNode(
						new HashSet<>(Collections.singleton(node)),
						new HashSet<>(node.nextFormulas),
						new HashSet<>(),
						new HashSet<>()
				);

				return expand(newNode, nodesSet);
			}
		} else {
			LTLFormula f = node.newFormulas.stream().findAny().get();
			node.newFormulas.remove(f);

			return f.process(processor, new ProcessingContext(node, nodesSet));
		}
	}

	private static class ProcessingContext {
		final TableauNode node;
		final Set<TableauNode> nodesSet;

		private ProcessingContext(TableauNode node, Set<TableauNode> nodesSet) {
			this.node = node;
			this.nodesSet = nodesSet;
		}
	}

	private static class FormulaProcessor implements LTLProcessor<Set<TableauNode>, ProcessingContext> {

		@Override
		public Set<TableauNode> process(LTLFormula.BooleanNode LTL, ProcessingContext context) {
			if (LTL.value) {
				return expand(context.node, context.nodesSet);
			}

			// current node contains a contradiction
			return context.nodesSet;
		}

		@Override
		public Set<TableauNode> process(LTLFormula.NegationNode LTL, ProcessingContext context) {
			if (context.node.oldFormulas.contains(LTL.child)) {
				// current node contains a contradiction
				return context.nodesSet;
			}

			context.node.oldFormulas.add(LTL);
			return expand(context.node, context.nodesSet);
		}

		@Override
		public Set<TableauNode> process(LTLFormula.NextNode LTL, ProcessingContext context) {
			context.node.oldFormulas.add(LTL);
			context.node.nextFormulas.add(LTL.child);
			return expand(context.node, context.nodesSet);
		}

		@Override
		public Set<TableauNode> process(LTLFormula.OperatorNode LTL, ProcessingContext context) {
			if (LTL.operator == LTLFormula.LTLOperator.AND) {
				if (!context.node.oldFormulas.contains(LTL.left)) {
					context.node.newFormulas.add(LTL.left);
				}
				if (!context.node.oldFormulas.contains(LTL.right)) {
					context.node.newFormulas.add(LTL.right);
				}
				context.node.oldFormulas.add(LTL);

				return expand(context.node, context.nodesSet);
			} else { // OR, UNTIL, RELEASE
				// compute the new and next sets of the first created node
				Set<LTLFormula> node1new = new1(LTL, context.node.newFormulas, context.node.oldFormulas);
				Set<LTLFormula> node1next = next1(LTL, context.node.nextFormulas);

				// for the second created node we reuse the old one by just modifying the old and new sets
				context.node.oldFormulas.add(LTL);
				new2(LTL, context.node.newFormulas, context.node.oldFormulas);

				TableauNode node1 = new TableauNode(
						new HashSet<>(context.node.incoming),
						node1new,
						new HashSet<>(context.node.oldFormulas),
						node1next
				);

				return expand(context.node, expand(node1, context.nodesSet));
			}
		}

		@Override
		public Set<TableauNode> process(LTLFormula.StateNode LTL, ProcessingContext context) {
			return handleActionAndStateFormulas(LTL, context);
		}

		@Override
		public Set<TableauNode> process(LTLFormula.ActionNode LTL, ProcessingContext context) {
			return handleActionAndStateFormulas(LTL, context);
		}

		private Set<TableauNode> handleActionAndStateFormulas(LTLFormula LTL, ProcessingContext context) {
			if (context.node.oldFormulas.stream().anyMatch(formula ->
					formula instanceof LTLFormula.NegationNode && ((LTLFormula.NegationNode) formula).child == LTL
			)) {
				// current node contains a contradiction
				return context.nodesSet;
			}

			context.node.oldFormulas.add(LTL);
			return AutomatonCreator.expand(context.node, context.nodesSet);
		}
	}

	private static Set<LTLFormula> new1(LTLFormula.OperatorNode f, Set<LTLFormula> newFormulas, Set<LTLFormula> oldFormulas) {
		Set<LTLFormula> res = new HashSet<>(newFormulas);
		if (f.operator == OR || f.operator == UNTIL) {
			if (!oldFormulas.contains(f.left)) {
				res.add(f.left);
			}
		} else if (f.operator == RELEASE) {
			if (!oldFormulas.contains(f.right)) {
				res.add(f.right);
			}
		}

		return res;
	}

	private static Set<LTLFormula> next1(LTLFormula.OperatorNode f, Set<LTLFormula> nextFormulas) {
		Set<LTLFormula> res = new HashSet<>(nextFormulas);
		if (f.operator == UNTIL || f.operator == RELEASE) {
			res.add(f);
		}

		return res;
	}

	private static void new2(LTLFormula.OperatorNode f, Set<LTLFormula> newFormulas, Set<LTLFormula> oldFormulas) {
		if (f.operator == OR || f.operator == UNTIL) {
			if (!oldFormulas.contains(f.right)) {
				newFormulas.add(f.right);
			}
		} else if (f.operator == RELEASE) {
			if (!oldFormulas.contains(f.left)) {
				newFormulas.add(f.left);
			}
			if (!oldFormulas.contains(f.right)) {
				newFormulas.add(f.right);
			}
		}
	}

	public static class TableauNode {
		public final Set<TableauNode> incoming;
		public final Set<LTLFormula> newFormulas;
		public final Set<LTLFormula> oldFormulas;
		public final Set<LTLFormula> nextFormulas;

		public TableauNode(Set<TableauNode> incoming, Set<LTLFormula> newFormulas,
				Set<LTLFormula> oldFormulas, Set<LTLFormula> nextFormulas) {
			this.incoming = incoming;
			this.newFormulas = newFormulas;
			this.oldFormulas = oldFormulas;
			this.nextFormulas = nextFormulas;
		}
	}

	private static class UntilSubformulaFinder implements LTLProcessor<Void, List<LTLFormula.OperatorNode>> {

		private static final UntilSubformulaFinder INSTANCE = new UntilSubformulaFinder();

		public static List<LTLFormula.OperatorNode> findUntilSubformulas(LTLFormula negationNormalForm) {
			List<LTLFormula.OperatorNode> untilSubformulas = new ArrayList<>();
			negationNormalForm.process(INSTANCE, untilSubformulas);
			return untilSubformulas;
		}

		@Override
		public Void process(LTLFormula.BooleanNode LTL, List<LTLFormula.OperatorNode> context) {
			return null;
		}

		@Override
		public Void process(LTLFormula.NegationNode LTL, List<LTLFormula.OperatorNode> context) {
			return null;
		}

		@Override
		public Void process(LTLFormula.NextNode LTL, List<LTLFormula.OperatorNode> context) {
			LTL.child.process(this, context);
			return null;
		}

		@Override
		public Void process(LTLFormula.OperatorNode LTL, List<LTLFormula.OperatorNode> context) {
			if (LTL.operator == UNTIL && LTL.right != LTLFormula.BooleanNode.TRUE) {
				context.add(LTL);
			}

			LTL.left.process(this, context);
			LTL.right.process(this, context);

			return null;
		}

		@Override
		public Void process(LTLFormula.StateNode LTL, List<LTLFormula.OperatorNode> context) {
			return null;
		}

		@Override
		public Void process(LTLFormula.ActionNode LTL, List<LTLFormula.OperatorNode> context) {
			return null;
		}
	}
}
// ----------------------------------------------------------------------------
// end of file
// ----------------------------------------------------------------------------

