// ---------------------------------------------------------------------------
// LTLPreprocessor.java
// A preprocessor for LTL formulas.
// $Id: LTLPreprocessor.java,v 1.1 2022/08/24 12:47:14 schreine Exp $
//
// Author: Ágoston Sütő <agoston.suetoe@jku.at>
// Copyright (C) 2021-, Research Institute for Symbolic Computation (RISC)
// Johannes Kepler University, Linz, Austria, http://www.risc.jku.at
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program.  If not, see <http://www.gnu.org/licenses/>.
// ----------------------------------------------------------------------------
package riscal.ltl.automaton;

import riscal.ltl.LTLChecker;
import riscal.ltl.LTLError;
import riscal.ltl.LTLFormula;
import riscal.ltl.Preprocessor;
import riscal.semantics.Translator;
import riscal.semantics.Types;
import riscal.syntax.AST;
import riscal.types.Environment;

import java.util.List;

import static riscal.ltl.LTLChecker.translator;
import static riscal.ltl.LTLFormula.LTLOperator.*;

/**
 * The automaton creation algorithm used in this implementation
 * {@link AutomatonCreator} requires the formula to contain only
 * the logical operators `and`, `or` and `not` and the temporal
 * operators `until`, `release` and `next time`. This class is
 * responsible for taking a general LTL formula, and replacing
 * all the other operators with these. It also expands the
 * quantifiers into a conjunction or disjunction and translates
 * the state and action formulas into a simpler form.
 */
public class LTLPreprocessor {

	private static final LTLPreprocessorImpl preprocessor = new LTLPreprocessorImpl();

	public static LTLFormula preprocess(AST.LTL ast, Types.Context context) {
		return preprocessor.preprocess(ast, context);
	}

	public static AST.ActionLTLTag.Kind getKind(AST.LTL.ActionLTL ast) {
		return ast.tag.kind;
	}

	public static Environment.Symbol.ActionSymbol getSymbol(AST.LTL.ActionLTL ast) {
		return (Environment.Symbol.ActionSymbol) ast.ident2.getSymbol();
	}

	public static Integer getInstance(AST.LTL.ActionLTL ast, Types.Context context) {
		// determine instance number (may be null)
		if (ast.exp != null) {
			Types.ExpSem isem = translator.process(ast.exp);
			if (!(isem instanceof Types.ExpSem.Single)) {
				throw new LTLError(ast.exp,
						"nondeterministic semantics not supported in LTL action formula expression");
			}
			Types.ExpSem.Single isem0 = (Types.ExpSem.Single) isem;
			Types.Value.Int ivalue = (Types.Value.Int) isem0.apply(context);
			return ivalue.getValue();
		}

		return null;
	}

	public static Types.Value[] getArgumentValues(AST.LTL.ActionLTL ast, Types.Context context) {
		// determine argument values (may be null)
		if (ast.exps != null) {
			int n = ast.exps.length;
			Types.Value[] args = new Types.Value[n];
			for (int i = 0; i < n; i++) {
				Types.ExpSem asem = translator.process(ast.exps[i]);
				if (!(asem instanceof Types.ExpSem.Single)) {
					throw new LTLError(ast.exps[i],
							"nondeterministic semantics not supported in LTL action formula expression");
				}
				Types.ExpSem.Single asem0 = (Types.ExpSem.Single) asem;
				args[i] = asem0.apply(context);
			}

			return args;
		}

		return null;
	}

	private static class LTLPreprocessorImpl implements Preprocessor<LTLFormula> {
		@Override
		public LTLFormula translate(AST.LTL.LetLTL ast, Types.Context context) {
			// compute context arising from binding
			Types.BinderSem[] bsem = translator.process(ast.binder);
			Types.Context context0 = context;
			for (int i = 0; i < bsem.length; i++) {
				if (!(bsem[i] instanceof Types.BinderSem.Single)) {
					throw new LTLError(ast.binder[i], "nondeterministic semantics not supported in LTL binder");
				}
				Types.BinderSem.Single bsem0 = (Types.BinderSem.Single) bsem[i];
				context0 = bsem0.apply(context0);
			}

			return preprocess(ast.ltl, context0);
		}

		@Override
		public LTLFormula translate(AST.LTL.LetParLTL ast, Types.Context context) {
			// compute context arising from binding
			Types.BinderSem bsem = translator.processPar(ast.binder);
			if (!(bsem instanceof Types.BinderSem.Single)) {
				throw new LTLError(ast, "nondeterministic semantics not supported in LTL binder");
			}

			Types.BinderSem.Single bsem0 = (Types.BinderSem.Single) bsem;
			Types.Context context0 = bsem0.apply(context);

			return preprocess(ast.ltl, context0);
		}

		private LTLFormula expandQuantifier(AST.LTL ltl, AST.QuantifiedVariable quantifier, Types.Context context,
				LTLFormula def, LTLFormula.LTLOperator connective) {
			List<Types.Context> contexts = LTLChecker.getContexts(context, quantifier);
			if (contexts.size() == 0) {
				return def;
			}

			LTLFormula result = preprocess(ltl, contexts.get(0));
			for (int i = 1; i < contexts.size(); i++) {
				LTLFormula current = preprocess(ltl, contexts.get(i));
				result = new LTLFormula.OperatorNode(connective, result, current);
			}

			return result;
		}

		@Override
		public LTLFormula translate(AST.LTL.ForallLTL ast, Types.Context context) {
			return expandQuantifier(ast.ltl, ast.qvar, context, LTLFormula.BooleanNode.TRUE, AND);
		}

		@Override
		public LTLFormula translate(AST.LTL.ExistsLTL ast, Types.Context context) {
			return expandQuantifier(ast.ltl, ast.qvar, context, LTLFormula.BooleanNode.FALSE, OR);
		}

		@Override
		public LTLFormula translate(AST.LTL.StateLTL ast, Types.Context context) {
			// compute expression semantics arising from state formula
			Types.ExpSem sem = translator.process(ast.exp);
			if (!(sem instanceof Types.ExpSem.Single)) {
				throw new LTLError(ast.exp,
						"nondeterministic semantics not supported in LTL state formula expression");
			}

			return new LTLFormula.StateNode(ast, (Types.ExpSem.Single) sem, context);
		}

		@Override
		public LTLFormula translate(AST.LTL.ActionLTL ast, Types.Context context) {
			switch (getKind(ast)) {
				case enabled:
				case executed:
					return translateActionFormula(ast, context);
				case weakfairness:
				case strongfairness:
					return translateFairnessFormula(ast, context);
				default:
					throw new Translator.RuntimeError(ast, "No kind set for action formula.");
			}
		}

		private LTLFormula translateActionFormula(AST.LTL.ActionLTL ast, Types.Context context) {
			return new LTLFormula.ActionNode(getKind(ast), getSymbol(ast),
					getInstance(ast, context), getArgumentValues(ast, context));
		}

		private LTLFormula translateFairnessFormula(AST.LTL.ActionLTL ast, Types.Context context) {
			AST.ActionLTLTag.Kind kind = getKind(ast);
			Environment.Symbol.ActionSymbol symbol = getSymbol(ast);
			Integer instance = getInstance(ast, context);
			Types.Value[] args = getArgumentValues(ast, context);

			// WeakFairness a == (<>[] Enabled a) => ([]<> Executed a)
			// StrongFairness a == ([]<> Enabled a) => ([]<> Executed a)
			LTLFormula lhs;
			if (kind == AST.ActionLTLTag.Kind.weakfairness) {
				// equivalent to !(<>[] Enabled a)
				lhs = new LTLFormula.OperatorNode(
						RELEASE,
						LTLFormula.BooleanNode.FALSE,
						new LTLFormula.OperatorNode(
								UNTIL,
								LTLFormula.BooleanNode.TRUE,
								new LTLFormula.NegationNode(
										new LTLFormula.ActionNode(AST.ActionLTLTag.Kind.enabled, symbol, instance, args)
								)
						)
				);
			} else {
				// equivalent to !([]<> Enabled a)
				lhs = new LTLFormula.OperatorNode(
						UNTIL,
						LTLFormula.BooleanNode.TRUE,
						new LTLFormula.OperatorNode(
								RELEASE,
								LTLFormula.BooleanNode.FALSE,
								new LTLFormula.NegationNode(
										new LTLFormula.ActionNode(AST.ActionLTLTag.Kind.enabled, symbol, instance, args)
								)
						)
				);
			}

			LTLFormula rhs = new LTLFormula.OperatorNode(
					RELEASE,
					LTLFormula.BooleanNode.FALSE,
					new LTLFormula.OperatorNode(
							UNTIL,
							LTLFormula.BooleanNode.TRUE,
							new LTLFormula.ActionNode(AST.ActionLTLTag.Kind.executed, symbol, instance, args)
					)
			);

			return new LTLFormula.OperatorNode(OR, lhs, rhs);
		}

		@Override
		public LTLFormula translate(AST.LTL.NotLTL ast, Types.Context context) {
			return new LTLFormula.NegationNode(preprocess(ast.ltl, context));
		}

		@Override
		public LTLFormula translate(AST.LTL.AndLTL ast, Types.Context context) {
			return new LTLFormula.OperatorNode(
					AND,
					preprocess(ast.ltl1, context),
					preprocess(ast.ltl2, context)
			);
		}

		@Override
		public LTLFormula translate(AST.LTL.OrLTL ast, Types.Context context) {
			return new LTLFormula.OperatorNode(
					OR,
				  preprocess(ast.ltl1, context),
					preprocess(ast.ltl2, context)
			);
		}

		@Override
		public LTLFormula translate(AST.LTL.ImpliesLTL ast, Types.Context context) {
			return new LTLFormula.OperatorNode(
					OR,
					new LTLFormula.NegationNode(preprocess(ast.ltl1, context)),
					preprocess(ast.ltl2, context)
			);
		}

		@Override
		public LTLFormula translate(AST.LTL.EquivLTL ast, Types.Context context) {
			// f <-> g == (f && g) || (!f && !g)

			LTLFormula left = preprocess(ast.ltl1, context);
			LTLFormula right = preprocess(ast.ltl2, context);

			return new LTLFormula.OperatorNode(
					OR,
					new LTLFormula.OperatorNode(AND, left, right),
					new LTLFormula.OperatorNode(
							AND,
							new LTLFormula.NegationNode(left),
							new LTLFormula.NegationNode(right)
					)
			);
		}

		@Override
		public LTLFormula translate(AST.LTL.IfThenElseLTL ast, Types.Context context) {
			// IF f THEN g ELSE h == (f && g) || (!f && h)

			LTLFormula condition = preprocess(ast.ltl1, context);
			return new LTLFormula.OperatorNode(
					OR,
					new LTLFormula.OperatorNode(
							AND,
							condition,
							preprocess(ast.ltl2, context)
					),
					new LTLFormula.OperatorNode(
							AND,
							new LTLFormula.NegationNode(condition),
							preprocess(ast.ltl3, context)
					)
			);
		}

		@Override
		public LTLFormula translate(AST.LTL.NextLTL ast, Types.Context context) {
			return new LTLFormula.NextNode(preprocess(ast.ltl, context));
		}

		@Override
		public LTLFormula translate(AST.LTL.FinallyLTL ast, Types.Context context) {
			// F f == True U f

			return new LTLFormula.OperatorNode(
					UNTIL,
					LTLFormula.BooleanNode.TRUE,
					preprocess(ast.ltl, context)
			);
		}

		@Override
		public LTLFormula translate(AST.LTL.GloballyLTL ast, Types.Context context) {
			// G f == False R f

			return new LTLFormula.OperatorNode(
					RELEASE,
					LTLFormula.BooleanNode.FALSE,
					preprocess(ast.ltl, context)
			);
		}

		@Override
		public LTLFormula translate(AST.LTL.UntilLTL ast, Types.Context context) {
			return new LTLFormula.OperatorNode(
					UNTIL,
					preprocess(ast.ltl1, context),
					preprocess(ast.ltl2, context)
			);
		}

		@Override
		public LTLFormula translate(AST.LTL.WeakUntilLTL ast, Types.Context context) {
			// f W g == (f U g) || G f == f U (g || G f) == g R (g || f)
			// here we use the last equivalence as that produces the simplest
			// automaton using this algorithm

			LTLFormula left = preprocess(ast.ltl1, context);
			LTLFormula right = preprocess(ast.ltl2, context);

			return new LTLFormula.OperatorNode(
					RELEASE,
					right,
					new LTLFormula.OperatorNode(OR, right, left)
			);
		}
	}
}
// ----------------------------------------------------------------------------
// end of file
// ----------------------------------------------------------------------------
