// ---------------------------------------------------------------------------
// EmptinessChecker.java
// Checking the emptiness of the product automaton in the LTL model checker.
// $Id: EmptinessChecker.java,v 1.4 2023/07/03 07:12:26 schreine Exp schreine $
//
// Author: Ágoston Sütő <agoston.suetoe@jku.at>
// Copyright (C) 2021-, Research Institute for Symbolic Computation (RISC)
// Johannes Kepler University, Linz, Austria, http://www.risc.jku.at
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program.  If not, see <http://www.gnu.org/licenses/>.
// ----------------------------------------------------------------------------
package riscal.ltl.search;

import riscal.Main;
import riscal.ltl.automaton.AutomatonNode;
import riscal.ltl.LTLChecker;
import riscal.ltl.Utilities.Pair;
import riscal.syntax.AST;

import java.util.*;

/**
 * This class is for checking if the product automaton
 * of the formula and the system contains a reachable
 * cycle with a state from every accepting set and one
 * that also satisfies the fairness conditions, if there
 * are any.
 */
public class EmptinessChecker {

	private final LTLChecker ltlChecker;

	private final AutomatonNode initialState;
	private final /* bitset */ long finalCheck;

	private final FairnessChecker fairnessChecker;

	public EmptinessChecker(LTLChecker ltlChecker, AutomatonNode initialState, List<AST.LTL> fairness) {
		this.ltlChecker = ltlChecker;

		this.initialState = initialState;
		this.finalCheck = (1L << initialState.finalStates) - 1;

		if (fairness != null) {
			this.fairnessChecker = new FairnessChecker(ltlChecker, finalCheck, fairness);
		} else {
			this.fairnessChecker = null;
		}
	}

	public CounterExample findCounterExample() {
		Long maxSearchDepth = Main.getCheckDepth();
		return findSCCs(new CheckingState(ltlChecker, initialState, null), maxSearchDepth);
	}

	/**
	 * The algorithm for finding SCCs which contain a state from
	 * each accepting set is taken from
	 * "A Note on On-the-Fly Verification Algorithms"
	 * by Stefan Schwoon and Javier Esparza.
	 * It is based on Couvreur's modifications of Tarjan's SCC
	 * algorithm. This is an iterative implementation to make sure
	 * it uses constant stack space and to make the search path
	 * available for printing.
	 * The algorithm is modified slightly, by always processing
	 * MSCCs instead of terminating early and by adding the additional
	 * fairness checking if needed.
	 */
	private CounterExample findSCCs(CheckingState initialState, Long maxSearchDepth) {
		int count = 0;
		Deque<CheckingState> searchPath = new ArrayDeque<>();

		Deque<Pair<CheckingState, Long>> roots = new ArrayDeque<>();
		Deque<CheckingState> actives = new ArrayDeque<>();

		searchPath.push(initialState);

		while (!searchPath.isEmpty()) {
			CheckingState state = searchPath.peek();

			if (state.isUnprocessed()) {
				count++;
				state.dfsnum = count;
				state.current = true;

				roots.push(new Pair<>(state, state.automatonNode.finalStates));
				actives.push(state);
			}

			for (CheckingState successor = state.getFirstSuccessor();
				 successor != null; successor = state.getNextSuccessor()) {
			  Main.checkStopped();
				if (successor.dfsnum == 0) {
					if (maxSearchDepth == null || searchPath.size() <= maxSearchDepth) {
						searchPath.push(successor);
					}
					break;
				} else if (successor.current) {
					long finalsFound = 0;

					Pair<CheckingState, Long> root;
					do {
						root = roots.pop();
						finalsFound |= root.second;
					} while (root.first.dfsnum > successor.dfsnum);

					roots.push(new Pair<>(root.first, finalsFound));
				}
			}

			if (state == searchPath.peek()) {
				if (roots.peek().first == state) {
					// we just finished processing an MSCC

					if (roots.peek().second == finalCheck
							&& (isDeadlockState(state) || actives.peek() != state)) {
						// admissible MSCC found
						roots.pop();

						List<CheckingState> MSCC = new ArrayList<>();

						CheckingState active;
						do {
							active = actives.pop();
							active.current = false;
							MSCC.add(active);
						} while (active != state);

						if (fairnessChecker != null) {
							// checking under fairness -> we have to check whether it contains a fair sub-SCC
							Pair<List<CheckingState>, List<LoopConstraint>> result = fairnessChecker.findFairSCC(MSCC);
							if (result != null) {
								return findValidLoopWithFairness(searchPath, MSCC, result.first, result.second);
							}
						} else {
							return findValidLoop(searchPath, MSCC);
						}
					} else {
						roots.pop();

						CheckingState active;
						do {
							active = actives.pop();
							active.current = false;
						} while (active != state);
					}
				}

				searchPath.pop();
			}
		}

		return null;
	}

	/**
	 * Find a cycle in the SCC which contains a state from each final set.
	 */
	private CounterExample findValidLoop(Deque<CheckingState> pathToScc, List<CheckingState> SCC) {
		List<LoopConstraint> finalSetConstraints =
				LoopConstraint.createFinalStateConstraints(initialState.finalStates);

		LoopFinder loopFinder = new LoopFinder(ltlChecker);
		List<CheckingState> loop = loopFinder.findLoop(SCC, finalSetConstraints);

		return createCounterExample(pathToScc, new ArrayList<>(), loop);
	}

	/**
	 * Find the path the initial state of the fair SCC and a cycle in the
	 * fair SCC which contains states:
	 *  from each final set
	 *  which satisfy every weak fairness
	 *  which satisfy every remaining strong fairness (this means there is
	 *   a state in the SCC in which the action is enabled, so we have to
	 *   find a state in which it is executed)
	 */
	private CounterExample findValidLoopWithFairness(Deque<CheckingState> pathToScc, List<CheckingState> MSCC,
			List<CheckingState> fairSCC, List<LoopConstraint> remainingConstraints) {
		List<LoopConstraint> constraints = new ArrayList<>();

		constraints.addAll(LoopConstraint.createFinalStateConstraints(initialState.finalStates));
		constraints.addAll(remainingConstraints);

		List<CheckingState> pathToFairSCC = new ArrayList<>(Collections.singleton(MSCC.get(MSCC.size() - 1)));
		LoopFinder.breadthFirstSearch(MSCC, pathToFairSCC, state -> state == fairSCC.get(fairSCC.size() - 1));

		LoopFinder loopFinder = new LoopFinder(ltlChecker);
		List<CheckingState> loop = loopFinder.findLoop(fairSCC, constraints);

		return createCounterExample(pathToScc, pathToFairSCC, loop);
	}

	private CounterExample createCounterExample(Deque<CheckingState> pathToScc,
			List<CheckingState> pathToCycle, List<CheckingState> loop) {
		if (loop.stream().anyMatch(this::isDeadlockState)) {
			// the counterexample resulted from a deadlock / a situation in which there were no further
			// states, so do not display a loop
			return new CounterExample(pathToScc, pathToCycle, null);
		} else {
			return new CounterExample(pathToScc, pathToCycle, loop);
		}
	}

	private Boolean isDeadlockState(CheckingState state) {
		return ltlChecker.getSuccessors(state.stateNode).isEmpty();
	}
}
// ----------------------------------------------------------------------------
// end of file
// ----------------------------------------------------------------------------

