// ---------------------------------------------------------------------------
// FairnessChecker.java
// Checking fairness constraints in the LTL model checker.
// $Id: FairnessChecker.java,v 1.3 2023/05/09 09:23:03 schreine Exp $
//
// Author: Ágoston Sütő <agoston.suetoe@jku.at>
// Copyright (C) 2021-, Research Institute for Symbolic Computation (RISC)
// Johannes Kepler University, Linz, Austria, http://www.risc.jku.at
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program.  If not, see <http://www.gnu.org/licenses/>.
// ----------------------------------------------------------------------------
package riscal.ltl.search;

import riscal.ltl.LTLChecker;
import riscal.ltl.Utilities.Pair;
import riscal.syntax.AST;

import java.util.*;

/**
 * The fairness checking is based on the algorithm FAIR-SUB from
 * The Temporal Verification of Reactive Systems: Progress by Manna and
 * Pnuelli available at https://theory.stanford.edu/~zm/tvors3.html
 */
public class FairnessChecker {

	private final LTLChecker ltlChecker;

	private final long finalCheck;

	private final List<FairnessConstraint> weakConstraints;
	private final List<FairnessConstraint> strongConstraints;

	public FairnessChecker(LTLChecker ltlChecker, long finalCheck, List<AST.LTL> fairness) {
		this.ltlChecker = ltlChecker;
		this.finalCheck = finalCheck;

		Pair<List<FairnessConstraint>, List<FairnessConstraint>> constraints
				= FairnessConstraint.makeFairnessConstraints(fairness);
		this.weakConstraints = constraints.first;
		this.strongConstraints = constraints.second;
	}

	public Pair<List<CheckingState>, List<LoopConstraint>> findFairSCC(List<CheckingState> SCC) {
		// copy the original SCC, because some states might be deleted from it when it is processed
		return findFairSCC(new ArrayList<>(SCC), new ArrayList<>(strongConstraints));
	}

	private Pair<List<CheckingState>, List<LoopConstraint>> findFairSCC(
			List<CheckingState> SCC, List<FairnessConstraint> remainingStrongConstraints) {
		List<FairnessConstraint> remainingWeakConstraints = new ArrayList<>(weakConstraints);

		for (CheckingState state : SCC) {
			if (remainingWeakConstraints.isEmpty()) {
				break;
			}

			remainingWeakConstraints.removeIf(weakFairness
					-> weakFairness.executed.isCompatibleWith(state.stateNode, ltlChecker)
					|| !weakFairness.enabled.isCompatibleWith(state.stateNode, ltlChecker));
		}

		if (!remainingWeakConstraints.isEmpty()) {
			return null;
		}

		remainingStrongConstraints.removeIf(strongFairness ->
				SCC.stream().noneMatch(state -> strongFairness.enabled.isCompatibleWith(state.stateNode, ltlChecker)));

		boolean allExecuted = remainingStrongConstraints.stream().allMatch(strongFairness ->
				SCC.stream().anyMatch(state -> strongFairness.executed.isCompatibleWith(state.stateNode, ltlChecker))
		);

		if (allExecuted) {
			List<LoopConstraint> remainingConstraints = new ArrayList<>();
			remainingConstraints.addAll(weakConstraints);
			remainingConstraints.addAll(remainingStrongConstraints);

			return new Pair<>(SCC, remainingConstraints);
		}

		for (FairnessConstraint strongFairness : remainingStrongConstraints) {
			SCC.removeIf(checkingState -> strongFairness.enabled.isCompatibleWith(checkingState.stateNode, ltlChecker));
		}

		while (SCC.size() > 0) {
			Pair<List<CheckingState>, List<LoopConstraint>> result = findSubSCC(SCC, remainingStrongConstraints);
			if (result != null) {
				return result;
			}
		}

		return null;
	}

	/**
	 * For finding the admissible sub-SCCs we use the same algorithm
	 * as in the initial partition into SCCs with some slight
	 * technical modifications, since we cannot reuse the fields of
	 * the CheckingState.
	 */
	private Pair<List<CheckingState>, List<LoopConstraint>> findSubSCC(List<CheckingState> remainingStates,
			List<FairnessConstraint> remainingStrongConstraints) {
		int count = 0;
		Deque<CheckingState> searchPath = new ArrayDeque<>();

		Deque<Pair<CheckingState, Long>> roots = new ArrayDeque<>();
		Deque<CheckingState> actives = new ArrayDeque<>();

		Map<CheckingState, Integer> dfsnums = new HashMap<>();
		Set<CheckingState> current = new HashSet<>();

		searchPath.push(remainingStates.get(0));

		while (!searchPath.isEmpty()) {
			CheckingState state = searchPath.peek();

			if (!dfsnums.containsKey(state)) {
				count++;
				dfsnums.put(state, count);
				current.add(state);

				roots.push(new Pair<>(state, state.automatonNode.finalStates));
				actives.push(state);
			}

			for (CheckingState successor = state.getFirstSuccessor(); successor != null; successor = state.getNextSuccessor()) {
				if (!remainingStates.contains(successor)) {
					continue;
				}

				if (dfsnums.getOrDefault(successor, 0) == 0) {
					searchPath.push(successor);
					break;
				} else if (current.contains(successor)) {
					long finalsFound = 0;

					Pair<CheckingState, Long> root;
					do {
						root = roots.pop();
						finalsFound |= root.second;
					} while (dfsnums.getOrDefault(root.first, 0) > dfsnums.getOrDefault(successor, 0) && roots.size() > 0);

					roots.push(new Pair<>(root.first, finalsFound));
				}
			}

			if (state == searchPath.peek()) {
				if (roots.peek().first == state) {
					if (roots.peek().second == finalCheck && actives.peek() != state) {
						roots.pop();

						List<CheckingState> MSCC = new ArrayList<>();

						CheckingState active;
						do {
							active = actives.pop();
							current.remove(active);
							remainingStates.remove(active);
							MSCC.add(active);
						} while (active != state);

						Pair<List<CheckingState>, List<LoopConstraint>> fairSCC
								= findFairSCC(MSCC, new ArrayList<>(remainingStrongConstraints));
						if (fairSCC != null) {
							return fairSCC;
						}
					} else {
						roots.pop();

						CheckingState active;
						do {
							active = actives.pop();
							current.remove(active);
							remainingStates.remove(active);
						} while (active != state);
					}
				}

				searchPath.pop();
			}
		}

		return null;
	}
}
// ----------------------------------------------------------------------------
// end of file
// ----------------------------------------------------------------------------

