// ---------------------------------------------------------------------------
// FairnessConstraint.java
// Fairness constraints in the LTL model checker.
// $Id: FairnessConstraint.java,v 1.1 2022/08/24 12:47:14 schreine Exp $
//
// Author: Ágoston Sütő <agoston.suetoe@jku.at>
// Copyright (C) 2021-, Research Institute for Symbolic Computation (RISC)
// Johannes Kepler University, Linz, Austria, http://www.risc.jku.at
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program.  If not, see <http://www.gnu.org/licenses/>.
// ----------------------------------------------------------------------------
package riscal.ltl.search;

import riscal.ltl.LTLChecker;
import riscal.ltl.LTLFormula;
import riscal.ltl.automaton.LTLPreprocessor;
import riscal.semantics.Translator;
import riscal.semantics.Types;
import riscal.ltl.Utilities.Pair;
import riscal.syntax.AST;
import riscal.syntax.AST.ActionLTLTag.Kind;
import riscal.types.Environment;

import java.util.ArrayList;
import java.util.List;

public class FairnessConstraint implements LoopConstraint {

	private enum Type { Weak, Strong }

	private final Type type;
	public final LTLFormula.ActionNode enabled;
	public final LTLFormula.ActionNode executed;

	public FairnessConstraint(Type type, LTLFormula.ActionNode enabled, LTLFormula.ActionNode executed) {
		this.type = type;
		this.enabled = enabled;
		this.executed = executed;
	}

	@Override
	public boolean checkState(CheckingState state, LTLChecker ltlChecker) {
		switch (type) {
		case Weak:
			return !enabled.isCompatibleWith(state.stateNode, ltlChecker)
					|| executed.isCompatibleWith(state.stateNode, ltlChecker);
		case Strong:
			return executed.isCompatibleWith(state.stateNode, ltlChecker);
		default:
			throw new RuntimeException("Invalid fairness constraint");
		}
	}

	static Pair<List<FairnessConstraint>, List<FairnessConstraint>> makeFairnessConstraints(List<AST.LTL> fairness) {
		List<FairnessConstraint> weakFairness = new ArrayList<>();
		List<FairnessConstraint> strongFairness = new ArrayList<>();

		for (AST.LTL formula : fairness) {
			if (formula instanceof AST.LTL.ActionLTL) {
				AST.LTL.ActionLTL actionFormula = (AST.LTL.ActionLTL) formula;
				FairnessConstraint constraint = makeFairnessConstraint(new Types.Context(0), actionFormula);

				Kind kind = LTLPreprocessor.getKind(actionFormula);
				if (kind == Kind.weakfairness) {
					weakFairness.add(constraint);
				} else if (kind == Kind.strongfairness) {
					strongFairness.add(constraint);
				} else {
					throw new Translator.RuntimeError(actionFormula, "Invalid fairness constraint");
				}
			} else if (formula instanceof AST.LTL.ForallLTL) {
				AST.LTL.ForallLTL forallLTL = (AST.LTL.ForallLTL) formula;
				int slots = forallLTL.qvar.qvcore.length;
				List<Types.Context> contexts = LTLChecker.getContexts(new Types.Context(slots), forallLTL.qvar);

				for (Types.Context context : contexts) {
					AST.LTL.ActionLTL actionFormula = (AST.LTL.ActionLTL) forallLTL.ltl;
					FairnessConstraint constraint = makeFairnessConstraint(context, actionFormula);

					Kind kind = LTLPreprocessor.getKind(actionFormula);
					if (kind == Kind.weakfairness) {
						weakFairness.add(constraint);
					} else if (kind == Kind.strongfairness) {
						strongFairness.add(constraint);
					} else {
						throw new Translator.RuntimeError(actionFormula, "Invalid fairness constraint");
					}
				}
			} else {
				throw new Translator.RuntimeError(formula, "Invalid fairness constraint");
			}
		}

		return new Pair<>(weakFairness, strongFairness);
	}

	private static FairnessConstraint makeFairnessConstraint(Types.Context context, AST.LTL.ActionLTL actionFormula) {
		Kind kind = LTLPreprocessor.getKind(actionFormula);
		Environment.Symbol.ActionSymbol symbol = LTLPreprocessor.getSymbol(actionFormula);
		Integer instance = LTLPreprocessor.getInstance(actionFormula, context);
		Types.Value[] args = LTLPreprocessor.getArgumentValues(actionFormula, context);

		LTLFormula.ActionNode enabled = new LTLFormula.ActionNode(Kind.enabled, symbol, instance, args);
		LTLFormula.ActionNode executed = new LTLFormula.ActionNode(Kind.executed, symbol, instance, args);

		return new FairnessConstraint(kind == Kind.weakfairness ? Type.Weak : Type.Strong, enabled, executed);
	}
}
// ----------------------------------------------------------------------------
// end of file
// ----------------------------------------------------------------------------
