// ---------------------------------------------------------------------------
// LoopFinder.java
// Finding loops in the product automaton for the LTL model checker.
// $Id: LoopFinder.java,v 1.1 2022/08/24 12:47:14 schreine Exp $
//
// Author: Ágoston Sütő <agoston.suetoe@jku.at>
// Copyright (C) 2021-, Research Institute for Symbolic Computation (RISC)
// Johannes Kepler University, Linz, Austria, http://www.risc.jku.at
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program.  If not, see <http://www.gnu.org/licenses/>.
// ----------------------------------------------------------------------------
package riscal.ltl.search;

import riscal.ltl.LTLChecker;

import java.util.*;
import java.util.function.Predicate;

public class LoopFinder {

	private final LTLChecker ltlChecker;

	public LoopFinder(LTLChecker ltlChecker) {
		this.ltlChecker = ltlChecker;
	}

	public List<CheckingState> findLoop(List<CheckingState> SCC, List<LoopConstraint> constraints) {
		List<CheckingState> loop = new ArrayList<>();

		CheckingState initialState = SCC.get(SCC.size() - 1);
		constraints.removeIf(constraint -> constraint.checkState(initialState, ltlChecker));
		loop.add(initialState);

		while (!constraints.isEmpty()) {
			breadthFirstSearch(SCC, loop, state -> constraints.removeIf(c -> c.checkState(state, ltlChecker)));
		}

		if (loop.size() == 1) {
			// we got really lucky and the initial state satisfies all the constraints
			for (CheckingState successor = initialState.getFirstSuccessor();
				 successor != null; successor = initialState.getNextSuccessor()) {
				if (SCC.contains(successor)) {
					loop.add(successor);
					break;
				}
			}
		}

		if (loop.get(loop.size() - 1) != initialState) {
			// if we're not back to the initial state yet, we run one last BFS to find it
			breadthFirstSearch(SCC, loop, state -> state == initialState);
		}

		// we indeed have a loop
		assert initialState == loop.get(loop.size() - 1);

		return loop;
	}

	public static void breadthFirstSearch(List<CheckingState> SCC, List<CheckingState> loop,
			Predicate<CheckingState> isGoalState) {
		CheckingState searchStart = loop.get(loop.size() - 1);

		Deque<CheckingState> states = new ArrayDeque<>();
		Map<CheckingState, CheckingState> ancestors = new HashMap<>();

		states.push(searchStart);
		ancestors.put(searchStart, null);

		while (!states.isEmpty()) {
			CheckingState current = states.pollLast();

			boolean goalFound = isGoalState.test(current);
			if (goalFound) {
				int last = loop.size();

				CheckingState x = current;
				while (ancestors.get(x) != null) {
					loop.add(last, x);
					x = ancestors.get(x);
				}
				return;
			} else {
				for (CheckingState successor = current.getFirstSuccessor(); successor != null; successor = current.getNextSuccessor()) {
					if (SCC.contains(successor) && !ancestors.containsKey(successor)) {
						ancestors.put(successor, current);
						states.addFirst(successor);
					}
				}
			}
		}

		throw new RuntimeException("No goal state found in breadth first search");
	}
}
// ----------------------------------------------------------------------------
// end of file
// ----------------------------------------------------------------------------
