// ---------------------------------------------------------------------------
// RISCTP.java
// Translation from RISCAL to the RISCTP theorem proving interface.
// $Id: RISCTP.java,v 1.54 2022/09/16 07:41:28 schreine Exp $
//
// Author: Wolfgang Schreiner <Wolfgang.Schreiner@risc.jku.at>
// Copyright (C) 2022-, Research Institute for Symbolic Computation (RISC)
// Johannes Kepler University, Linz, Austria, https://www.risc.jku.at
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program.  If not, see <https://www.gnu.org/licenses/>.
// ----------------------------------------------------------------------------
package riscal.risctp;

import java.util.*;
import java.util.function.*;

import riscal.Main;
import risctp.syntax.*;
import risctp.syntax.AST.*;
import risctp.syntax.AST.Decl.*;
import risctp.syntax.AST.Decl.Function;
import risctp.syntax.AST.Exp.*;
import risctp.syntax.AST.Pattern.*;
import risctp.types.*;
import risctp.types.Symbol.*;
import risctp.problem.*;

public class RISCTP
{
  /****************************************************************************
   * Translate RISCAL declarations to a RISCTP proof problem.
   * @param decls a sequence of RISCAL declarations.
   * @return a type-checked RISCTP proof problem (null, in case of an error)
   ***************************************************************************/
  public static ProofProblem translate(List<riscal.syntax.AST.Declaration> decls) 
  {
    RISCTP translator = new RISCTP();
    translator.translateCore(decls);
    if (translator.checker == null) return null;
    return translator.checker.getResult();
  }
  
  /***************************************************************************
   * The exception raised by the translator on an error.
   **************************************************************************/
  private static final class TranslationError extends RuntimeException
  {
    private static final long serialVersionUID = 13062022L;
    
    /*************************************************************************
     * Create a type error with the denoted message.
     * @param msg the message
     ************************************************************************/
    private TranslationError(String msg) 
    { 
      super(msg); 
    }
    public static TranslationError raise(String msg)
    {
      throw new TranslationError(msg);
    }
  }
  
  // the caller map
  private riscal.tasks.Recursion.CallMap callers;

  // the typechecker used for processing the declarations
  private TypeChecker checker;
  
  // the symbols of the already declared entities
  private Set<riscal.types.Environment.Symbol> symbols;
  
  // the currently processed declaration
  private riscal.syntax.AST.Declaration currentDecl;
  
  // the commands to be executed after processing the declaration
  // private List<Runnable> postDecl;
  
  // the map of integer bound expressions (as strings) to the bounded integer type
  private Map<String,Map<String,Type>> intMap;
  private int intMapSize; // the total number of types in the map
  
  // the map of string representations of structured types to type names
  private Map<String,String> typeMap;
  
  // map of element types to the corresponding set types
  // private Map<TypeSymbol,TypeSymbol> setMap;

  private RISCTP()
  {
    this.callers = riscal.Main.getCallers();
    this.checker = new TypeChecker();
    this.symbols = new HashSet<riscal.types.Environment.Symbol>();
    this.currentDecl = null;
    // this.postDecl = new ArrayList<Runnable>();
    this.intMap = new HashMap<String,Map<String,Type>>();
    this.intMapSize = 0;
    this.typeMap = new HashMap<String,String>();
    // this.setMap = new HashMap<TypeSymbol,TypeSymbol>();
  }

  /****************************************************************************
   * Process a declaration, adding it to the generated proof problem.
   * @param decl the declaration
   ***************************************************************************/
  private void declare(Decl decl)
  {
    if (checker != null) checker.checkDeclaration(decl);
  }
  
  // --------------------------------------------------------------------------
  //
  // builtin entities
  //
  // --------------------------------------------------------------------------

  // special names
  private static final String UNITNAME = "_Unit";
  private static final String SUMNAME = "_sum";
  private static final String PRODUCTNAME = "_product";
  private static final String SIZENAME = "_size";
  private static final String POWERNAME = "_power";
  
  /****************************************************************************
   * Declare builtin entities.
   ***************************************************************************/
  private void declareBuiltins()
  {
    declare(TypeDecl.create(Id.create(UNITNAME),
        type(AST.BOOLNAME),
        Equiv.create(var(AST.VALUENAME), False.create())));
    intNumberFun(POWERNAME, "1", AST.MULTNAME, (Exp e)->e);
    intIntervalFun(SUMNAME, "0", AST.PLUSNAME, (Exp e)->e);
    intIntervalFun(PRODUCTNAME, "1", AST.MULTNAME, (Exp e)->e);
    intSetFun(SUMNAME, "0", AST.PLUSNAME, (Exp e)->e);
    intSetFun(PRODUCTNAME, "1", AST.MULTNAME, (Exp e)->e);
  }
 
  /****************************************************************************
   * Create integer function iterating a certain number of times
   * @param name the name of the function.
   * @param base the value of the function if the iteration number is zero.
   * @param combine the combination function.
   * @param value the function to apply to the element value.
   ***************************************************************************/
  private void intNumberFun(String name, String base, String combine,
    java.util.function.Function<Exp,Exp> value)
  {
    declare(Function.create(Id.create(name),
        Arrays.asList(new TypedVar[] { tvar("x", AST.INTNAME), tvar("y", AST.INTNAME) }),
        type(AST.INTNAME), null));
    declare(Axiom.create(Id.create(name),
        Forall.create(tvar("x", AST.INTNAME),
            Forall.create(tvar("y", AST.INTNAME),
                apply(AST.EQUALNAME,
                    apply(name, var("x"), var("y")),
                    If.create(apply(AST.LESSEQUALNAME, var("y"), apply("0")),
                        apply(base),
                        apply(combine, value.apply(var("x")),
                            apply(name,
                                var("x"),
                                apply(AST.MINUSNAME, var("y"), apply("1"))))))))));
  }
  
  /****************************************************************************
   * Create integer function iterating over an integer interval.
   * @param name the name of the function.
   * @param base the value of the function if the interval is empty.
   * @param combine the combination function.
   * @param value the function to apply to the element value.
   ***************************************************************************/
  private void intIntervalFun(String name, String base, String combine,
    java.util.function.Function<Exp,Exp> value)
  {
    declare(Function.create(Id.create(name),
        Arrays.asList(new TypedVar[] { tvar("x", AST.INTNAME), tvar("y", AST.INTNAME) }),
        type(AST.INTNAME), null));
    declare(Axiom.create(Id.create(name),
        Forall.create(tvar("x", AST.INTNAME),
            Forall.create(tvar("y", AST.INTNAME),
                apply(AST.EQUALNAME,
                    apply(name, var("x"), var("y")),
                    If.create(apply(AST.GREATERNAME, var("x"), var("y")),
                        apply(base),
                        apply(combine, value.apply(var("x")),
                            apply(name,
                                apply(AST.PLUSNAME, var("x"), apply("1")),
                                var("y")))))))));
  }
  
  /****************************************************************************
   * Create integer function iterating over an integer set.
   * @param name the name of the function.
   * @param base the value of the function if the set is empty.
   * @param combine the combination function.
   * @param value the function to apply to the element value.
   ***************************************************************************/
  private void intSetFun(String name, String base, String combine, 
    java.util.function.Function<Exp,Exp> value)
  {
    declare(Function.create(Id.create(name),
        Arrays.asList(new TypedVar[] { 
            TypedVar.create(Id.create("set"), setType(type(AST.INTNAME))) }),
        type(AST.INTNAME), null));
    declare(Axiom.create(Id.create(name + "_set"),
        Forall.create(TypedVar.create(Id.create("set"), setType(type(AST.INTNAME))),
            And.create(
                Imp.create(
                    Forall.create(tvar("elem", AST.INTNAME),
                        Not.create(apply(AST.MAPSELECTNAME, var("set"), var("elem")))),
                    apply(AST.EQUALNAME, apply(name, var("set")), apply(base))),
                Forall.create(tvar("elem", AST.INTNAME),
                    Imp.create(apply(AST.MAPSELECTNAME, var("set"), var("elem")),
                        apply(AST.EQUALNAME,
                            apply(name, var("set")),  
                            apply(combine, value.apply(var("elem")),
                                apply(name,
                                    apply(AST.MAPSTORENAME, var("set"), var("elem"), 
                                        False.create()))))))))));
  }
  
  /****************************************************************************
   * Check whether there exists a function with given name and argument types.
   * @param name the name.
   * @param types the types of the arguments.
   * @return true if and only if such a function exists.
   ***************************************************************************/
  private boolean existsFun(String name, List<Type> types)
  {
    List<TypeSymbol> tsymbols = new ArrayList<TypeSymbol>(types.size());
    for (Type type : types) tsymbols.add(checker.check(type));
    FunctionSymbol fsymbol = checker.getFunction(Id.create(name), tsymbols); 
    return fsymbol != null;  
  }
  
  /****************************************************************************
   * Create set size function for denoted element type.
   * @param type the element type.
   ***************************************************************************/
  private void createSizeFun(Type etype)
  {
    if (existsFun(SIZENAME, Arrays.asList(new Type[] { setType(etype) }))) return;
    TypeSymbol tsymbol = checker.check(etype);
    createSetFun(SIZENAME, tsymbol.root.type, "0", AST.PLUSNAME, (Exp e)->apply("1"));
  }
  
  /****************************************************************************
   * Create generic function iterating over a set.
   * @param name the name of the function.
   * @param etype the element type.
   * @param base the value of the function if the set is empty.
   * @param combine the combination function.
   * @param value the function to apply to the element value.
   ***************************************************************************/
  private int setFunCounter = -1;
  private void createSetFun(String name, Type etype,
    String base, String combine, java.util.function.Function<Exp,Exp> value)
  {
    declare(Function.create(Id.create(name),
        Arrays.asList(new TypedVar[] { 
            TypedVar.create(Id.create("set"), setType(etype)) }),
        type(AST.INTNAME), null));
    setFunCounter++;
    declare(Axiom.create(Id.create(name + setFunCounter),
        Forall.create(TypedVar.create(Id.create("set"), setType(etype)),
            And.create(
                Imp.create(
                    Forall.create(TypedVar.create(Id.create("elem"), etype),
                        Not.create(apply(AST.MAPSELECTNAME, var("set"), var("elem")))),
                    apply(AST.EQUALNAME, apply(name, var("set")), apply(base))),
                Forall.create(TypedVar.create(Id.create("elem"), etype),
                    Imp.create(apply(AST.MAPSELECTNAME, var("set"), var("elem")),
                        apply(AST.EQUALNAME,
                            apply(name, var("set")),  
                            apply(combine, value.apply(var("x")),
                                apply(name,
                                    apply(AST.MAPSTORENAME, var("set"), var("elem"), 
                                        False.create()))))))))));
  }
  
  // --------------------------------------------------------------------------
  //
  // syntax auxiliaries
  //
  // --------------------------------------------------------------------------
  
  /****************************************************************************
   * Create type from n-ary type constructor.
   * @param name the name of the constructor.
   * @param types the list of argument types.
   * @return the type.
   ***************************************************************************/
  private Type type(String name, List<Type> types)
  {
    return Type.create(Id.create(name), types);
  }
  
  /****************************************************************************
   * Create named type.
   * @param name the name of the constructor.
   * @param types the argument types.
   * @return the type.
   ***************************************************************************/
  private Type type(String name, Type... types)
  {
    return type(name, Arrays.asList(types));
  }
  
  /****************************************************************************
   * Create a set type.
   * @param etype the element type.
   * @return the set type.
   ***************************************************************************/
  private Type setType(Type etype)
  {
    return type(AST.MAPNAME, etype, type(AST.BOOLNAME));
    // if "finite" is given an actual definition,
    // the following prevents the inconsistency that arises
    // when "sum" is applied to an infinite set
    // with this, also "elemType" has to be redefined as shown below
    /*
    TypeSymbol esymbol = checker.check(etype);
    TypeSymbol ssymbol = setMap.get(esymbol);
    if (ssymbol != null) return ssymbol.type;
    Type stype = type(AST.MAPNAME, etype, type(AST.BOOLNAME));
    List<TypedVar> tvars = new ArrayList<TypedVar>(1);
    tvars.add(TypedVar.create(Id.create("set"), stype));
    String key = AST.internal(Integer.toString(setMap.size()));
    String pred = "finite" + key;
    declare(Decl.Function.create(Id.create(pred), tvars, type(AST.BOOLNAME), null));
    Id id = Id.create("settype" + key);
    List<Exp> exps = new ArrayList<Exp>(1);
    exps.add(var(AST.VALUENAME));
    declare(Decl.TypeDecl.create(id, stype, apply(pred, exps)));
    ssymbol = (TypeSymbol)id.getSymbol();
    setMap.put(esymbol, ssymbol);
    return ssymbol.type;
    */
  }
  
  /****************************************************************************
   * Get the element type.
   * @param type the set type.
   * @return the element type.
   ***************************************************************************/
  private Type elemType(Type type)
  {
    return type.types[0];
    // to be applied when "setType" is redefined as shown above
    /*
    TypeSymbol tsymbol = (TypeSymbol)type.id.getSymbol();
    Type type0 = tsymbol.base.type;
    return type0.types[0];
    */
  }
  
  /****************************************************************************
   * Create typed variable.
   * @param vname the name of the variable.
   * @param tname the name of the type.
   * @return the typed variable.
   ***************************************************************************/
  private TypedVar tvar(String vname, String tname)
  {
    return TypedVar.create(Id.create(vname), type(tname));
  }
  
  /****************************************************************************
   * Create variable expression.
   * @param name the name of the variable.
   * @return the expression.
   ***************************************************************************/
  private Exp var(String name)
  {
    return Var.create(Id.create(name));
  }

  /****************************************************************************
   * Create application of n-ary function.
   * @param name the name of the function.
   * @param exps the list of arguments.
   * @return the expression.
   ***************************************************************************/
  private Exp apply(String name, List<Exp> exps)
  {
    return Apply.create(Id.create(name), exps);
  }
  
  /****************************************************************************
   * Create application of function.
   * @param name the name of the function.
   * @param exps the arguments.
   * @return the expression.
   ***************************************************************************/
  private Exp apply(String name, Exp... exps)
  {
    return apply(name, Arrays.asList(exps));
  }
  
  /****************************************************************************
   * Create the empty set.
   * @param type the element type.
   * @return the set.
   ***************************************************************************/
  private Exp emptySet(Type type)
  {
    String tname1 = getTypeName(type);
    String tname2 = AST.BOOLNAME;
    Id id = Id.create(new String[] { AST.MAPCONSTRUCTNAME, tname1, tname2 });
    List<Exp> args = new ArrayList<Exp>(1);
    args.add(False.create());
    return Apply.create(id, args);
  }
  
  // --------------------------------------------------------------------------
  //
  // declaration auxiliaries
  //
  // --------------------------------------------------------------------------
  
  // unique suffix for generation of unused ids
  private static int idNumber = 0;
  
  /****************************************************************************
   * Return a new id not in use by RISCAL.
   * @param base the base string for the id.
   * @return the id.
   ***************************************************************************/
  private Id unusedId(String base)
  {
    String name = "__" + base + idNumber;
    idNumber++;
    return Id.create(name);
  }

  
  /****************************************************************************
   * Check whether identifier symbol was already declared.
   * @param id an identifier with a symbol.
   * @return true if the symbol was already declared; otherwise
   *         the symbol is declared now.
   ***************************************************************************/
  private boolean declared(riscal.syntax.AST.Identifier id)
  {
    riscal.types.Environment.Symbol symbol = id.getSymbol();
    if (symbols.contains(symbol)) return true;
    symbols.add(symbol);
    return false;
  }
    
  /****************************************************************************
   * Determine whether function is recursive.
   * @param id the function identifier.
   * @return true if the function is (directly or indirectly) recursive
   ***************************************************************************/
  private boolean isRecursive(riscal.syntax.AST.Identifier id)
  {
    riscal.types.Environment.Symbol.FunctionSymbol symbol = 
        (riscal.types.Environment.Symbol.FunctionSymbol)id.getSymbol();
    return callers.calls(symbol, symbol);
  }
  
  /****************************************************************************
   * Does true specification contain the "modular" keyword.
   * @param specs the function specification.
   * @return true if the "modular" keyword occurs.
   ***************************************************************************/
  private boolean isModular(riscal.syntax.AST.FunctionSpec[] specs)
  {
    for (riscal.syntax.AST.FunctionSpec spec : specs)
    {
      if (spec instanceof riscal.syntax.AST.FunctionSpec.ContractSpec)
        return true;
    }
    return false;
  }
  
  /****************************************************************************
   * Get precondition from function specifications.
   * @param specs the function specification.
   * @return the precondition.
   ***************************************************************************/
  private Exp getPrecondition(riscal.syntax.AST.FunctionSpec[] specs)
  {
    Exp result = null;
    for (riscal.syntax.AST.FunctionSpec spec : specs)
    {
      if (!(spec instanceof riscal.syntax.AST.FunctionSpec.RequiresSpec)) continue;
      riscal.syntax.AST.FunctionSpec.RequiresSpec spec0 =
          (riscal.syntax.AST.FunctionSpec.RequiresSpec)spec;
      result = Exp.and(result, translate(spec0.exp));
    }
    return result;
  }
  
  /****************************************************************************
   * Create universally quantified application.
   * @param id the name of a function predicate.
   * @param params the parameters.
   * @param transform a transformation function to be applied to the application.
   * @return the formula expression.
   ***************************************************************************/
  private Exp forallFormula(Id id, List<TypedVar> params,
    java.util.function.Function<Exp,Exp> transform)
  {
    int n = params.size();
    List<Exp> exps = new ArrayList<Exp>(n);
    for (int i = 0; i < n; i++)
      exps.add(Var.create(params.get(i).id));
    Exp exp = Apply.create(id, exps);
    exp = transform.apply(exp);
    for (int i = n-1; i >= 0; i--)
      exp = Forall.create(params.get(i), exp);
    return exp;
  }
  
  /****************************************************************************
   * Translate a function definition.
   * @param id the (original) function name.
   * @param id0 the (already translated) predicate name.
   * @param tvars the (already translated) predicate parameters.
   * @param type the result type (null if function denotes a predicate)
   * @param exp the (already translated) predicate body.
   * @param specs the function specification
   * @param declared was predicate already declared?
   * @param recursive is the predicate recursive?
   * @param modular is the contract to be used rather than the definition?
   ***************************************************************************/
  private void translateFunDef(riscal.syntax.AST.Identifier id,
    Id id0, List<TypedVar> tvars, Type type, Exp exp, 
    riscal.syntax.AST.FunctionSpec[] specs,
    boolean declared, boolean recursive, boolean modular)
  {
    if (!declared)
    {
      if (type == null) type = type(AST.BOOLNAME);
      if (!recursive && !modular)
      {
        Decl decl0 = Function.create(id0, tvars, type, exp);
        declare(decl0);
        return;
      }
      Decl decl0 = Function.create(id0, tvars, type, null);
      declare(decl0);
    }
    if (modular)
    {
      translateFunSpec(id, id0, tvars, specs);
      return;
    }
    Type type0 = type;
    Exp exp1 = forallFormula(id0, tvars, (Exp exp0)->
    {
      if (type0 == null) return Equiv.create(exp0, exp);
      List<Exp> exps = new ArrayList<Exp>(2);
      exps.add(exp0);
      exps.add(exp);
      return apply(AST.EQUALNAME, exps);
    });
    Id id1 = Id.create(new String[] { id0.toString(), "def" });
    Decl decl1 = Axiom.create(id1, exp1);
    declare(decl1);
  }
  
  /****************************************************************************
   * translate function specification.
   * @param id the (original) function name.
   * @param id0 the (translated) function name.
   * @param tvars the (translated) function parameters.
   * @param specs the function specification.
   ***************************************************************************/
  private void translateFunSpec(riscal.syntax.AST.Identifier id,
    Id id0, List<TypedVar> tvars, 
    riscal.syntax.AST.FunctionSpec[] specs)
  {
    riscal.types.Environment.Symbol.FunctionSymbol fsymbol =
        (riscal.types.Environment.Symbol.FunctionSymbol)id.getSymbol();
    // collect preconditions
    Exp precondition = null;
    for (riscal.syntax.AST.FunctionSpec spec : specs)
    {
      if (!(spec instanceof riscal.syntax.AST.FunctionSpec.RequiresSpec)) continue;
      riscal.syntax.AST.FunctionSpec.RequiresSpec spec0 = 
          (riscal.syntax.AST.FunctionSpec.RequiresSpec)spec;
      precondition = Exp.and(precondition, translate(spec0.exp));
    }
    Exp precondition0 = precondition;
    // create axiom for every postcondition
    int counter = 0;
    for (riscal.syntax.AST.FunctionSpec spec : specs)
    {
      if (!(spec instanceof riscal.syntax.AST.FunctionSpec.EnsuresSpec)) continue;
      riscal.syntax.AST.FunctionSpec.EnsuresSpec spec0 = 
          (riscal.syntax.AST.FunctionSpec.EnsuresSpec)spec;
      String suffix = counter == 0 ? "" : "_" + Integer.toString(counter);
      counter++;
      Id id1 = Id.create(new String[] { Main.shortName(fsymbol), "axiom" + suffix });
      Exp exp1 = forallFormula(id0, tvars, (Exp exp0)->
      {
        List<LetBinder> binders = new ArrayList<LetBinder>(1);
        Id var = Id.create(riscal.types.TypeChecker.resultName);
        binders.add(LetBinder.create(var, exp0));
        return Exp.imp(precondition0, Let.create(binders, translate(spec0.exp)));
      });
      Decl decl1 = Axiom.create(id1, exp1);
      declare(decl1);
    }
  }
  
  // --------------------------------------------------------------------------
  //
  // declarations
  //
  // --------------------------------------------------------------------------
  
  private void translateCore(List<riscal.syntax.AST.Declaration> decls)
  {
    declareBuiltins();
    try
    {
      for (riscal.syntax.AST.Declaration decl : decls)
        translate(decl);
    }
    catch(Exception e)
    {
      e.printStackTrace();
      checker = null;
      TranslationError.raise(e.getMessage());
    }
  }
  
  private void translate(riscal.syntax.AST.Declaration decl)
  {
    // set new declaration context
    riscal.syntax.AST.Declaration oldDecl = currentDecl;
    currentDecl = decl;
    
    translateDecl(decl);
    
    // process post-commands, new list necessary for recursive processing
    // List<Runnable> ps = postDecl;
    // postDecl = new ArrayList<Runnable>();
    // for (Runnable p : ps) p.run();
    
    // restore declaration context
    currentDecl = oldDecl;
  }
  
  private void translateDecl(riscal.syntax.AST.Declaration decl)
  {
    if (decl instanceof riscal.syntax.AST.Declaration.ValueDeclaration)
      translate((riscal.syntax.AST.Declaration.ValueDeclaration)decl);
    else if (decl instanceof riscal.syntax.AST.Declaration.FunctionDeclaration)
      translate((riscal.syntax.AST.Declaration.FunctionDeclaration)decl);
    else if (decl instanceof riscal.syntax.AST.Declaration.PredicateDeclaration)
      translate((riscal.syntax.AST.Declaration.PredicateDeclaration)decl);
    else if (decl instanceof riscal.syntax.AST.Declaration.ProcedureDeclaration)
      translate((riscal.syntax.AST.Declaration.ProcedureDeclaration)decl);
    else if (decl instanceof riscal.syntax.AST.Declaration.TypeDefinition)
      translate((riscal.syntax.AST.Declaration.TypeDefinition)decl);
    else if (decl instanceof riscal.syntax.AST.Declaration.RecTypeDefinition)
      translate((riscal.syntax.AST.Declaration.RecTypeDefinition)decl);
    else if (decl instanceof riscal.syntax.AST.Declaration.EnumTypeDefinition)
      translate((riscal.syntax.AST.Declaration.EnumTypeDefinition)decl);
    else if (decl instanceof riscal.syntax.AST.Declaration.ValueDefinition)
      translate((riscal.syntax.AST.Declaration.ValueDefinition)decl);
    else if (decl instanceof riscal.syntax.AST.Declaration.PredicateValueDefinition)
      translate((riscal.syntax.AST.Declaration.PredicateValueDefinition)decl);
    else if (decl instanceof riscal.syntax.AST.Declaration.FunctionDefinition)
      translate((riscal.syntax.AST.Declaration.FunctionDefinition)decl);
    else if (decl instanceof riscal.syntax.AST.Declaration.PredicateDefinition)
      translate((riscal.syntax.AST.Declaration.PredicateDefinition)decl);
    else if (decl instanceof riscal.syntax.AST.Declaration.TheoremDefinition)
      translate((riscal.syntax.AST.Declaration.TheoremDefinition)decl);
    else if (decl instanceof riscal.syntax.AST.Declaration.TheoremParamDefinition)
      translate((riscal.syntax.AST.Declaration.TheoremParamDefinition)decl);
    else if (decl instanceof riscal.syntax.AST.Declaration.ProcedureDefinition)
      translate((riscal.syntax.AST.Declaration.ProcedureDefinition)decl);
    else if (decl instanceof riscal.syntax.AST.Declaration.SharedSystem)
      translate((riscal.syntax.AST.Declaration.SharedSystem)decl);
    else if (decl instanceof riscal.syntax.AST.Declaration.DistributedSystem)
      translate((riscal.syntax.AST.Declaration.DistributedSystem)decl);
  }
  
  private void translate(riscal.syntax.AST.Declaration.ValueDeclaration decl)
  {
    Decl decl0 = Function.create(translate(decl.ident), 
        new ArrayList<TypedVar>(0), type(AST.NATNAME), null);
    declare(decl0);
  }
  
  private void translate(riscal.syntax.AST.Declaration.FunctionDeclaration decl)
  {
    declared(decl.ident);
    Decl decl0 = Function.create(translate(decl.ident), 
        translate(decl.param), translate(decl.type), null);
    declare(decl0);
  }
  
  private void translate(riscal.syntax.AST.Declaration.PredicateDeclaration decl)
  {
    declared(decl.ident);
    Decl decl0 = Function.create(translate(decl.ident), 
        translate(decl.param), type(AST.BOOLNAME), null);
    declare(decl0);
  }
  
  private void translate(riscal.syntax.AST.Declaration.ProcedureDeclaration decl)
  {
    declared(decl.ident);
    Decl decl0 = Function.create(translate(decl.ident), 
        translate(decl.param), translate(decl.type), null);
    declare(decl0);
  }
  
  private void translate(riscal.syntax.AST.Declaration.TypeDefinition decl)
  {
    Decl decl0 = TypeDecl.create(translate(decl.ident), 
        translate(decl.type), decl.exp == null ? null : translate(decl.exp));
    declare(decl0);
  }
  
  private void translate(riscal.syntax.AST.Declaration.RecTypeDefinition decl)
  {
    Exp exp = translate(decl.exp);
    List<DataTypeItem> items = translate(exp, decl.ritems);
    DataType decl0 = DataType.create(items);
    declare(decl0);
    axiomatizeHeight(items, decl.ritems);
  }
  
  private List<DataTypeItem> translate(Exp exp,
    riscal.syntax.AST.RecursiveTypeItem[] items)
  {
    List<DataTypeItem> items0 = new ArrayList<DataTypeItem>(items.length);
    for (riscal.syntax.AST.RecursiveTypeItem item : items)
      items0.add(translate(exp, item));
    return items0;
  }
  
  // number of selectors in current datatype
  private int selectors = 0;
  
  private DataTypeItem translate(Exp exp, riscal.syntax.AST.RecursiveTypeItem item)
  {
    Id id = translate(item.ident);
    List<DataTypeConstr> constrs = new ArrayList<DataTypeConstr>(item.ridents.length);
    selectors = 0;
    for (riscal.syntax.AST.RecursiveIdentifier rident : item.ridents)
      constrs.add(translate(id, rident));
    List<Exp> exps0 = new ArrayList<Exp>(1);
    exps0.add(Var.create(Id.create(AST.VALUENAME)));
    Exp exp0 = apply(AST.HEIGHTNAME, exps0);
    List<Exp> exps1 = new ArrayList<Exp>(2);
    exps1.add(exp0);
    exps1.add(exp);
    Exp exp1 = apply(AST.LESSEQUALNAME, exps1);
    return DataTypeItem.create(id, constrs, exp1);
  }
  
  private DataTypeConstr translate(Id id, riscal.syntax.AST.RecursiveIdentifier rident)
  {
    if (rident instanceof riscal.syntax.AST.RecursiveIdentifier.RecIdentifier)
      return translate(id, (riscal.syntax.AST.RecursiveIdentifier.RecIdentifier)rident);
    else if (rident instanceof riscal.syntax.AST.RecursiveIdentifier.RecApplication)
      return translate(id, (riscal.syntax.AST.RecursiveIdentifier.RecApplication)rident);
    TranslationError.raise("unknown recursive identifier: " + rident);
    return null;
  }
  
  private DataTypeConstr translate(Id id, riscal.syntax.AST.RecursiveIdentifier.RecIdentifier recident)
  {
    Id id0 = translate(recident.ident);
    String name = dtname(id, id0);
    return DataTypeConstr.create(Id.create(name), new ArrayList<TypedVar>(0));
  }
  
  private DataTypeConstr translate(Id id, riscal.syntax.AST.RecursiveIdentifier.RecApplication recident)
  {
    Id id0 = translate(recident.ident);
    List<TypedVar> tvars = new ArrayList<TypedVar>(recident.types.length);
    for (riscal.syntax.AST.Type type : recident.types)
    {
      Id id1 = Id.create("sel_" + Integer.toString(selectors));
      String sname = dtname(id, id1);
      selectors++; // global variable, reset to zero for every datatype
      tvars.add(TypedVar.create(Id.create(sname), translate(type)));
    }
    String cname = dtname(id, id0);
    return DataTypeConstr.create(Id.create(cname), tvars);
  }
  
  private void translate(riscal.syntax.AST.Declaration.EnumTypeDefinition decl)
  {
    translate(new riscal.syntax.AST.Declaration.RecTypeDefinition(
        new riscal.syntax.AST.Expression.NumberLiteralExp(new riscal.syntax.AST.Decimal("0")),
        Arrays.asList(new riscal.syntax.AST.RecursiveTypeItem[] { decl.ritem })));
  }
  
  private void axiomatizeHeight(List<DataTypeItem> items, 
    riscal.syntax.AST.RecursiveTypeItem[] items0)
  {
    int n = items.size();
    for (int i = 0; i < n; i++)
      axiomatizeHeight(items.get(i), items0[i]);
  }
  
  private void axiomatizeHeight(DataTypeItem item, 
    riscal.syntax.AST.RecursiveTypeItem item0)
  {
    int n = item.constrs.length;
    List<MatchBinder> binders = new ArrayList<MatchBinder>(n);
    for (int i = 0; i < n; i++)
    {
      ConstrPattern pattern = pattern(item.constrs[i]);
      Exp exp = height(pattern, item0.ridents[i]);
      binders.add(MatchBinder.create(pattern, exp));
    }
    Id axiomId = Id.create(new String[]{ "__" + AST.HEIGHTNAME });
    Id var = Id.create("x");
    Id tid = item.id;
    Exp apply = apply(AST.HEIGHTNAME, Var.create(var));
    Match match = Match.create(Var.create(var), binders);
    Exp axiom = Forall.create(TypedVar.create(var, Type.create(tid)),
        Apply.create(Id.create(AST.EQUALNAME), 
            Arrays.asList(new Exp[] { apply, match })));
    Decl decl = Axiom.create(axiomId, axiom);
    declare(decl); 
    FormulaSymbol symbol = (FormulaSymbol)axiomId.getSymbol();
    ProofProblem result = checker.getResult();
    result.heightAxioms.add(symbol);
  }
  
  private ConstrPattern pattern(DataTypeConstr constr)
  {
    List<TypedVar> tvars = new ArrayList<TypedVar>(constr.tvars.length);
    for (TypedVar tvar : constr.tvars)
    {
      FunctionSymbol fsymbol = (FunctionSymbol)tvar.id.getSymbol();
      Id id = Id.create(tvar.id.getStrings());
      VariableSymbol vsymbol = new VariableSymbol(id, fsymbol.tsymbol);
      id.setSymbol(vsymbol);
      tvars.add(TypedVar.create(id, tvar.type));
    }
    return ConstrPattern.create(constr.id, tvars);
  }
    
  private Exp height(ConstrPattern pattern, 
    riscal.syntax.AST.RecursiveTypeItem.RecursiveIdentifier rident)
  {
    int n = pattern.tvars.length;
    if (n == 0) return apply("0");
    riscal.syntax.AST.RecursiveTypeItem.RecursiveIdentifier.RecApplication rident0 =
        (riscal.syntax.AST.RecursiveTypeItem.RecursiveIdentifier.RecApplication) rident;
    riscal.types.Environment.Symbol.FunctionSymbol symbol = 
        (riscal.types.Environment.Symbol.FunctionSymbol)rident0.ident.getSymbol();
    List<Exp> exps = new ArrayList<Exp>(n);
    for (int i = 0; i < n; i++)
    {
      exps.add(height(Var.create(pattern.tvars[i].id), symbol.types[i]));
    }
    return apply(AST.PLUSNAME, apply("1"), applyMax(exps));
  }
  
  /***************************************************************************
   * Create RISCTP term for height of expression denoting a value of 
   * a (canonical) RISCAL type.
   * @param exp the RISCTP expression
   * @param type the canonical type of the RISCAL value.
   * @return the height expression
   **************************************************************************/
  Exp height(Exp exp, riscal.syntax.AST.Type type)
  {
    if (type instanceof riscal.syntax.AST.Type.UnitType)
      return apply("0");
    if (type instanceof riscal.syntax.AST.Type.BoolType)
      return apply("0");
    if (type instanceof riscal.syntax.AST.Type.IntType)
      return apply("0");
    if (type instanceof riscal.syntax.AST.Type.SubType)
    {
      riscal.syntax.AST.Type.SubType type0 = (riscal.syntax.AST.Type.SubType)type;
      return height(exp, type0.type);
    }
    if (type instanceof riscal.syntax.AST.Type.RecursiveType)
      return apply(AST.HEIGHTNAME, exp);
    if (type instanceof riscal.syntax.AST.Type.TupleType)
    {
      riscal.syntax.AST.Type.TupleType type0 = (riscal.syntax.AST.Type.TupleType)type;
      return tupleHeight(exp, type0.types);
    }
    if (type instanceof riscal.syntax.AST.Type.RecordType)
    {
      riscal.syntax.AST.Type.RecordType type0 = (riscal.syntax.AST.Type.RecordType)type;
      int n = type0.param.length;
      riscal.syntax.AST.Type[] types = new riscal.syntax.AST.Type[n];
      for (int i = 0; i < n; i++)
      {
        riscal.types.Environment.Symbol.ValueSymbol symbol =
            (riscal.types.Environment.Symbol.ValueSymbol)type0.param[i].ident.getSymbol();
        types[i] = symbol.type;
      }
      return tupleHeight(exp, types);
    }
    if (type instanceof riscal.syntax.AST.Type.MapType)
      return mapHeight(exp, (riscal.syntax.AST.Type.MapType)type);
    if (type instanceof riscal.syntax.AST.Type.SetType)
      return setHeight(exp, (riscal.syntax.AST.Type.SetType)type);
    TranslationError.raise("unknown canonical type: " + type);
    return null;
  }
    
  private Exp tupleHeight(Exp exp, riscal.syntax.AST.Type.Type[] types)
  {
    int n = types.length;
    List<Exp> exps = new ArrayList<Exp>(n);
    for (int i = 0; i < n; i++)
    {
      Id id = Id.create(new String[] { AST.TUPLESELECTNAME, Integer.toString(i+1) });
      exps.add(height(Apply.create(id, Arrays.asList(new Exp[] { exp })), types[i]));
    }
    return applyMax(exps);
  }
  
  private Exp mapHeight(Exp exp, riscal.syntax.AST.Type.MapType type)
  {
    Quantification quant = new Quantification();
    Id kid = unusedId("key");
    Type ktype = translate(type.type1);
    TypedVar tvar = TypedVar.create(kid, ktype);
    quant.ids.add(kid);
    quant.types.add(ktype);
    quant.tvars.add(tvar);
    Exp exp0 = apply(AST.MAPSELECTNAME, exp, Var.create(kid));
    return createUniqueChoice(AST.GREATEREQUALNAME, quant, 
        height(exp0, type.type2), type(AST.INTNAME));
  }
  
  private Exp setHeight(Exp exp, riscal.syntax.AST.Type.SetType type)
  {
    Quantification quant = new Quantification();
    Id eid = unusedId("elem");
    Type etype = translate(type.type);
    TypedVar tvar = TypedVar.create(eid, etype);
    quant.ids.add(eid);
    quant.types.add(etype);
    quant.tvars.add(tvar);
    quant.exp = apply(AST.MAPSELECTNAME, exp, Var.create(eid));
    return createUniqueChoice(AST.GREATEREQUALNAME, quant, 
        height(Var.create(eid), type.type), type(AST.INTNAME));
  }
  
  private void translate(riscal.syntax.AST.Declaration.ValueDefinition decl)
  {
    riscal.syntax.AST.Type type = decl.type == null ? decl.exp.getType() : decl.type;
    Decl decl0 = Function.create(translate(decl.ident), 
        new ArrayList<TypedVar>(0), translate(type), translate(decl.exp));
    declare(decl0);
  }
  
  private void translate(riscal.syntax.AST.Declaration.PredicateValueDefinition decl)
  {
    Decl decl0 = Function.create(translate(decl.ident), 
        new ArrayList<TypedVar>(0), type(AST.BOOLNAME), translate(decl.exp));
    declare(decl0);
  }
  
  private void translate(riscal.syntax.AST.Declaration.FunctionDefinition decl)
  {
    boolean declared = declared(decl.ident);
    boolean recursive = isRecursive(decl.ident);
    boolean modular = isModular(decl.spec);
    Id id = translate(decl.ident);
    List<TypedVar> tvars = translate(decl.param);
    Type type = translate(decl.type);
    Exp exp = translate(decl.exp);
    translateFunDef(decl.ident, id, tvars, type, exp, decl.spec, declared, recursive, modular);
  }
  
  private void translate(riscal.syntax.AST.Declaration.PredicateDefinition decl)
  {
    boolean declared = declared(decl.ident);
    boolean recursive = isRecursive(decl.ident);
    boolean modular = isModular(decl.spec);
    Id id = translate(decl.ident);
    List<TypedVar> tvars = translate(decl.param);
    Exp exp = translate(decl.exp);
    translateFunDef(decl.ident, id, tvars, null, exp, decl.spec, declared, recursive, modular);
  }
  
  private void translate(riscal.syntax.AST.Declaration.TheoremDefinition decl)
  {
    // in RISCAL, a theorem can be also used as a predicate
    Id id0 = translate(decl.ident);
    Decl decl0 = Function.create(id0, 
        new ArrayList<TypedVar>(0), type(AST.BOOLNAME), translate(decl.exp));
    declare(decl0);
    // the actual theorem or axiom
    Exp formula = Apply.create(id0, new ArrayList<Exp>(0));
    Id id1 = Id.create(decl.ident.toString());
    Decl decl1 = decl.axiom ? Axiom.create(id1, formula) : Theorem.create(id1, formula);
    declare(decl1);
  }
  
  private void translate(riscal.syntax.AST.Declaration.TheoremParamDefinition decl)
  {
    // in RISCAL, a theorem can be also used as a predicate
    boolean declared = declared(decl.ident);
    boolean recursive = isRecursive(decl.ident);
    boolean modular = isModular(decl.spec);
    Id id0 = translate(decl.ident);
    List<TypedVar> tvars = translate(decl.param);
    Exp exp = translate(decl.exp);
    translateFunDef(decl.ident, id0, tvars, null, exp, decl.spec, declared, recursive, modular);
    // the actual theorem or axiom
    Exp pre = getPrecondition(decl.spec);
    Exp formula = forallFormula(id0, translate(decl.param), (Exp exp0)->Exp.imp(pre, exp0));
    // theorem names may be overloaded
    riscal.types.Environment.Symbol.FunctionSymbol fsymbol =
        (riscal.types.Environment.Symbol.FunctionSymbol)decl.ident.getSymbol();
    Id id2 = Id.create(new String[] { riscal.Main.shortName(fsymbol) });
    Decl decl2 = decl.axiom ? Axiom.create(id2, formula) : Theorem.create(id2, formula);
    declare(decl2);
  }
  
  private void translate(riscal.syntax.AST.Declaration.ProcedureDefinition decl)
  {
    Id id0 = translate(decl.ident);
    List<TypedVar> tvars = translate(decl.param);
    Decl decl0 = Function.create(id0, tvars, translate(decl.type), null);
    declare(decl0);
    translateFunSpec(decl.ident, id0, tvars, decl.spec);
  }
    
  private void translate(riscal.syntax.AST.Declaration.SharedSystem decl)
  {
    // ignore
  }
  
  private void translate(riscal.syntax.AST.Declaration.DistributedSystem decl)
  {
    // ignore
  }
  
  // --------------------------------------------------------------------------
  //
  // type auxiliaries
  //
  // --------------------------------------------------------------------------
  
  /****************************************************************************
   * Get integer type for bounds.
   * @param exp1 the lower bound.
   * @param exp2 the upper bound.
   * @return the correspondingly bounded integer type.
   ***************************************************************************/
  private Type intType(Exp exp1, Exp exp2)
  {
    String key1 = exp1.toString();
    String key2 = exp2.toString();
    Map<String,Type> map = intMap.get(key1);
    if (map == null) 
    {
      map = new HashMap<String,Type>();
      intMap.put(key1, map);
    }
    Type type = map.get(key2);
    if (type != null) return type;
    String name = AST.INTNAME+ AST.internal(Integer.toString(intMapSize));
    intMapSize++;
    Decl decl =  TypeDecl.create(Id.create(name), type(AST.INTNAME),
        And.create(
            apply(AST.LESSEQUALNAME, exp1, Var.create(Id.create(AST.VALUENAME))),
            apply(AST.LESSEQUALNAME, Var.create(Id.create(AST.VALUENAME)), exp2)));
    declare(decl);
    type = type(name);
    map.put(key2, type);
    return type;
  }
  
  // --------------------------------------------------------------------------
  //
  // types
  //
  // --------------------------------------------------------------------------
  
  private Type translate(riscal.syntax.AST.Type type)
  {
    if (type instanceof riscal.syntax.AST.Type.IdentifierType)
      return translate((riscal.syntax.AST.Type.IdentifierType)type);
    if (type instanceof riscal.syntax.AST.Type.UnitType)
      return translate((riscal.syntax.AST.Type.UnitType)type);
    if (type instanceof riscal.syntax.AST.Type.BoolType)
      return translate((riscal.syntax.AST.Type.BoolType)type);
    if (type instanceof riscal.syntax.AST.Type.IntType)
      return translate((riscal.syntax.AST.Type.IntType)type);
    if (type instanceof riscal.syntax.AST.Type.NatType)
      return translate((riscal.syntax.AST.Type.NatType)type);
    if (type instanceof riscal.syntax.AST.Type.MapType)
      return translate((riscal.syntax.AST.Type.MapType)type);
    if (type instanceof riscal.syntax.AST.Type.TupleType)
      return translate((riscal.syntax.AST.Type.TupleType)type);
    if (type instanceof riscal.syntax.AST.Type.RecordType)
      return translate((riscal.syntax.AST.Type.RecordType)type);
    if (type instanceof riscal.syntax.AST.Type.SetType)
      return translate((riscal.syntax.AST.Type.SetType)type);
    if (type instanceof riscal.syntax.AST.Type.ArrayType)
      return translate((riscal.syntax.AST.Type.ArrayType)type);
    if (type instanceof riscal.syntax.AST.Type.SubType)
      return translate((riscal.syntax.AST.Type.SubType)type);
    if (type instanceof riscal.syntax.AST.Type.RecursiveType)
      return translate((riscal.syntax.AST.Type.RecursiveType)type);
    TranslationError.raise("unknown type: " + type);
    return null;
  }
  
  private Type translate(riscal.syntax.AST.Type.IdentifierType type)
  {
    return type(type.ident.string);
  }
  
  private Type translate(riscal.syntax.AST.Type.UnitType type)
  {
    return type(UNITNAME);
  }
  
  private Type translate(riscal.syntax.AST.Type.BoolType type)
  {
    return type(AST.BOOLNAME);
  }
  
  private Type translate(riscal.syntax.AST.Type.IntType type)
  {
    // try to extract bounds from syntactic type or from semantic type
    // if both fails, we have to do without bound information
    riscal.syntax.AST.Expression exp1 = type.exp1 != null ? type.exp1 : type.lowerBound();
    riscal.syntax.AST.Expression exp2 = type.exp2 != null ? type.exp2 : type.upperBound();
    if (exp1 == null || exp2 == null) return type(AST.INTNAME);
    return intType(translate(exp1), translate(exp2));
  }
  
  private Type translate(riscal.syntax.AST.Type.NatType type)
  {
    // ignore bounds in semantic type
    if (type.exp == null) return type(AST.NATNAME);
    Exp exp = apply("0");
    return intType(exp, translate(type.exp));
  }
  
  private Type translate(riscal.syntax.AST.Type.MapType type)
  {
    return type(AST.MAPNAME, translate(type.type1), translate(type.type2));
  }
  
  private Type translate(riscal.syntax.AST.Type.TupleType type)
  {
    List<Type> types = new ArrayList<Type>(type.types.length);
    for (riscal.syntax.AST.Type type0 : type.types)
      types.add(translate(type0));
    return type(AST.TUPLENAME, types);
  }
  
  private Type translate(riscal.syntax.AST.Type.RecordType type)
  {
    if (type.param != null)
    {
      List<Type> types = new ArrayList<Type>(type.param.length);
      for (riscal.syntax.AST.Parameter param : type.param)
        types.add(translate(param.type));
      return type(AST.TUPLENAME, types);
    }
    List<Type> types = new ArrayList<Type>(type.symbols.length);
    for (riscal.types.Environment.Symbol.ValueSymbol symbol : type.symbols)
      types.add(translate(symbol.type));
    return type(AST.TUPLENAME, types);
  }
  
  private Type translate(riscal.syntax.AST.Type.SetType type)
  {
    return setType(translate(type.type));
  }
  
  private Type translate(riscal.syntax.AST.Type.ArrayType type)
  {
    Exp exp1 = apply("0");
    Exp exp2 = apply(AST.MINUSNAME, translate(type.exp), apply("1"));
    return type(AST.MAPNAME, intType(exp1, exp2), translate(type.type));
  }
  
  private Type translate(riscal.syntax.AST.Type.SubType type)
  {
    // return just name of type
    return type(type.ident.string);
  }
  
  private Type translate(riscal.syntax.AST.Type.RecursiveType type)
  {
    // return just name of type
    return type(type.ritem.ident.string);
  }
  
  // --------------------------------------------------------------------------
  //
  // expression auxiliaries
  //
  // --------------------------------------------------------------------------
  
 
  /****************************************************************************
   * Get index of record selection.
   * @param exp an expression denoting a record.
   * @param ident an identifier denoting a component in the record.
   * @return the 1-based index of the component.
   **************************************************************************/
  private int getRecordIndex(riscal.syntax.AST.Expression exp,
    riscal.syntax.AST.Identifier ident)
  {
    riscal.syntax.AST.Type.RecordType type = 
        (riscal.syntax.AST.Type.RecordType)exp.getType();
    riscal.types.Environment.Symbol.ValueSymbol symbol = 
        (riscal.types.Environment.Symbol.ValueSymbol)ident.getSymbol();
    return 1+riscal.semantics.Translator.getIndex(type, symbol);
  }
  
  /***************************************************************************
   * Create binders for the components of a tuple.
   * @param id the name of the tuple.
   * @param ids the names of the bound variables.
   * @return the binders.
   **************************************************************************/
  private List<LetBinder> tupleBinders(Id id, List<Id> ids)
  {
    int n = ids.size();
    List<LetBinder> binders = new ArrayList<LetBinder>(n);
    int i = 1;
    for (Id id0 : ids)
    {
      Id id1 = Id.create(new String[] { AST.TUPLESELECTNAME, Integer.toString(i) });
      i++;
      Exp val = Apply.create(id1, Arrays.asList(new Exp[] { Var.create(id) }));
      binders.add(LetBinder.create(id0, val));
    }
    return binders;
  }
    
  /***************************************************************************
   * Bind names to components of a tuple in ane expression.
   * @param id the name of the tuple.
   * @param ids the names of the bound variables.
   * @param exp the expression base.
   * @return the bound expression.
   **************************************************************************/
  private Exp tupleBind(Id id, List<Id> ids, Exp exp)
  {
    List<LetBinder> binders = tupleBinders(id, ids);
    int n = binders.size();
    for (int i = n-1; i >= 0; i--)
      exp = Let.create(Arrays.asList(new LetBinder[] { binders.get(i) }), exp); 
    return exp;
  }
  
  // result of analyzing a quantified variable
  private static class Quantification
  {
    List<TypedVar> tvars = new ArrayList<TypedVar>();
    List<Id> ids = new ArrayList<Id>();
    List<Type> types = new ArrayList<Type>();
    Exp exp = null; // the conjunction of constraints (or null)
  }
 
  /****************************************************************************
   * Analyze a quantified variable.
   * @param qvar the quantified variable.
   * @return the analysis result.
   ***************************************************************************/
  private Quantification quantification(riscal.syntax.AST.QuantifiedVariable qvar)
  {
    Quantification quant = new Quantification();
    translate(qvar, quant,
        (Quantification q, TypedVar tvar)->{ 
          q.tvars.add(tvar); 
          q.ids.add(tvar.id);
          q.types.add(tvar.type);
          return q; 
        },
        (Quantification q, Exp exp0)->{ q.exp = Exp.and(exp0, q.exp); return q; });
    return quant;
  }
    
  /****************************************************************************
   * Create a choose expression from the quantification analysis.
   * @param sat is the choice formula universally satisfiable?
   * @param quant the result of the quantification analysis.
   * @return the expression.
   ***************************************************************************/
  private Choose createChoose(boolean sat, Quantification quant)
  {
    Exp exp0 = quant.exp == null ? True.create() : quant.exp;
    int n = quant.tvars.size();
    if (n == 1) return Choose.create(sat, quant.tvars.get(0), exp0);
    List<Id> ids = new ArrayList<Id>(n);
    List<Type> types = new ArrayList<Type>(n);
    for (int i = n-1; i >= 0; i--) 
    {
      TypedVar tvar = quant.tvars.get(i);
      ids.add(Id.create(tvar.id));
      types.add(tvar.type);
    }
    Id id = unusedId("choose");
    TypedVar tvar = TypedVar.create(id, type(AST.TUPLENAME, types)); 
    return Choose.create(sat, tvar, tupleBind(id, ids, exp0));
  }
 
  /****************************************************************************
   * Create a "choose in" expression from the quantification analysis.
   * @param quant the result of the quantification analysis.
   * @param exp the (translated) body of the expression.
   * @return the expression.
   ***************************************************************************/
  private Exp createChoose(Quantification quant, Exp exp)
  {
    Choose choose = createChoose(false, quant);
    int n = quant.tvars.size();
    Id id = n == 1 ? Id.create(choose.tvar.id) : unusedId("choose");
    LetBinder binder = LetBinder.create(id, choose);
    List<Id> ids = new ArrayList<Id>(n);
    for (int i = n-1; i >= 0; i--) 
    {
      TypedVar tvar = quant.tvars.get(i);
      ids.add(Id.create(tvar.id));
    }
    if (n > 1) exp = tupleBind(id, ids, exp);
    return Let.create(Arrays.asList(new LetBinder[] { binder }), exp);
  }
  
  /****************************************************************************
   * Create a formula that is true for every element of a set builder.
   * @param eid the identifier of the element
   * @param quant the result of the quantification analysis.
   * @param exp the (translated) body of the set builder expression.
   * @return the formula.
   ***************************************************************************/
  private Exp createElemFormula(Id eid, Quantification quant, Exp exp)
  {
    Exp formula = Exp.and(quant.exp, apply(AST.EQUALNAME, Var.create(eid), exp));
    int n = quant.tvars.size();
    for (int i = n-1; i >= 0; i--)
      formula = Exists.create(quant.tvars.get(i), formula);
    return formula;
  }
  
  /****************************************************************************
   * Create a set builder expression from quantification analysis.
   * @param etype the element type.
   * @param quant the result of the quantification analysis.
   * @param exp the (translated) body of the set builder expression.
   * @return the expression.
   ***************************************************************************/
  private Exp createSetBuilder(Type etype, Quantification quant, Exp exp)
  {
    Id sid = unusedId("set");
    Id eid = unusedId("elem");
    Type stype = setType(etype);
    Exp formula = createElemFormula(eid, quant, exp);
    return Choose.create(true, TypedVar.create(sid, stype),
        Forall.create(TypedVar.create(eid, etype),
            Equiv.create(
                apply(AST.MAPSELECTNAME, Var.create(sid), Var.create(eid)),
                formula)));
  }
  
  /****************************************************************************
   * Create expression for a unique choice from a quantified expression.
   * @param name the name of the predicate that makes the choice unique.
   * @param quant the result of the quantification analysis
   * @param exp the expression denoting the possible values.
   * @param etype the type of the values.
   * @return the expression.
   ***************************************************************************/
  private Exp createUniqueChoice(String pred, 
    Quantification quant, Exp exp, Type etype)
  {
    Id eid = unusedId("val0");
    Id eid0 = unusedId("val1");
    return Choose.create(false, TypedVar.create(eid, etype),
        And.create(
            createElemFormula(eid, quant, exp),
            Forall.create(TypedVar.create(eid0, etype),
                Imp.create(
                    createElemFormula(eid0, quant, exp),
                    apply(pred, Var.create(eid), Var.create(eid0))))));
  }
  
  /****************************************************************************
   * Get name for type.
   * @param type a type.
   * @return its name (generated by type definition, if type is not atomic)
   ***************************************************************************/
  private String getTypeName(Type type)
  {
    String name = type.id.toString();
    if (type.types.length == 0) return name;
    String key = type.toString();
    name = typeMap.get(key);
    if (name != null) return name;
    Id id = unusedId("type");
    Decl decl = TypeDecl.create(id, type, null);
    declare(decl);
    name = id.toString();
    typeMap.put(key, name);
    return name;
  }
  
  /****************************************************************************
   * Create a (generalized) powerset expression.
   * @param etype the element type (of the basic set).
   * @param exp the basic set.
   * @param cond null or a condition to be applied to the sets in the power set.
   * @return the powerset expression.
   ***************************************************************************/
  private Exp powerSet(Type etype, Exp exp, java.util.function.Function<Exp,Exp> cond)
  {
    Type stype = setType(etype);
    Id pid = Id.create("pset");
    Id sid0 = Id.create("set0");
    Id sid1 = Id.create("set1");
    Id eid = Id.create("elem");
    LetBinder binder = LetBinder.create(sid0, exp);
    return Let.create(Arrays.asList(new LetBinder[] { binder }), 
        Choose.create(true, TypedVar.create(pid, setType(stype)),
            Forall.create(TypedVar.create(Id.create(sid1), stype),
                Equiv.create(
                    apply(AST.MAPSELECTNAME, Var.create(pid), Var.create(sid1)),
                    Exp.and(cond == null ? null : cond.apply(Var.create(sid1)), 
                        Forall.create(TypedVar.create(eid, etype),
                            Imp.create(
                                apply(AST.MAPSELECTNAME, Var.create(sid1), Var.create(eid)),
                                apply(AST.MAPSELECTNAME, Var.create(sid0), Var.create(eid)))))))));
  }
  
  /****************************************************************************
   * Create map construction expression.
   * @param type1 the key type.
   * @param type2 the value type.
   * @param exp the value.
   * @return the expressions.
   ***************************************************************************/
  private Exp createMap(Type type1, Type type2, Exp exp)
  {
    String tname1 = getTypeName(type1);
    String tname2 = getTypeName(type2);
    Id id = Id.create(new String[] { AST.MAPCONSTRUCTNAME, tname1, tname2 });
    return Apply.create(id, Arrays.asList(new Exp[] { exp }));
  }
  
  /****************************************************************************
   * Get name of datatype function.
   * @param id1 the datatype identifier.
   * @param id2 the function identifier.
   * @return the name of the datatype function.
   ***************************************************************************/
  private String dtname(Id id1, Id id2)
  {
    return AST.unquote(id1.toString()) + "!" + AST.unquote(id2.toString());
  }
  
  /****************************************************************************
   * create maximum of expressions.
   * @param exps a list of expressions.
   * @return their maximum.
   ***************************************************************************/
  private Exp applyMax(List<Exp> exps)
  {
    int n = exps.size();
    if (n == 1) return exps.get(0);
    Exp exp0 = null;
    for (int i = n-1; i >= 1; i--)
    {
      Exp exp1 = exps.get(i);
      Exp exp2 = i == 1 ? exps.get(0) : var("_max" + (i-1));
      Exp exp3 = If.create(apply(AST.LESSNAME, exp1, exp2), exp1, exp2);
      if (i == n-1)
        exp0 = exp3;
      else
      {
        LetBinder binder = LetBinder.create(Id.create("_max" + i), exp3);
        exp0 = Let.create(Arrays.asList(new LetBinder[] { binder }), exp0);
      }
    }
    return exp0;
  }
  
  // --------------------------------------------------------------------------
  //
  // expressions
  //
  // --------------------------------------------------------------------------
  
  private Exp translate(riscal.syntax.AST.Expression exp)
  {
    if (exp instanceof riscal.syntax.AST.Expression.UnitExp)
      return translate((riscal.syntax.AST.Expression.UnitExp)exp);
    if (exp instanceof riscal.syntax.AST.Expression.TrueExp)
      return translate((riscal.syntax.AST.Expression.TrueExp)exp);
    if (exp instanceof riscal.syntax.AST.Expression.FalseExp)
      return translate((riscal.syntax.AST.Expression.FalseExp)exp);
    if (exp instanceof riscal.syntax.AST.Expression.NumberLiteralExp)
      return translate((riscal.syntax.AST.Expression.NumberLiteralExp)exp);
    if (exp instanceof riscal.syntax.AST.Expression.EmptySetExp)
      return translate((riscal.syntax.AST.Expression.EmptySetExp)exp);
    if (exp instanceof riscal.syntax.AST.Expression.RecIdentifierExp)
      return translate((riscal.syntax.AST.Expression.RecIdentifierExp)exp);
    if (exp instanceof riscal.syntax.AST.Expression.RecApplicationExp)
      return translate((riscal.syntax.AST.Expression.RecApplicationExp)exp);
    if (exp instanceof riscal.syntax.AST.Expression.IdentifierExp)
      return translate((riscal.syntax.AST.Expression.IdentifierExp)exp);
    if (exp instanceof riscal.syntax.AST.Expression.ApplicationExp)
      return translate((riscal.syntax.AST.Expression.ApplicationExp)exp);
    if (exp instanceof riscal.syntax.AST.Expression.PowersetExp)
      return translate((riscal.syntax.AST.Expression.PowersetExp)exp);
    if (exp instanceof riscal.syntax.AST.Expression.Powerset1Exp)
      return translate((riscal.syntax.AST.Expression.Powerset1Exp)exp);
    if (exp instanceof riscal.syntax.AST.Expression.Powerset2Exp)
      return translate((riscal.syntax.AST.Expression.Powerset2Exp)exp);
    if (exp instanceof riscal.syntax.AST.Expression.HeightExp)
      return translate((riscal.syntax.AST.Expression.HeightExp)exp);
    if (exp instanceof riscal.syntax.AST.Expression.MapSelectionExp)
      return translate((riscal.syntax.AST.Expression.MapSelectionExp)exp);
    if (exp instanceof riscal.syntax.AST.Expression.TupleSelectionExp)
      return translate((riscal.syntax.AST.Expression.TupleSelectionExp)exp);
    if (exp instanceof riscal.syntax.AST.Expression.RecordSelectionExp)
      return translate((riscal.syntax.AST.Expression.RecordSelectionExp)exp);
    if (exp instanceof riscal.syntax.AST.Expression.NegationExp)
      return translate((riscal.syntax.AST.Expression.NegationExp)exp);
    if (exp instanceof riscal.syntax.AST.Expression.FactorialExp)
      return translate((riscal.syntax.AST.Expression.FactorialExp)exp);
    if (exp instanceof riscal.syntax.AST.Expression.BigUnionExp)
      return translate((riscal.syntax.AST.Expression.BigUnionExp)exp);
    if (exp instanceof riscal.syntax.AST.Expression.BigIntersectExp)
      return translate((riscal.syntax.AST.Expression.BigIntersectExp)exp);
    if (exp instanceof riscal.syntax.AST.Expression.PowerExp)
      return translate((riscal.syntax.AST.Expression.PowerExp)exp);
    if (exp instanceof riscal.syntax.AST.Expression.TimesExp)
      return translate((riscal.syntax.AST.Expression.TimesExp)exp);
    if (exp instanceof riscal.syntax.AST.Expression.TimesExpMult)
      return translate((riscal.syntax.AST.Expression.TimesExpMult)exp);
    if (exp instanceof riscal.syntax.AST.Expression.DividesExp)
      return translate((riscal.syntax.AST.Expression.DividesExp)exp);
    if (exp instanceof riscal.syntax.AST.Expression.RemainderExp)
      return translate((riscal.syntax.AST.Expression.RemainderExp)exp);
    if (exp instanceof riscal.syntax.AST.Expression.PlusExp)
      return translate((riscal.syntax.AST.Expression.PlusExp)exp);
    if (exp instanceof riscal.syntax.AST.Expression.PlusExpMult)
      return translate((riscal.syntax.AST.Expression.PlusExpMult)exp);
    if (exp instanceof riscal.syntax.AST.Expression.MinusExp)
      return translate((riscal.syntax.AST.Expression.MinusExp)exp);
    if (exp instanceof riscal.syntax.AST.Expression.EnumeratedSetExp)
      return translate((riscal.syntax.AST.Expression.EnumeratedSetExp)exp);
    if (exp instanceof riscal.syntax.AST.Expression.CartesianExp)
      return translate((riscal.syntax.AST.Expression.CartesianExp)exp);
    if (exp instanceof riscal.syntax.AST.Expression.IntervalExp)
      return translate((riscal.syntax.AST.Expression.IntervalExp)exp);
    if (exp instanceof riscal.syntax.AST.Expression.TupleExp)
      return translate((riscal.syntax.AST.Expression.TupleExp)exp);
    if (exp instanceof riscal.syntax.AST.Expression.RecordExp)
      return translate((riscal.syntax.AST.Expression.RecordExp)exp);
    if (exp instanceof riscal.syntax.AST.Expression.SetSizeExp)
      return translate((riscal.syntax.AST.Expression.SetSizeExp)exp);
    if (exp instanceof riscal.syntax.AST.Expression.UnionExp)
      return translate((riscal.syntax.AST.Expression.UnionExp)exp);
    if (exp instanceof riscal.syntax.AST.Expression.IntersectExp)
      return translate((riscal.syntax.AST.Expression.IntersectExp)exp);
    if (exp instanceof riscal.syntax.AST.Expression.WithoutExp)
      return translate((riscal.syntax.AST.Expression.WithoutExp)exp);
    if (exp instanceof riscal.syntax.AST.Expression.NumberExp)
      return translate((riscal.syntax.AST.Expression.NumberExp)exp);
    if (exp instanceof riscal.syntax.AST.Expression.SetBuilderExp)
      return translate((riscal.syntax.AST.Expression.SetBuilderExp)exp);
    if (exp instanceof riscal.syntax.AST.Expression.SumExp)
      return translate((riscal.syntax.AST.Expression.SumExp)exp);
    if (exp instanceof riscal.syntax.AST.Expression.ProductExp)
      return translate((riscal.syntax.AST.Expression.ProductExp)exp);
    if (exp instanceof riscal.syntax.AST.Expression.MinExp)
      return translate((riscal.syntax.AST.Expression.MinExp)exp);
    if (exp instanceof riscal.syntax.AST.Expression.MaxExp)
      return translate((riscal.syntax.AST.Expression.MaxExp)exp);
    if (exp instanceof riscal.syntax.AST.Expression.ArrayBuilderExp)
      return translate((riscal.syntax.AST.Expression.ArrayBuilderExp)exp);
    if (exp instanceof riscal.syntax.AST.Expression.MapBuilderExp)
      return translate((riscal.syntax.AST.Expression.MapBuilderExp)exp);
    if (exp instanceof riscal.syntax.AST.Expression.ChooseExp)
      return translate((riscal.syntax.AST.Expression.ChooseExp)exp);
    if (exp instanceof riscal.syntax.AST.Expression.MapUpdateExp)
      return translate((riscal.syntax.AST.Expression.MapUpdateExp)exp);
    if (exp instanceof riscal.syntax.AST.Expression.TupleUpdateExp)
      return translate((riscal.syntax.AST.Expression.TupleUpdateExp)exp);
    if (exp instanceof riscal.syntax.AST.Expression.RecordUpdateExp)
      return translate((riscal.syntax.AST.Expression.RecordUpdateExp)exp);
    if (exp instanceof riscal.syntax.AST.Expression.EqualsExp)
      return translate((riscal.syntax.AST.Expression.EqualsExp)exp);
    if (exp instanceof riscal.syntax.AST.Expression.NotEqualsExp)
      return translate((riscal.syntax.AST.Expression.NotEqualsExp)exp);
    if (exp instanceof riscal.syntax.AST.Expression.LessExp)
      return translate((riscal.syntax.AST.Expression.LessExp)exp);
    if (exp instanceof riscal.syntax.AST.Expression.LessEqualExp)
      return translate((riscal.syntax.AST.Expression.LessEqualExp)exp);
    if (exp instanceof riscal.syntax.AST.Expression.GreaterExp)
      return translate((riscal.syntax.AST.Expression.GreaterExp)exp);
    if (exp instanceof riscal.syntax.AST.Expression.GreaterEqualExp)
      return translate((riscal.syntax.AST.Expression.GreaterEqualExp)exp);
    if (exp instanceof riscal.syntax.AST.Expression.InSetExp)
      return translate((riscal.syntax.AST.Expression.InSetExp)exp);
    if (exp instanceof riscal.syntax.AST.Expression.NotInSetExp)
      return translate((riscal.syntax.AST.Expression.NotInSetExp)exp);
    if (exp instanceof riscal.syntax.AST.Expression.SubsetExp)
      return translate((riscal.syntax.AST.Expression.SubsetExp)exp);
    if (exp instanceof riscal.syntax.AST.Expression.NotSubsetExp)
      return translate((riscal.syntax.AST.Expression.NotSubsetExp)exp);
    if (exp instanceof riscal.syntax.AST.Expression.NotExp)
      return translate((riscal.syntax.AST.Expression.NotExp)exp);
    if (exp instanceof riscal.syntax.AST.Expression.AndExp)
      return translate((riscal.syntax.AST.Expression.AndExp)exp);
    if (exp instanceof riscal.syntax.AST.Expression.OrExp)
      return translate((riscal.syntax.AST.Expression.OrExp)exp);
    if (exp instanceof riscal.syntax.AST.Expression.ImpliesExp)
      return translate((riscal.syntax.AST.Expression.ImpliesExp)exp);
    if (exp instanceof riscal.syntax.AST.Expression.EquivExp)
      return translate((riscal.syntax.AST.Expression.EquivExp)exp);
    if (exp instanceof riscal.syntax.AST.Expression.IfThenElseExp)
      return translate((riscal.syntax.AST.Expression.IfThenElseExp)exp);
    if (exp instanceof riscal.syntax.AST.Expression.MatchExp)
      return translate((riscal.syntax.AST.Expression.MatchExp)exp);
    if (exp instanceof riscal.syntax.AST.Expression.ForallExp)
      return translate((riscal.syntax.AST.Expression.ForallExp)exp);
    if (exp instanceof riscal.syntax.AST.Expression.ExistsExp)
      return translate((riscal.syntax.AST.Expression.ExistsExp)exp);
    if (exp instanceof riscal.syntax.AST.Expression.LetExp)
      return translate((riscal.syntax.AST.Expression.LetExp)exp);
    if (exp instanceof riscal.syntax.AST.Expression.LetParExp)
      return translate((riscal.syntax.AST.Expression.LetParExp)exp);
    if (exp instanceof riscal.syntax.AST.Expression.ChooseInExp)
      return translate((riscal.syntax.AST.Expression.ChooseInExp)exp);
    if (exp instanceof riscal.syntax.AST.Expression.ChooseInElseExp)
      return translate((riscal.syntax.AST.Expression.ChooseInElseExp)exp);
    if (exp instanceof riscal.syntax.AST.Expression.AssertExp)
      return translate((riscal.syntax.AST.Expression.AssertExp)exp);
    if (exp instanceof riscal.syntax.AST.Expression.PrintExp)
      return translate((riscal.syntax.AST.Expression.PrintExp)exp);
    if (exp instanceof riscal.syntax.AST.Expression.PrintInExp)
      return translate((riscal.syntax.AST.Expression.PrintInExp)exp);
    if (exp instanceof riscal.syntax.AST.Expression.PrintTraceInExp)
      return translate((riscal.syntax.AST.Expression.PrintTraceInExp)exp);
    if (exp instanceof riscal.syntax.AST.Expression.CheckExp)
      return translate((riscal.syntax.AST.Expression.CheckExp)exp);
    TranslationError.raise("unknown expression. " + exp);
    return null;
  }
  
  private List<Exp> translate(riscal.syntax.AST.Expression[] exps)
  {
    List<Exp> exps0 = new ArrayList<Exp>(exps.length);
    for (riscal.syntax.AST.Expression exp : exps)
      exps0.add(translate(exp));
    return exps0;
  }
  
  private Exp translate(riscal.syntax.AST.Expression.UnitExp exp)
  {
    return False.create();
  }
  
  private Exp translate(riscal.syntax.AST.Expression.TrueExp exp)
  {
    return True.create();
  }
  
  private Exp translate(riscal.syntax.AST.Expression.FalseExp exp)
  {
    return False.create();
  }
  
  private Exp translate(riscal.syntax.AST.Expression.NumberLiteralExp exp)
  {
    return apply(exp.literal.value);
  }
  
  private Exp translate(riscal.syntax.AST.Expression.EmptySetExp exp)
  {
    return emptySet(translate(exp.type));
  }
  
  private Exp translate(riscal.syntax.AST.Expression.RecIdentifierExp exp)
  {
    String name = dtname(translate(exp.ident1), translate(exp.ident2));
    return apply(name);
  }
  
  private Exp translate(riscal.syntax.AST.Expression.RecApplicationExp exp)
  {
    String name = dtname(translate(exp.ident1), translate(exp.ident2));
    List<Exp> exps = new ArrayList<Exp>(exp.exps.length);
    for (riscal.syntax.AST.Expression exp0 : exp.exps)
      exps.add(translate(exp0));
    return apply(name, exps);
  }
  
  private Exp translate(riscal.syntax.AST.Expression.IdentifierExp exp)
  {
    return apply(exp.ident.string);
  }
  
  private Exp translate(riscal.syntax.AST.Expression.ApplicationExp exp)
  {
    return apply(functionName(exp.ident.string), translate(exp.exps));
  }
  
  private Exp translate(riscal.syntax.AST.Expression.PowersetExp exp)
  {
    Type etype = elemType(translate(exp.exp.getType()));    
    return powerSet(etype, translate(exp.exp), null);
  }
  
  private Exp translate(riscal.syntax.AST.Expression.Powerset1Exp exp)
  {
    Type etype = elemType(translate(exp.exp1.getType())); 
    Exp exp1 = translate(exp.exp1); 
    Exp exp2 = translate(exp.exp2);
    createSizeFun(etype);
    return powerSet(etype, exp1,
        (Exp e)-> apply(AST.EQUALNAME, apply(SIZENAME, e),  exp2));
  }
  
  private Exp translate(riscal.syntax.AST.Expression.Powerset2Exp exp)
  {
    Type etype = elemType(translate(exp.exp1.getType()));
    Exp exp1 = translate(exp.exp1); 
    Exp exp2 = translate(exp.exp2);
    Exp exp3 = translate(exp.exp3);
    createSizeFun(etype);
    return powerSet(etype, exp1,
        (Exp e)-> And.create(
            apply(AST.LESSEQUALNAME, exp2, apply(SIZENAME, e)),
            apply(AST.LESSEQUALNAME, apply(SIZENAME, e), exp3)));
  }
  
  private Exp translate(riscal.syntax.AST.Expression.HeightExp exp)
  {
    return apply(AST.HEIGHTNAME, translate(exp.exp));
  }
  
  private Exp translate(riscal.syntax.AST.Expression.MapSelectionExp exp)
  {
    return apply(AST.MAPSELECTNAME, translate(exp.exp1), translate(exp.exp2));
  }
  
  private Exp translate(riscal.syntax.AST.Expression.TupleSelectionExp exp)
  {
    Id id = Id.create(new String[] { AST.TUPLESELECTNAME, exp.index.value });
    return Apply.create(id, Arrays.asList(new Exp[] { translate(exp.exp) }));
  }
  
  private Exp translate(riscal.syntax.AST.Expression.RecordSelectionExp exp)
  {
    int index = getRecordIndex(exp.exp, exp.ident);
    Id id = Id.create(new String[] { AST.TUPLESELECTNAME, Integer.toString(index) });
    return Apply.create(id, Arrays.asList(new Exp[] { translate(exp.exp) }));
  }
  
  private Exp translate(riscal.syntax.AST.Expression.NegationExp exp)
  {
    return apply(AST.MINUSNAME, translate(exp.exp));
  }
  
  private Exp translate(riscal.syntax.AST.Expression.FactorialExp exp)
  {
    return apply(PRODUCTNAME, apply("1"), translate(exp.exp));
  }
  
  private Exp translate(riscal.syntax.AST.Expression.BigUnionExp exp)
  {
    Type stype = translate(exp.getType());
    Type etype = elemType(stype);
    LetBinder binder = LetBinder.create(Id.create("bigset"), translate(exp.exp));
    return Let.create(Arrays.asList(new LetBinder[] { binder }),
        Choose.create(true, TypedVar.create(Id.create("bigunion"), stype),
            Forall.create(TypedVar.create(Id.create("elem"), etype),
                Equiv.create(
                    apply(AST.MAPSELECTNAME, var("bigunion"), var("elem")),
                    Exists.create(TypedVar.create(Id.create("smallset"), stype),
                        And.create(
                            apply(AST.MAPSELECTNAME, var("bigset"), var("smallset")), 
                            apply(AST.MAPSELECTNAME, var("smallset"), var("elem"))))))));
  }
  
  private Exp translate(riscal.syntax.AST.Expression.BigIntersectExp exp)
  {
    Type stype = translate(exp.getType());
    Type etype = elemType(stype);
    LetBinder binder = LetBinder.create(Id.create("bigset"), translate(exp.exp));
    return Let.create(Arrays.asList(new LetBinder[] { binder }),
        Choose.create(true, TypedVar.create(Id.create("bigunion"), stype),
            Forall.create(TypedVar.create(Id.create("elem"), etype),
                Equiv.create(
                    apply(AST.MAPSELECTNAME, var("bigunion"), var("elem")),
                    Forall.create(TypedVar.create(Id.create("smallset"), stype),
                        Imp.create(
                            apply(AST.MAPSELECTNAME, var("bigset"), var("smallset")), 
                            apply(AST.MAPSELECTNAME, var("smallset"), var("elem"))))))));
  }
  
  private Exp translate(riscal.syntax.AST.Expression.PowerExp exp)
  {
    return apply(POWERNAME, translate(exp.exp1), translate(exp.exp2));
  }
  
  private Exp translate(riscal.syntax.AST.Expression.TimesExp exp)
  {
    return apply(AST.MULTNAME, translate(exp.exp1), translate(exp.exp2));
  }
  
  private Exp translate(riscal.syntax.AST.Expression.TimesExpMult exp)
  {
    return apply(PRODUCTNAME, translate(exp.exp1), translate(exp.exp2));
  }
  
  private Exp translate(riscal.syntax.AST.Expression.DividesExp exp)
  {
    return apply(AST.DIVNAME, translate(exp.exp1), translate(exp.exp2));
  }
  
  private Exp translate(riscal.syntax.AST.Expression.RemainderExp exp)
  {
    return apply(AST.MODNAME, translate(exp.exp1), translate(exp.exp2));
  }
  
  private Exp translate(riscal.syntax.AST.Expression.PlusExp exp)
  {
    return apply(AST.PLUSNAME, translate(exp.exp1), translate(exp.exp2));
  }
  
  private Exp translate(riscal.syntax.AST.Expression.PlusExpMult exp)
  {
    return apply(SUMNAME, translate(exp.exp1), translate(exp.exp2));
  }
  
  private Exp translate(riscal.syntax.AST.Expression.MinusExp exp)
  {
    return apply(AST.MINUSNAME, translate(exp.exp1), translate(exp.exp2));
  }
  
  private Exp translate(riscal.syntax.AST.Expression.EnumeratedSetExp exp)
  {
    Type stype = translate(exp.getType());
    Exp set = createMap(elemType(stype), type(AST.BOOLNAME), False.create());
    for (riscal.syntax.AST.Expression exp0 : exp.exps)
    {
      Exp exp1 = translate(exp0);
      set = apply(AST.MAPSTORENAME, set, exp1, True.create());
    }
    return set;
  }
  
  private Exp translate(riscal.syntax.AST.Expression.CartesianExp exp)
  {
    int n = exp.exps.length;
    if (n == 1) return translate(exp.exps[0]);
    List<Exp> exps = new ArrayList<Exp>(n);
    List<Type> types = new ArrayList<Type>(n);
    List<LetBinder> binders = new ArrayList<LetBinder>(n);
    Exp formula = null;
    for (int i = 0; i < n; i++)
    {
      exps.add(translate(exp.exps[i]));
      riscal.syntax.AST.Type.SetType stype = 
          (riscal.syntax.AST.Type.SetType)exp.exps[i].getType();
      types.add(translate(stype.type));
      Id setid = Id.create("set" + i);
      Id selectid = Id.create(new String[] { AST.TUPLESELECTNAME, Integer.toString(i+1) });
      binders.add(LetBinder.create(setid, exps.get(i)));
      formula = Exp.and(formula,
          apply(AST.MAPSELECTNAME, Var.create(setid),
              Apply.create(selectid, Arrays.asList(new Exp[] { var("tuple") }))));
    }
    Type type = type(AST.TUPLENAME, types);
    return Let.create(binders,
        Choose.create(true, TypedVar.create(Id.create("product"), setType(type)),
            Forall.create(TypedVar.create(Id.create("tuple"), type),
                Equiv.create(
                    apply(AST.MAPSELECTNAME, var("product"), var("tuple")),
                    formula))));
  }
  
  private Exp translate(riscal.syntax.AST.Expression.IntervalExp exp)
  {
    Quantification quant = new Quantification();
    Id id = unusedId("elem");
    Type type = type(AST.INTNAME);
    TypedVar tvar = TypedVar.create(id, type);
    quant.ids.add(id);
    quant.types.add(type);
    quant.tvars.add(tvar);
    quant.exp = And.create(
        apply(AST.LESSEQUALNAME, translate(exp.exp1), Var.create(id)),
        apply(AST.LESSEQUALNAME, Var.create(id), translate(exp.exp2)));
    return createSetBuilder(type, quant, Var.create(id));
  }
  
  private Exp translate(riscal.syntax.AST.Expression.TupleExp exp)
  {
    return apply(AST.TUPLECONSTRUCTNAME, translate(exp.exps));
  }
  
  private Exp translate(riscal.syntax.AST.Expression.RecordExp exp)
  {
    List<Exp> exps = new ArrayList<Exp>(exp.eidents.length);
    for (riscal.syntax.AST.ExpressionIdentifier eident : exp.eidents)
      exps.add(translate(eident.exp));
    return apply(AST.TUPLECONSTRUCTNAME, exps);
  }
  
  private Exp translate(riscal.syntax.AST.Expression.SetSizeExp exp)
  {
    Type etype = elemType(translate(exp.exp.getType()));
    createSizeFun(etype);
    return apply(SIZENAME, translate(exp.exp));
  }
  
  private Exp translate(riscal.syntax.AST.Expression.UnionExp exp)
  {
    Id sid = Id.create("set");
    Id eid = Id.create("elem");
    Type stype = translate(exp.exp1.getType());
    Type etype = elemType(stype);
    LetBinder binder1 = LetBinder.create(Id.create("set1"), translate(exp.exp1));
    LetBinder binder2 = LetBinder.create(Id.create("set2"), translate(exp.exp2));
    return Let.create(Arrays.asList(new LetBinder[] { binder1, binder2 }),        
        Choose.create(true, TypedVar.create(sid, stype),
            Forall.create(TypedVar.create(eid, etype),
                Equiv.create(
                    apply(AST.MAPSELECTNAME, Var.create(sid), Var.create(eid)),
                    Or.create(
                        apply(AST.MAPSELECTNAME, Var.create(binder1.id), Var.create(eid)),
                        apply(AST.MAPSELECTNAME, Var.create(binder2.id), Var.create(eid)))))));
  }
  
  private Exp translate(riscal.syntax.AST.Expression.IntersectExp exp)
  {
    Id sid = Id.create("set");
    Id eid = Id.create("elem");
    Type stype = translate(exp.exp1.getType());
    Type etype = elemType(stype);
    LetBinder binder1 = LetBinder.create(Id.create("set1"), translate(exp.exp1));
    LetBinder binder2 = LetBinder.create(Id.create("set2"), translate(exp.exp2));
    return Let.create(Arrays.asList(new LetBinder[] { binder1, binder2 }),        
        Choose.create(true, TypedVar.create(sid, stype),
            Forall.create(TypedVar.create(eid, etype),
                Equiv.create(
                    apply(AST.MAPSELECTNAME, Var.create(sid), Var.create(eid)),
                    And.create(
                        apply(AST.MAPSELECTNAME, Var.create(binder1.id), Var.create(eid)),
                        apply(AST.MAPSELECTNAME, Var.create(binder2.id), Var.create(eid)))))));
  }
  
  private Exp translate(riscal.syntax.AST.Expression.WithoutExp exp)
  {
    Id sid = Id.create("set");
    Id eid = Id.create("elem");
    Type stype = translate(exp.exp1.getType());
    Type etype = elemType(stype);
    LetBinder binder1 = LetBinder.create(Id.create("set1"), translate(exp.exp1));
    LetBinder binder2 = LetBinder.create(Id.create("set2"), translate(exp.exp2));
    return Let.create(Arrays.asList(new LetBinder[] { binder1, binder2 }),        
        Choose.create(true, TypedVar.create(sid, stype),
            Forall.create(TypedVar.create(eid, etype),
                Equiv.create(
                    apply(AST.MAPSELECTNAME, Var.create(sid), Var.create(eid)),
                    And.create(
                        apply(AST.MAPSELECTNAME, Var.create(binder1.id), Var.create(eid)),
                        Not.create(apply(AST.MAPSELECTNAME, Var.create(binder2.id), Var.create(eid))))))));
  }
  
  private Exp translate(riscal.syntax.AST.Expression.NumberExp exp)
  {
    int n = exp.qvar.qvcore.length;
    List<Type> types = new ArrayList<Type>(n);
    List<Exp> exps = new ArrayList<Exp>(n);
    for (int i = 0; i < n; i++)
    {
      riscal.types.Environment.Symbol.ValueSymbol symbol =
          (riscal.types.Environment.Symbol.ValueSymbol)exp.qvar.qvcore[i].ident.getSymbol();
      types.add(translate(symbol.type));
      exps.add(Var.create(translate(symbol.ident)));
    }
    Type etype = n == 1 ? types.get(0) : type(AST.TUPLENAME, types);
    Exp exp0 = n == 1 ? exps.get(0) : apply(AST.TUPLECONSTRUCTNAME, exps);
    Quantification quant = quantification(exp.qvar);
    createSizeFun(etype);
    return apply(SIZENAME, createSetBuilder(etype, quant, exp0));
  }
  
  private Exp translate(riscal.syntax.AST.Expression.SetBuilderExp exp)
  {
    Type stype = translate(exp.getType());
    Type etype = elemType(stype);
    Quantification quant = quantification(exp.qvar);
    return createSetBuilder(etype, quant, translate(exp.exp));
  }

  private Exp translate(riscal.syntax.AST.Expression.SumExp exp)
  {
    Type etype = translate(exp.getType());
    Quantification quant = quantification(exp.qvar);
    return apply(SUMNAME, createSetBuilder(etype, quant, translate(exp.exp)));
  }
  
  private Exp translate(riscal.syntax.AST.Expression.ProductExp exp)
  {
    Type etype = translate(exp.getType());
    Quantification quant = quantification(exp.qvar);
    return apply(PRODUCTNAME, createSetBuilder(etype, quant, translate(exp.exp)));
  }
  
  private Exp translate(riscal.syntax.AST.Expression.MinExp exp)
  {
    Quantification quant = quantification(exp.qvar);
    Exp exp0 = translate(exp.exp);
    Type etype = translate(exp.getType());
    return createUniqueChoice(AST.LESSEQUALNAME, quant, exp0, etype);
  }
  
  private Exp translate(riscal.syntax.AST.Expression.MaxExp exp)
  {
    Quantification quant = quantification(exp.qvar);
    Exp exp0 = translate(exp.exp);
    Type etype = translate(exp.getType());
    return createUniqueChoice(AST.GREATEREQUALNAME, quant, exp0, etype);
  }
  
  private Exp translate(riscal.syntax.AST.Expression.ArrayBuilderExp exp)
  {
    Exp exp1 = apply("0");
    Exp exp2 = apply(AST.MINUSNAME, translate(exp.exp1), apply("1"));
    return createMap(intType(exp1, exp2), translate(exp.type), translate(exp.exp2));
  }
  
  private Exp translate(riscal.syntax.AST.Expression.MapBuilderExp exp)
  {
    return createMap(translate(exp.type1), translate(exp.type2), translate(exp.exp));
  }
  
  private Exp translate(riscal.syntax.AST.Expression.MapUpdateExp exp)
  {
    return apply(AST.MAPSTORENAME, 
        Arrays.asList(new Exp[] { translate(exp.exp1), 
            translate(exp.exp2), translate(exp.exp3)}));
  }
  
  private Exp translate(riscal.syntax.AST.Expression.TupleUpdateExp exp)
  {
    Id id = Id.create(new String[] { AST.TUPLESTORENAME, exp.index.value });
    return Apply.create(id, Arrays.asList(new Exp[] { translate(exp.exp1), translate(exp.exp2) }));
  }
  
  private Exp translate(riscal.syntax.AST.Expression.RecordUpdateExp exp)
  {
    int index = getRecordIndex(exp.exp1, exp.ident);
    Id id = Id.create(new String[] { AST.TUPLESTORENAME, Integer.toString(index) });
    return Apply.create(id, Arrays.asList(new Exp[] { translate(exp.exp1), translate(exp.exp2) }));
  }
  
  private Exp translate(riscal.syntax.AST.Expression.EqualsExp exp)
  {
    return apply(AST.EQUALNAME, translate(exp.exp1), translate(exp.exp2));
  }
  
  private Exp translate(riscal.syntax.AST.Expression.NotEqualsExp exp)
  {
    return apply(AST.NOTEQUALNAME, translate(exp.exp1), translate(exp.exp2));
  }
  
  private Exp translate(riscal.syntax.AST.Expression.LessExp exp)
  {
    return apply(AST.LESSNAME, translate(exp.exp1), translate(exp.exp2));
  }
  
  private Exp translate(riscal.syntax.AST.Expression.LessEqualExp exp)
  {
    return apply(AST.LESSEQUALNAME, translate(exp.exp1), translate(exp.exp2));
  }
  
  private Exp translate(riscal.syntax.AST.Expression.GreaterExp exp)
  {
    return apply(AST.GREATERNAME, translate(exp.exp1), translate(exp.exp2));
  }
  
  private Exp translate(riscal.syntax.AST.Expression.GreaterEqualExp exp)
  {
    return apply(AST.GREATEREQUALNAME, translate(exp.exp1), translate(exp.exp2));
  }
  
  private Exp translate(riscal.syntax.AST.Expression.InSetExp exp)
  {
    Exp exp1 = translate(exp.exp1);
    Exp exp2 = translate(exp.exp2);
    return apply(AST.MAPSELECTNAME, exp2, exp1);
  }
  
  private Exp translate(riscal.syntax.AST.Expression.NotInSetExp exp)
  {
    Exp exp1 = translate(exp.exp1);
    Exp exp2 = translate(exp.exp2);
    return Not.create(apply(AST.MAPSELECTNAME, exp2, exp1));
  }
  
  private Exp translate(riscal.syntax.AST.Expression.SubsetExp exp)
  {
    Id eid = Id.create("elem");
    Type stype = translate(exp.exp1.getType());
    Type etype = elemType(stype);
    LetBinder binder1 = LetBinder.create(Id.create("set1"), translate(exp.exp1));
    LetBinder binder2 = LetBinder.create(Id.create("set2"), translate(exp.exp2));
    return Let.create(Arrays.asList(new LetBinder[] { binder1, binder2 }),
        Forall.create(TypedVar.create(eid, etype),
        Imp.create(
            apply(AST.MAPSELECTNAME, Var.create(binder1.id), Var.create(eid)),
            apply(AST.MAPSELECTNAME, Var.create(binder2.id), Var.create(eid)))));      
  }
  
  private Exp translate(riscal.syntax.AST.Expression.NotSubsetExp exp)
  {
    Id eid = Id.create("elem");
    Type stype = translate(exp.exp1.getType());
    Type etype = elemType(stype);
    LetBinder binder1 = LetBinder.create(Id.create("set1"), translate(exp.exp1));
    LetBinder binder2 = LetBinder.create(Id.create("set2"), translate(exp.exp2));
    return Let.create(Arrays.asList(new LetBinder[] { binder1, binder2 }),
        Exists.create(TypedVar.create(eid, etype),
        And.create(
            apply(AST.MAPSELECTNAME, Var.create(binder1.id), Var.create(eid)),
            Not.create(apply(AST.MAPSELECTNAME, Var.create(binder2.id), Var.create(eid))))));  
  }
  
  private Exp translate(riscal.syntax.AST.Expression.NotExp exp)
  {
    return Not.create(translate(exp.exp));
  }
  
  private Exp translate(riscal.syntax.AST.Expression.AndExp exp)
  {
    return And.create(translate(exp.exp1), translate(exp.exp2));
  }
  
  private Exp translate(riscal.syntax.AST.Expression.OrExp exp)
  {
    return Or.create(translate(exp.exp1), translate(exp.exp2));
  }
  
  private Exp translate(riscal.syntax.AST.Expression.ImpliesExp exp)
  {
    return Imp.create(translate(exp.exp1), translate(exp.exp2));
  }
  
  private Exp translate(riscal.syntax.AST.Expression.EquivExp exp)
  {
    return Equiv.create(translate(exp.exp1), translate(exp.exp2));
  }

  private Exp translate(riscal.syntax.AST.Expression.IfThenElseExp exp)
  {
    return If.create(translate(exp.exp1), translate(exp.exp2), translate(exp.exp3));
  }
  
  private Exp translate(riscal.syntax.AST.Expression.MatchExp exp)
  {
    Exp exp0 = translate(exp.exp);
    Type type = translate(exp.exp.getType());
    List<MatchBinder> binders = new ArrayList<MatchBinder>(exp.pexp.length);
    for (riscal.syntax.AST.PatternExpression pexp : exp.pexp)
      binders.add(translate(type.id, pexp));
    return Match.create(exp0, binders);
  }
  
  private MatchBinder translate(Id id, riscal.syntax.AST.PatternExpression pexp)
  {
    if (pexp instanceof riscal.syntax.AST.PatternExpression.DefaultPatternExp)
      return translate(id, (riscal.syntax.AST.PatternExpression.DefaultPatternExp)pexp);
    if (pexp instanceof riscal.syntax.AST.PatternExpression.IdentifierPatternExp)
      return translate(id, (riscal.syntax.AST.PatternExpression.IdentifierPatternExp)pexp);
    if (pexp instanceof riscal.syntax.AST.PatternExpression.ApplicationPatternExp)
      return translate(id, (riscal.syntax.AST.PatternExpression.ApplicationPatternExp)pexp);
    TranslationError.raise("unknown pattern expression: " + pexp);
    return null;
  }
  
  private MatchBinder translate(Id id, riscal.syntax.AST.PatternExpression.DefaultPatternExp pexp)
  {
    return MatchBinder.create(Pattern.DefaultPattern.create(), translate(pexp.exp));
  }
  
  private MatchBinder translate(Id id, riscal.syntax.AST.PatternExpression.IdentifierPatternExp pexp)
  {
    String name = dtname(id, translate(pexp.ident));
    Pattern pattern = Pattern.ConstrPattern.create(Id.create(name), new ArrayList<TypedVar>(0));
    return MatchBinder.create(pattern, translate(pexp.exp));
  }
  
  private MatchBinder translate(Id id, riscal.syntax.AST.PatternExpression.ApplicationPatternExp pexp)
  {
    String name = dtname(id, translate(pexp.ident));
    List<TypedVar> tvars = new ArrayList<TypedVar>(pexp.params.length);
    for (riscal.syntax.AST.Parameter param : pexp.params)
      tvars.add(translate(param));
    Pattern pattern = Pattern.ConstrPattern.create(Id.create(name), tvars);
    return MatchBinder.create(pattern, translate(pexp.exp));
  }
  
  private Exp translate(riscal.syntax.AST.Expression.ForallExp exp)
  {
    return translate(exp.qvar, translate(exp.exp),
        (Exp exp0, TypedVar tvar)->Forall.create(tvar, exp0),
        (Exp exp0, Exp exp1)->Imp.create(exp1, exp0));
  }
  
  private Exp translate(riscal.syntax.AST.Expression.ExistsExp exp)
  {
    return translate(exp.qvar, translate(exp.exp),
        (Exp exp0, TypedVar tvar)->Exists.create(tvar, exp0),
        (Exp exp0, Exp exp1)->And.create(exp1, exp0));
  }
  
  private Exp translate(riscal.syntax.AST.Expression.LetExp exp)
  {
    // in RISCTP, let binders bind simultaneously, not sequentially
    List<LetBinder> binders = translate(exp.binder);
    Exp exp0 = translate(exp.exp);
    int n = binders.size();
    for (int i = n-1; i >= 0; i--)
      exp0 = Let.create(Arrays.asList(new LetBinder[] { binders.get(i) }), exp0);
    return exp0;
  }
  
  private Exp translate(riscal.syntax.AST.Expression.LetParExp exp)
  {
    // in RISCTP, let binders bind simultaneously, not sequentially
    return Let.create(translate(exp.binder), translate(exp.exp));
  }
   
  private Exp translate(riscal.syntax.AST.Expression.ChooseExp exp)
  {
    Quantification quant = quantification(exp.qvar);
    return createChoose(exp.getSat(), quant);
  }
  
  private Exp translate(riscal.syntax.AST.Expression.ChooseInExp exp)
  {
    Quantification quant = quantification(exp.qvar);
    return createChoose(quant, translate(exp.exp));
  }
  
  private Exp translate(riscal.syntax.AST.Expression.ChooseInElseExp exp)
  {
    Quantification quant = quantification(exp.qvar);
    Exp choice = createChoose(quant, translate(exp.exp1));
    if (quant.exp == null) return choice;
    Exp cond = translate(exp.qvar, True.create(),
        (Exp exp0, TypedVar tvar)->Exists.create(tvar, exp0),
        (Exp exp0, Exp exp1)->And.create(exp1, exp0));
    return If.create(cond, choice, translate(exp.exp2));
  }
  
  private Exp translate(riscal.syntax.AST.Expression.AssertExp exp)
  {
    // ignore assertion
    return translate(exp.exp2);
  }
  
  private Exp translate(riscal.syntax.AST.Expression.PrintExp exp)
  {
    return translate(exp.exp);
  }
  
  private Exp translate(riscal.syntax.AST.Expression.PrintInExp exp)
  {
    return translate(exp.exp);
  }
  
  private Exp translate(riscal.syntax.AST.Expression.PrintTraceInExp exp)
  {
    return translate(exp.exp);
  }
  
  private Exp translate(riscal.syntax.AST.Expression.CheckExp exp)
  {
    // ignore check
    return True.create();
  }
  
  // --------------------------------------------------------------------------
  //
  // miscellaneous
  //
  // --------------------------------------------------------------------------
  
  private <T> T translate(riscal.syntax.AST.QuantifiedVariable qvar, T exp,
    BiFunction<T,TypedVar,T> addVar, BiFunction<T,Exp,T> addFormula)
  {
    if (qvar.exp != null) exp = addFormula.apply(exp, translate(qvar.exp));
    int n = qvar.qvcore.length;
    for (int i = n-1; i >= 0; i--)
    {
      riscal.syntax.AST.QuantifiedVariableCore qvc = qvar.qvcore[i];
      exp = translate(qvc, exp, addVar, addFormula);
    }
    return exp;
  }
  
  private <T> T translate(riscal.syntax.AST.QuantifiedVariableCore qvc, T exp,
    BiFunction<T,TypedVar,T> addVar, BiFunction<T,Exp,T> addFormula)
  {
    if (qvc.wexp != null) exp = addFormula.apply(exp, translate(qvc.wexp));
    if (qvc instanceof riscal.syntax.AST.QuantifiedVariableCore.IdentifierTypeQuantifiedVar)
      return translate((riscal.syntax.AST.QuantifiedVariableCore.IdentifierTypeQuantifiedVar)qvc, exp, addVar, addFormula);
    if (qvc instanceof riscal.syntax.AST.QuantifiedVariableCore.IdentifierSetQuantifiedVar)
      return translate((riscal.syntax.AST.QuantifiedVariableCore.IdentifierSetQuantifiedVar)qvc, exp, addVar, addFormula);
    TranslationError.raise("unknown quantified variable core: " + qvc);
    return null;
  }
  
  private <T> T translate(riscal.syntax.AST.QuantifiedVariableCore.IdentifierTypeQuantifiedVar qvc, T exp,
    BiFunction<T,TypedVar,T> addVar, BiFunction<T,Exp,T> addFormula)
  {
    return addVar.apply(exp, TypedVar.create(translate(qvc.ident), translate(qvc.type)));
  }
  
  private <T> T translate(riscal.syntax.AST.QuantifiedVariableCore.IdentifierSetQuantifiedVar qvc, T exp,
    BiFunction<T,TypedVar,T> addVar, BiFunction<T,Exp,T> addFormula)
  {
    riscal.syntax.AST.Expression exp0 = new riscal.syntax.AST.Expression.InSetExp(
        new riscal.syntax.AST.Expression.IdentifierExp(qvc.ident), qvc.exp);
    exp = addFormula.apply(exp, translate(exp0));
    // unfortunately, just a semantic type, thus we may have no static bounds on element type
    riscal.syntax.AST.Type.SetType stype = (riscal.syntax.AST.Type.SetType)qvc.exp.getType();
    return addVar.apply(exp, TypedVar.create(translate(qvc.ident), translate(stype.type)));
  }
  
  private List<LetBinder> translate(riscal.syntax.AST.Binder[] binders)
  {
    List<LetBinder> binders0 = new ArrayList<LetBinder>(binders.length);
    for (riscal.syntax.AST.Binder binder : binders)
      binders0.add(translate(binder));
    return binders0;
  }
  
  private LetBinder translate(riscal.syntax.AST.Binder binder)
  {
    return LetBinder.create(translate(binder.ident), translate(binder.exp));
  }
  
  private List<TypedVar> translate(riscal.syntax.AST.Parameter[] params)
  {
    List<TypedVar> tvars = new ArrayList<TypedVar>(params.length);
    for (riscal.syntax.AST.Parameter param : params)
      tvars.add(translate(param));
    return tvars;
  }
  
  private TypedVar translate(riscal.syntax.AST.Parameter param)
  {
    TypedVar tvar = TypedVar.create(translate(param.ident), translate(param.type));
    return tvar;
  }
  
  private Id translate(riscal.syntax.AST.Identifier id)
  {
    // in RISCAL, constants and functions are in separate namespaces
    // in RISCTP, they are in the same, thus we have to rename one to avoid conflicts
    // constant names are stored in some symbols, so we rename functions
    riscal.types.Environment.Symbol symbol = id.getSymbol();
    String name = id.string;
    if (symbol instanceof riscal.types.Environment.Symbol.FunctionSymbol)
      name = functionName(name);
    return Id.create(name);
  }
  
  // the translated name of a function
  private static String functionName(String name) 
  { return name + AST.internal("fun"); }
}
// ---------------------------------------------------------------------------
// end of file
// ---------------------------------------------------------------------------
