// ---------------------------------------------------------------------------
// DeclarationTranslator.java
// A class which is dedicated to translate declarations to SMT-LIB strings.
//
// Author: Franz Reichl <franz.x.reichl@gmail.com>
// $Id: DeclarationTranslator.java,v 1.7 2020/05/25 11:42:22 schreine Exp $
// 
// Author: Franz Reichl <franz.x.reichl@gmail.com>
// Copyright (C) 2019-, Research Institute for Symbolic Computation (RISC)
// Johannes Kepler University, Linz, Austria, http://www.risc.jku.at
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program.  If not, see <http://www.gnu.org/licenses/>.
// ---------------------------------------------------------------------------
package riscal.smt;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.stream.Collectors;
import java.util.stream.Stream;

import riscal.smt.DeclarationPreprocessor.TranslationInformation;
import riscal.syntax.AST.Declaration;
import riscal.syntax.AST.Declaration.FunctionDeclaration;
import riscal.syntax.AST.Declaration.FunctionDefinition;
import riscal.syntax.AST.Declaration.TheoremDefinition;
import riscal.syntax.AST.Declaration.TypeDefinition;
import riscal.syntax.AST.Declaration.ValueDefinition;
import riscal.syntax.AST.Type;
import riscal.syntax.AST.Type.BoolType;
import riscal.syntax.AST.Type.SetType;
import riscal.syntax.AST.Type.SubType;
import riscal.types.Environment.Symbol.FunctionSymbol;
import riscal.types.Environment.Symbol.TypeSymbol;
import riscal.types.Environment.Symbol.ValueSymbol;
import riscal.types.TypeChecker;


public class DeclarationTranslator 
{
	/**
	 * Used to translate the expressions in the declarations.
	 */
	private ExpressionTranslator expTrans;
	/**
	 * If in the processing of a declaration an additional declaration
	 * has to be introduced it is added here.
	 * E.g. the tuple access and update functions.
	 */
	private List<String> additionalDeclarations;
	
	private boolean useRemoveableDeclarations=false;
	private List<String> additionalDeclarations_RemoveableInIterativeMode;
	/**
	 * A collection of names of different entities that must not be used for new
	 * (function) names.
	 */
	private Set<String> forbiddenNames;

	/**
	 * Contains the names of functions of the original specifications,
	 * that shall be considered as defined.
	 */
	private Set<String> definedFunctions;
	/**
	 * Used to assign the names of functions that represent the condition
	 * of a subtype to the respective subtype.
	 */
	private Map<Type,String> typeConditions=new HashMap<Type, String>();
	/**
	 * Collection of the valuesymbol of the constants in the current setup.
	 */
	private Set<ValueSymbol> values=new HashSet<ValueSymbol>();
	/**
	 * @return The values of the constants.
	 */
	Set<ValueSymbol> getVals()
	{
		return values;
	}
	
	private boolean unquantify;
	boolean getUnquantify()
	{
		return unquantify;
	}
	
	/**
	 * The representation of the single value of the unit type as a bit vector without prefix.
	 */
	static final String unitBareVector="0";
	/**
	 * The representation of the single value of the unit type as a bit vector.
	 */
	static final String unit="#b0";
	
	/***************************************************************************
	 * A Declaration Translator for the given function names.
	 * @param tI Contains the names part of the declarations,
	 * strings that must not be used for new identifiers/variables/...
	 * and the names of the function which shall be considered as defined.
	 * @param eliminateQuantifiers Determines whether quantifiers shall be eliminated.
	 **************************************************************************/
	private DeclarationTranslator(TranslationInformation tI, boolean eliminateQuantifiers)
	{
		this.unquantify=eliminateQuantifiers;
		this.forbiddenNames=tI.forbiddenNames;
		this.definedFunctions=tI.definedFunctions;
		additionalDeclarations=new ArrayList<String>();
		additionalDeclarations_RemoveableInIterativeMode=new ArrayList<String>();
		this.expTrans=new ExpressionTranslator(this);
	}
	
	/***************************************************************************
	 * Translates a given list of declarations to a SMT-LIB script.
	 * This method requires that the DeclarationPreprocessor was already
	 * applied to these declarations.
	 * @param declarations Declarations which were already processed by the
	 * 		DeclarationPreprocessor and which shall be translated to a SMT-LIB script.
	 * @param tI Contains the names part of the declarations,
	 * strings that must not be used for new identifiers/variables/...
	 * (this set has to contain SMT-LIB keywords)
	 * and the names of the function which shall be considered as defined.	 
	 * @param eliminateQuantifiers If true, quantifiers are eliminated.
	 * Otherwise SMT-LIB quantifiers are used to represent quantifiers,
	 * note that this is an experimental option.
	 * @return A SMT-LIB script corresponding to the list of declarations.
	 **************************************************************************/
	public static String translate(List<Declaration> declarations,DeclarationPreprocessor.TranslationInformation tI,boolean eliminateQuantifiers)
	{
		DeclarationTranslator decTrans=new DeclarationTranslator(tI,eliminateQuantifiers);
//		StringBuilder stB=new StringBuilder();
		List<String> strlst=new ArrayList<String>();
		for(Declaration dec:declarations)
		{
//			stB.append(decTrans.process(dec));
			strlst.addAll(decTrans.process(dec));
		}
		if(SMT.checkDependencies)
		{
			if(strlst.size()>0)
			{
				strlst.add(strlst.size()-1, "(check-sat)");
			}
		}
		strlst.add("(check-sat)");
		return String.join("", strlst);
//		stB.append("(check-sat)");
//		return stB.toString();
	}
	
	public static String translateMultiple(List<Declaration> dependencies,List<List<Declaration>> toCheck,
			DeclarationPreprocessor.TranslationInformation tI, boolean eliminateQuantifiers)
	{
		DeclarationTranslator decTrans=new DeclarationTranslator(tI,eliminateQuantifiers);
		List<String> strlst=new ArrayList<String>();
		for(Declaration dec:dependencies)
		{
			strlst.addAll(decTrans.process(dec));
		}
		for(List<Declaration> declst:toCheck)
		{
			List<String> removable=new ArrayList<String>();
			List<String> nonRemovable=new ArrayList<String>();
			for(Declaration dec:declst)
			{
				Utilities.Pair<List<String>, List<String>> p=decTrans.processGiveNonRemovableDeclarations(dec);
				nonRemovable.addAll(p.first);
				removable.addAll(p.second);
			}
			if(nonRemovable.size()==0)
				strlst.add("(push 1)\n");
			else
			{
				strlst.add("(push 1)\n");
				strlst.addAll(nonRemovable);
				strlst.add("(push 1)\n");
			}
			strlst.addAll(removable);				
			if(SMT.checkDependencies)
			{
				if(strlst.size()>0)
				{
					strlst.add(strlst.size()-1, "(check-sat)\n");
				}
			}
			strlst.add("(check-sat)\n");
			strlst.add("(pop 1)\n");
		}
		return String.join("", strlst);
	}
	
	/***************************************************************************
	 * Adds a declaration in the SMT-LIB format to the list of additional declarations.
	 * @param str The declarations which shall be added.
	 **************************************************************************/
	public void addDeclaration(String str)
	{
		additionalDeclarations.add(str+"\n");
	}
	
	public void addDeclarationRemovable(String str)
	{
		if(useRemoveableDeclarations)
			additionalDeclarations_RemoveableInIterativeMode.add(str+"\n");
		else
			additionalDeclarations.add(str+"\n");
	}
	
	/***************************************************************************
	 * Generates a (function) name which does not occur in forbiddenNames and
	 * adds this name to forbiddenNames.
	 * @param s A suggestion for the name.
	 * @return A fresh (function) name.
	 **************************************************************************/
	public String getNewFunctionName(String s)
	{
		return Utilities.NewIdentString(s, forbiddenNames);
	}
	
	/***************************************************************************
	 * Retrieves the name of the predicate which corresponds to the
	 * condition of a type.
	 * @param ty The type whose condition we want to retrieve.
	 * @return If the type is not a subtype null.
	 * Otherwise the name of the function which corresponds to the
	 * condition of the type.
	 **************************************************************************/
	String getTypeCondition(Type ty)
	{
		return typeConditions.get(ty);
	}
	
	/***************************************************************************
	 * Translates the denoted declaration to SMT-LIB.
	 * @param decl The declaration which shall be translated.
	 * @return SMT-LIB command(s) which correspond to the denoted declaration.
	 **************************************************************************/
	private List<String> process(Declaration decl)
	{
		useRemoveableDeclarations=false;
		additionalDeclarations.clear();
		String str;
		if(decl instanceof ValueDefinition)
			str=process((ValueDefinition)decl);
		else if(decl instanceof TypeDefinition)
			str=process((TypeDefinition)decl);
		else if(decl instanceof FunctionDeclaration)
			str=process((FunctionDeclaration)decl);
		else if(decl instanceof FunctionDefinition)
			str=process((FunctionDefinition)decl);
		else if(decl instanceof TheoremDefinition)
			str=process((TheoremDefinition)decl);
		else
		{
			SMT.notImplemented(decl, "The declaration could not be translated to SMT-LIB");
			return null;
		}
		List<String> strlst=new ArrayList<String>();
		strlst.addAll(additionalDeclarations);
		strlst.add(str);
		return strlst;
	}
	
	private Utilities.Pair<List<String>,List<String>> processGiveNonRemovableDeclarations(Declaration decl)
	{
		useRemoveableDeclarations=true;
		additionalDeclarations.clear();
		additionalDeclarations_RemoveableInIterativeMode.clear();
		String str;
		if(decl instanceof ValueDefinition)
			str=process((ValueDefinition)decl);
		else if(decl instanceof TypeDefinition)
			str=process((TypeDefinition)decl);
		else if(decl instanceof FunctionDeclaration)
			str=process((FunctionDeclaration)decl);
		else if(decl instanceof FunctionDefinition)
			str=process((FunctionDefinition)decl);
		else if(decl instanceof TheoremDefinition)
			str=process((TheoremDefinition)decl);
		else
		{
			SMT.notImplemented(decl, "The declaration could not be translated to SMT-LIB");
			return null;
		}
		List<String> strlst=new ArrayList<String>();
		strlst.addAll(additionalDeclarations_RemoveableInIterativeMode);
		strlst.add(str);
		return new Utilities.Pair<List<String>, List<String>>(additionalDeclarations,strlst);
	}
	
	/***************************************************************************
	 * Translates a RISCAL type to a SMT-LIB sort.
	 * @param type The sort which shall be translated. Shall be canonical.
	 * @return The corresponding SMT-LIB sort.
	 **************************************************************************/
	private String translateType(Type type)
	{
		return expTrans.translateType(type);
	}
	
	private String process(TypeDefinition dec)
	{
		Type ty=((TypeSymbol)dec.ident.getSymbol()).getType();
		Type baseType=Utilities.getBaseType(ty); 
		String str= "(define-sort "+dec.ident.string+" ()"+translateType(baseType)+")\n";
		if(ty instanceof SubType)
		{
			String name=getNewFunctionName(dec.ident.string+"_condition");
			String def=expTrans.compute(dec.exp);
			Type parentType=((SubType)ty).type;
			String s=typeConditions.get(parentType);
			String variable=TypeChecker.paramName;
			if(s!=null)
			{
				def="(and "+def+" ("+s+" "+variable+"))";
			}
			else
			{
				String cond=expTrans.generateTypeConditions(parentType, variable);
				if(cond!=null)
					def="(and "+cond+" "+def+")";
			}
			str+="(define-fun "+name+" (("+variable+" "+translateType(baseType)+")) Bool "+def+")\n";
			typeConditions.put(ty, name);
		}
		return str;
	}
	
	private String process(FunctionDeclaration dec)
	{
		String str="(declare-fun "+dec.ident.string;
		FunctionSymbol funSym=((FunctionSymbol)dec.ident.getSymbol());
		Stream<Type> s=Arrays.stream(funSym.types);
		String params=s.map(type -> this.translateType(type)).collect(Collectors.joining(" "));
		str+=" ("+params+") ";
		String resultType=this.translateType(funSym.type);
		str+=resultType+")";	
//		if(namesFromDeclaration.contains(dec.ident.string)&&!definedFunctions.contains(dec.ident.string))
		if(!definedFunctions.contains(dec.ident.string))
		{
			Type ty=funSym.type;
			if((ty instanceof BoolType)||
					((ty instanceof SetType)&&SMT.expandBaseTypesOfSubTypes))
			{
				return str+"\n";
			}
			List<String> functionArgumentNames=new ArrayList<String>();
			for(ValueSymbol val:((FunctionSymbol)dec.ident.getSymbol()).params)
			{
				functionArgumentNames.add(val.ident.string);
			}	
			String cond=expTrans.generateTypeConditions(ty, expTrans.generateSMTLIBFunctionApplication(dec.ident.string, functionArgumentNames));
			if(cond==null)
				return str+"\n";
			addDeclarationRemovable(str);
			List<String> variableNames=new ArrayList<String>();
			List<Type>variableTypes=new ArrayList<Type>();
			List<ValueSymbol> syms=Arrays.asList(((FunctionSymbol)dec.ident.getSymbol()).params);
			for(ValueSymbol val:Utilities.sortSymbols(syms))
			{
				variableNames.add(val.ident.string);
				variableTypes.add(val.type);
			}			
			str="(assert "+expTrans.generateQuantifiedSMTLibExpression(cond, variableNames,variableTypes)+")";						
		}	
		return str+"\n";
	}
	
	private String process(FunctionDefinition decl)
	{
		String str="(define-fun "+decl.ident.string;
		FunctionSymbol funSym=((FunctionSymbol)decl.ident.getSymbol());
		Stream<ValueSymbol> s=Arrays.stream(funSym.params);
		String params=s.map(sym -> "("+ sym.ident.string+" "+this.translateType(sym.type)+")").collect(Collectors.joining(" "));
		str+=" ("+params+")";
		String resultType=this.translateType(funSym.type);
		str+=resultType+" ";	
		Type ty=((FunctionSymbol)decl.ident.getSymbol()).type;
		String expStr=expTrans.compute(decl.exp,ty);
		str+=expStr+")";
		str+="\n";
		return str;
	}
	
	private String process(ValueDefinition decl)
	{
		values.add((ValueSymbol) decl.ident.getSymbol());
		Type ty=((ValueSymbol)decl.ident.getSymbol()).type;
		String str=expTrans.compute(decl.exp,ty);
		return "(define-fun "+decl.ident.string+" () "+this.translateType(ty)+" "+str+")\n";
	}
	
	private String process(TheoremDefinition decl)
	{
		addDeclarationRemovable(";"+decl.ident.string);//this shall only increase readability of the resulting translation.
		String str=expTrans.compute(decl.exp,true);
		return "(assert "+str+")\n";
	}	
}
//---------------------------------------------------------------------------
// end of file
//---------------------------------------------------------------------------
