// ---------------------------------------------------------------------------
// ExpressionPreprocessor.java
// A class which is dedicated to translate expressions to an intermediate format.
//
// Author: Franz Reichl <franz.x.reichl@gmail.com>
// $Id: ExpressionPreprocessor.java,v 1.15 2020/11/21 12:00:15 schreine Exp $
// 
// Author: Franz Reichl <franz.x.reichl@gmail.com>
// Copyright (C) 2019-, Research Institute for Symbolic Computation (RISC)
// Johannes Kepler University, Linz, Austria, http://www.risc.jku.at
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program.  If not, see <http://www.gnu.org/licenses/>.
// ---------------------------------------------------------------------------
package riscal.smt;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.List;
import java.util.ListIterator;
import java.util.Map;
import java.util.Set;
import java.util.Stack;
import java.util.stream.Collectors;

import riscal.smt.DeclarationPreprocessor.FactorialInformation;
import riscal.smt.Utilities.Triple;
import riscal.syntax.AST;
import riscal.syntax.AST.Binder;
import riscal.syntax.AST.Decimal;
import riscal.syntax.AST.Declaration;
import riscal.syntax.AST.Declaration.FunctionDeclaration;
import riscal.syntax.AST.Declaration.TheoremDefinition;
import riscal.syntax.AST.Expression;
import riscal.syntax.AST.Expression.AndExp;
import riscal.syntax.AST.Expression.ApplicationExp;
import riscal.syntax.AST.Expression.ArrayBuilderExp;
import riscal.syntax.AST.Expression.AssertExp;
import riscal.syntax.AST.Expression.CheckExp;
import riscal.syntax.AST.Expression.ChooseExp;
import riscal.syntax.AST.Expression.ChooseInElseExp;
import riscal.syntax.AST.Expression.ChooseInExp;
import riscal.syntax.AST.Expression.EqualsExp;
import riscal.syntax.AST.Expression.EquivExp;
import riscal.syntax.AST.Expression.ExistsExp;
import riscal.syntax.AST.Expression.FactorialExp;
import riscal.syntax.AST.Expression.FalseExp;
import riscal.syntax.AST.Expression.ForallExp;
import riscal.syntax.AST.Expression.GreaterEqualExp;
import riscal.syntax.AST.Expression.GreaterExp;
import riscal.syntax.AST.Expression.HeightExp;
import riscal.syntax.AST.Expression.IdentifierExp;
import riscal.syntax.AST.Expression.IfThenElseExp;
import riscal.syntax.AST.Expression.ImpliesExp;
import riscal.syntax.AST.Expression.InSetExp;
import riscal.syntax.AST.Expression.LessEqualExp;
import riscal.syntax.AST.Expression.LessExp;
import riscal.syntax.AST.Expression.LetExp;
import riscal.syntax.AST.Expression.LetParExp;
import riscal.syntax.AST.Expression.MapBuilderExp;
import riscal.syntax.AST.Expression.MatchExp;
import riscal.syntax.AST.Expression.MaxExp;
import riscal.syntax.AST.Expression.MinExp;
import riscal.syntax.AST.Expression.NotEqualsExp;
import riscal.syntax.AST.Expression.NotExp;
import riscal.syntax.AST.Expression.NotInSetExp;
import riscal.syntax.AST.Expression.NotSubsetExp;
import riscal.syntax.AST.Expression.NumberExp;
import riscal.syntax.AST.Expression.OrExp;
import riscal.syntax.AST.Expression.PlusExp;
import riscal.syntax.AST.Expression.PlusExpMult;
import riscal.syntax.AST.Expression.PowerExp;
import riscal.syntax.AST.Expression.PrintExp;
import riscal.syntax.AST.Expression.PrintInExp;
import riscal.syntax.AST.Expression.PrintTraceInExp;
import riscal.syntax.AST.Expression.ProductExp;
import riscal.syntax.AST.Expression.RecApplicationExp;
import riscal.syntax.AST.Expression.RecIdentifierExp;
import riscal.syntax.AST.Expression.RecordExp;
import riscal.syntax.AST.Expression.RecordSelectionExp;
import riscal.syntax.AST.Expression.RecordUpdateExp;
import riscal.syntax.AST.Expression.SetBuilderExp;
import riscal.syntax.AST.Expression.SubsetExp;
import riscal.syntax.AST.Expression.SumExp;
import riscal.syntax.AST.Expression.TimesExp;
import riscal.syntax.AST.Expression.TimesExpMult;
import riscal.syntax.AST.Expression.TrueExp;
import riscal.syntax.AST.Expression.TupleExp;
import riscal.syntax.AST.Expression.TupleSelectionExp;
import riscal.syntax.AST.Expression.TupleUpdateExp;
import riscal.syntax.AST.Identifier;
import riscal.syntax.AST.Multiple;
import riscal.syntax.AST.Parameter;
import riscal.syntax.AST.QuantifiedVariable;
import riscal.syntax.AST.QuantifiedVariableCore;
import riscal.syntax.AST.QuantifiedVariableCore.IdentifierSetQuantifiedVar;
import riscal.syntax.AST.QuantifiedVariableCore.IdentifierTypeQuantifiedVar;
import riscal.syntax.AST.Type;
import riscal.syntax.AST.Type.BoolType;
import riscal.syntax.AST.Type.IdentifierType;
import riscal.syntax.AST.Type.IntType;
import riscal.syntax.AST.Type.RecordType;
import riscal.syntax.AST.Type.SetType;
import riscal.syntax.AST.Type.TupleType;
import riscal.syntax.ASTCloner;
import riscal.types.Environment.Symbol;
import riscal.types.Environment.Symbol.FunctionSymbol;
import riscal.types.Environment.Symbol.TypeSymbol;
import riscal.types.Environment.Symbol.ValueSymbol;


public class ExpressionPreprocessor extends ASTCloner
{	
	/**
	 * The underlying DeclarationPreprocessor.
	 */
	DeclarationPreprocessor decPre;
	
	/**
	 * This set corresponds to the ValueSymbols of the constants which can occur in the expressions.
	 */
	private Set<ValueSymbol> values;

	/**
	 * Contains the already encountered variables (introduced by quantifiers or lets) 
	 * in the expression that shall be processed.
	 */
	private Set<String> visitedQuantifiedVars=new HashSet<String>();
	
	/**
	 * The existentially quantified variables from the current context.
	 */
	private Stack<ValueSymbol> existentiallyQuantifiedVariables=new Stack<ValueSymbol>();
	
	/***************************************************************************
	 * Getter for visitedQuantifiedVars.
	 * @return The set of the already encountered variables.
	 **************************************************************************/
	Set<String> getVariables()
	{
		return visitedQuantifiedVars;
	}
	
	
	 /**
	 * This Map is used for renaming values. If we rename a function we assign to the
	 * associated value symbol a string which represents its new name.
	 */
	private Map<ValueSymbol,String> variableRenamings=new HashMap<ValueSymbol,String>();
	
	/**
	 * This list contains all additional declarations which are generated 
	 * during the procession of an expression.
	 */
	private List<Declaration> generatedDeclarations=new ArrayList<Declaration>();
	/***************************************************************************
	 * Getter for generatedDeclarations.
	 * @return The list of additionally generated declarations.
	 **************************************************************************/
	List<Declaration> getGeneratedDeclarations()
	{
		return generatedDeclarations;
	}
	
	/**
	 * Contains information about the occurrence of factorial expressions.
	 */
	private FactorialInformation factorialInfo;
	/***************************************************************************
	 * Getter for factorialInfo.
	 * @return The current factorial information.
	 **************************************************************************/
	public FactorialInformation getFactorialInfo()
	{
		return factorialInfo;
	}
	
	/**
	 * Auxiliary cloner.
	 */
	private static ASTCloner cloner=new ASTCloner();
	
	
	/***************************************************************************
	 * Create ExpressionPreprocessor.
	 * @param d The underlying DeclarationPreprocessor
	 * @param vals The ExpressionPreprocessor can only process expressions that 
	 * only contain constants whose valuessymbols are part of this list.
	 **************************************************************************/
	ExpressionPreprocessor(DeclarationPreprocessor d,Set<ValueSymbol> vals)
	{
		this.decPre=d;
		this.values=vals;
		String factorialName=decPre.giveNewName("factorial");
		factorialInfo=new FactorialInformation(factorialName);
	}
	
	/***************************************************************************
	 * Preprocess the given expression. If additional declarations 
	 * were generated during this process they can be retrieved by 
	 * getGeneratedDeclarations as long as no further expression is processed.
	 * Conditions:
	 *  - exp does not contain AssertExp
	 *  - exp does not contain expressions related to recursive types (e.g. MatchExp)
	 * @param exp The expression that shall be preprocessed.
	 * @return A copied and modified version of exp.
	 **************************************************************************/
	public Expression processExpression(Expression exp)
	{
		generatedDeclarations.clear();
		visitedQuantifiedVars.clear();
		Expression exp0=(Expression) exp.accept(this);
		variableRenamings.clear();
		return exp0;
	}	
	
	 /***************************************************************************
	 * Quantify the given expression with respect to vars 
	 * (variable, type pairs according to the parameters).
	 * If toProve is
	 * true, then the quantified expression is negated. 
	 * Afterwards this expression gets preprocessed.
	 * * Conditions:
	 *  - exp does not contain AssertExp
	 *  - exp does not contain expressions related to recursive types (e.g. MatchExp)
	 * @param exp The expression that shall be quantified and preprocessed.
	 * @param vars Contains the variable names/types for the quantification.
	 * @param toProve If true the quantified expression is negated.
	 * @return The preprocessed quantified expression.
	 **************************************************************************/
	public Expression processAndQuantifyExpression(Expression exp,Parameter[] vars,boolean toProve)
	{
		generatedDeclarations.clear();
		visitedQuantifiedVars.clear();
		Expression exp0=constructForAllExpression(exp, vars);
		if(toProve&&!SMT.debug_DoNotNegateGoal)
			exp0=new NotExp(exp0);
		exp0=(Expression) exp0.accept(this);		
		variableRenamings.clear();	
		return exp0;
	}
	
	/***************************************************************************
	 * Preprocess the given type.
	 * @param type The type that shall be preprocessed.
	 * @return The preprocessed type.
	 **************************************************************************/
	public Type translateType(Type type)
	{
		if(type.isCanonical())
			return Utilities.canonic2syntacticType(type);
		return processType(type);
	}
	
	/***************************************************************************
	 * Preprocess the given type.
	 * Remark we also process types in the "Expression"-Preprocessor 
	 * since it is not worth it to create a own class for it.
	 * @param type The type that shall be preprocessed.
	 * @return The preprocessed type.
	 **************************************************************************/
	public Type processType(Type type)
	{
		return (Type) type.accept(this);
	}
	
	
	/***************************************************************************
	 * This method constructs forall expressions. 
	 * The given expression is not copied but the variables 
	 * and types in the quantifier are.
	 * @param exp An expression that shall be universally quantified.
	 * @param varTypes The types over which the expression shall be quantified.
	 * @param varNames The names of the variables which shall be quantified. 
	 * varTypes.size()==varNames.size() must hold.
	 * @return If the array is not empty and the given expression
	 * contains a free variable that is part of vars a ForallExp.
	 * Otherwise exp.
	 **************************************************************************/
	private Expression constructForAllExpression(Expression exp,Parameter[] vars)
	{
		if(vars.length==0)
			return exp;
		Set<ValueSymbol> free=TypedFreeVariables.compute(exp);
		for(int i=vars.length-1;i>=0;i--)
		{
			String name=vars[i].ident.string;
			Symbol sym=vars[i].ident.getSymbol();
			if(free.contains(sym))
			{
				Identifier ident=new Identifier(name);
				ident.setSymbol(sym);
//				Type t=translateType(vars[i].type);
				Type t=vars[i].type;
				List<QuantifiedVariableCore> qvcore0=new ArrayList<AST.QuantifiedVariableCore>(1);
				qvcore0.add(new QuantifiedVariableCore.IdentifierTypeQuantifiedVar(ident,t));
				QuantifiedVariable qvar=new QuantifiedVariable(qvcore0, null);
				exp=new ForallExp(qvar, exp);
			}			
		}
		return exp;
	}
	
	/***************************************************************************
	 * Generate a name for a variable. This new name is not part of
	 * visitedQuantifiedVars and identString. This name is added to
	 * visitedQuantifiedVars.
	 * @param suggestion A bases for the string that shall be generated.
	 * @return A new variable name.
	 **************************************************************************/
	private String giveNewVariable(String suggestion)
	{
		String variable=decPre.giveNewNameConstraint(suggestion, visitedQuantifiedVars);
		return variable;
	}
	

	//#############################Cloner################################################//
	
	
	/***************************************************************************
	 * Order qvcore with respect to SMT.qvOrdering.
	 * NoOrdering
	 * 		The ordering of qvcore is preserved
	 * OrderAscending_TypeSize
	 * 		Elements with smaller types come first.
	 * 		If two elements have the same type size the element
	 * 		whose name is lexicographically smaller comes first.
	 * OrderDescending_TypeSize
	 * 		Elements with larger types come first.
	 * 		If two elements have the same type size the element
	 * 		whose name is lexicographically smaller comes first.
	 * @param qvcore The array that shall be sorted.
	 * @return A sorted list containing the elements of qvcore.
	 **************************************************************************/
	private static List<QuantifiedVariableCore> orderQV(QuantifiedVariableCore[] qvcore)
	{
		List<QuantifiedVariableCore> qvclst=new ArrayList<AST.QuantifiedVariableCore>(Arrays.asList(qvcore));
		switch(SMT.qvOrdering)
		{
		case NoOrdering:
			break;
		case OrderAscending_TypeSize:
			qvclst.sort(new Utilities.QVAscendingComparator());
			break;
		case OrderDescending_TypeSize:
			qvclst.sort(new Utilities.QVDescendingComparator());
			break;
		}
		return qvclst;
	}
	
	/***************************************************************************
	 * Extracts and processes the QuantifiedVariableCore from the given QuantifiedVariable.
	 * IdentifierSetQuantifiedVar are transformed into IdentifierTypeQuantifiedVar.
	 * Moreover all conditions part of qVar and the QuantifiedVariableCore are
	 * extracted, this includes conditions generated by the transformation of 
	 * IdentifierSetQuantifiedVar. 
	 * Note: The expressions representing the conditions are not yet
	 * processed.
	 * @param qVar The QuantifiedVariable that shall be processed.
	 * @return A pair. The first element contains a list of all withExpressions 
	 * part of qVar in the order of occurrence.
	 * If qVar itself has a with expression this expression is added with a 
	 * conjunction to the last condition.
	 * The second element contains a list of the processed quantifiedvariablecore from qVar.
	 **************************************************************************/
	private Utilities.Pair<List<Expression>, List<QuantifiedVariableCore>> traverseQuantifiedVariable(QuantifiedVariable qVar)
	{
		List<QuantifiedVariableCore> qvcorelst=new ArrayList<AST.QuantifiedVariableCore>(qVar.qvcore.length);
		List<Expression> premises=new ArrayList<AST.Expression>(qVar.qvcore.length);			
		Set<Utilities.Pair<Expression, Set<ValueSymbol>>> condsWithFreeVars=new HashSet<Utilities.Pair<Expression,Set<ValueSymbol>>>();
		Set<ValueSymbol> currentBound=new HashSet<ValueSymbol>();
		for(QuantifiedVariableCore qvcore:qVar.qvcore)
		{
			currentBound.add((ValueSymbol) qvcore.ident.getSymbol());
			Expression wexp=qvcore.wexp;
			
			Set<ValueSymbol> free=new HashSet<ValueSymbol>();
			if(wexp!=null)
			{
				free.addAll(TypedFreeVariables.compute(wexp));
			}
			if(qvcore instanceof IdentifierSetQuantifiedVar)
			{
				Expression set=((IdentifierSetQuantifiedVar)qvcore).exp;
				Identifier setIdent=new Identifier(qvcore.ident.string);
				setIdent.setSymbol(qvcore.ident.getSymbol());
				wexp=AST.and(wexp, new InSetExp(new IdentifierExp(setIdent), set));
				free.addAll(TypedFreeVariables.compute(set));
				free.add((ValueSymbol) qvcore.ident.getSymbol());
			}
			if(wexp!=null)
			{
				free.retainAll(currentBound);
				condsWithFreeVars.add(new Utilities.Pair<AST.Expression, Set<ValueSymbol>>(wexp, free));
			}
		}
		List<QuantifiedVariableCore> qvclst=orderQV(qVar.qvcore);
		currentBound.clear();
		for(QuantifiedVariableCore qvcore:qvclst)
		{
			qvcorelst.add((QuantifiedVariableCore) qvcore.accept(this));
			currentBound.add((ValueSymbol) qvcore.ident.getSymbol());
			Expression expPremise=null;
			for(Iterator<Utilities.Pair<Expression, Set<ValueSymbol>>>it=condsWithFreeVars.iterator();it.hasNext();)
			{
				Utilities.Pair<Expression, Set<ValueSymbol>> p=it.next();
				if(currentBound.containsAll(p.second))
				{
					expPremise=AST.and(expPremise, p.first);
					it.remove();
				}
			}			
			premises.add(expPremise);
		}	
		if(qVar.exp!=null)
		{
			Expression e=qVar.exp;
			premises.set(qVar.qvcore.length-1, AST.and(e, premises.get(qVar.qvcore.length-1)));
		}			
		return new Utilities.Pair<List<Expression>, List<QuantifiedVariableCore>>(premises, qvcorelst);
	}
	
	private QuantifiedVariable processQuantifiedVariable(QuantifiedVariable qVar)
	{
		Utilities.Pair<List<Expression>, List<QuantifiedVariableCore>> conds=traverseQuantifiedVariable(qVar);
		List<QuantifiedVariableCore> corelst=new ArrayList<AST.QuantifiedVariableCore>();
		Iterator<Expression> itConds=conds.first.iterator();
  	Iterator<QuantifiedVariableCore> itQVC=conds.second.iterator();
  	while(itConds.hasNext()&&itQVC.hasNext())
  	{
  		Expression ex=itConds.next();
  		Expression qVarCondition=ex==null?null:(Expression) ex.accept(this);
  		IdentifierTypeQuantifiedVar aux=(IdentifierTypeQuantifiedVar) itQVC.next();
  		QuantifiedVariableCore qvc=new IdentifierTypeQuantifiedVar(aux.ident, aux.type , qVarCondition);
  		corelst.add(qvc);
  	}
  	return new QuantifiedVariable(corelst, null);
	}
	

	/***************************************************************************
	 * Used to incorporate the condition of an universal quantifier
	 * into the corresponding quantified expression.
	 * @param condition A not yet processed expression.
	 * @param exp An already processed expression.
	 * @return A processed expression.
	 **************************************************************************/
	private Expression incorporateForallCondition(Expression condition,Expression exp)
	{
		if(condition==null)
			return exp;
		if(SMT.getTransformToNegationNormalForm())
		{
			Expression exp0=new NotExp(condition);
			exp0=(Expression) exp0.accept(this);
			return new OrExp(exp0, exp);
		}
		else
		{
			Expression exp0=(Expression) condition.accept(this);	
			return Utilities.constructImplicationExpression(exp0, exp);
		}
	}
	
	/***************************************************************************
	 * Universally quantifies the given expression. The universal quantifiers
	 * are applied in a nested way such that each QuantifiedVariable contains
	 * only one QuantifiedVariableCore.
	 * @param qvar The QuantifiedVariable that shall be used for the
	 * quantification.
	 * @param exp The expression that shall be quantified.
	 * @return The quantified and processed expression.
	 **************************************************************************/
	private Expression universallyQuantifyExpression(QuantifiedVariable qvar,Expression exp)
	{
		Utilities.Pair<List<Expression>, List<QuantifiedVariableCore>> conds=traverseQuantifiedVariable(qvar);
		Expression exp0=(Expression) exp.accept(this);
		for(int i=conds.second.size()-1;i>=0;i--)
		{
			List<QuantifiedVariableCore> corelist=new ArrayList<QuantifiedVariableCore>(1);
			corelist.add(conds.second.get(i));
			QuantifiedVariable qv=new QuantifiedVariable(corelist , null);
			exp0=incorporateForallCondition(conds.first.get(i), exp0);
			exp0=new ForallExp(qv, exp0);
		}
		return exp0;
	}
	
	/***************************************************************************
	 * Used to incorporate the condition of an existential quantifier
	 * into the corresponding quantified expression.
	 * @param condition A not yet processed expression.
	 * @param exp An already processed expression.
	 * @return A processed expression.
	 **************************************************************************/
	private Expression incorporateExistsCondition(Expression condition,Expression exp)
	{
		if(condition==null)
			return exp;
		Expression exp0=(Expression) condition.accept(this);
		return AST.and(exp0, exp);
	}
	
	/***************************************************************************
	 * Existentially quantifies the given expression. The existential quantifiers
	 * are applied in a nested way such that each QuantifiedVariable contains
	 * only one QuantifiedVariableCore.
	 * @param qvar The QuantifiedVariable that shall be used for the
	 * quantification.
	 * @param exp The expression that shall be quantified.
	 * @return The quantified and processed expression.
	 **************************************************************************/
	private Expression existentiallyQuantifyExpression(QuantifiedVariable qvar,Expression exp)
	{
		Utilities.Pair<List<Expression>, List<QuantifiedVariableCore>> conds=traverseQuantifiedVariable(qvar);
		Expression exp0=(Expression) exp.accept(this);
		for(int i=conds.second.size()-1;i>=0;i--)
		{
			List<QuantifiedVariableCore> corelist=new ArrayList<QuantifiedVariableCore>(1);
			corelist.add(conds.second.get(i));
			QuantifiedVariable qv=new QuantifiedVariable(corelist , null);
			exp0=incorporateExistsCondition(conds.first.get(i), exp0);
			exp0=new ExistsExp(qv, exp0);
		}
		return exp0;
	}
		
  public Expression visit(AndExp exp)
  {
  	if(exp.exp1 instanceof FalseExp||exp.exp2 instanceof FalseExp)
  		return new FalseExp();
  	if(exp.exp1 instanceof TrueExp)
  	{
  		return (Expression) exp.exp2.accept(this);
  	}
  	if(exp.exp2 instanceof TrueExp)
  	{
  		return (Expression) exp.exp1.accept(this);
  	}
  	if(SMT.considerGuardsForChoices())
  		decPre.conds.analyseWithSimpleConditions(exp.exp2);
    Expression exp10 = (Expression)exp.exp1.accept(this);
    if(SMT.considerGuardsForChoices())
    {
    	decPre.conds.revertSimpleConditions();
    	decPre.conds.analyseWithSimpleConditions(exp.exp1);
    }
    Expression exp20 = (Expression)exp.exp2.accept(this);
    if(SMT.considerGuardsForChoices())
    	decPre.conds.revertSimpleConditions();
    return new AndExp(exp10, exp20);
  }

	public Expression visit(ApplicationExp exp)
  {
    List<Expression> exps0 = new ArrayList<Expression>(exp.exps.length);
    for (Expression e : exp.exps) exps0.add((Expression)e.accept(this));
    String funName=decPre.getIdentRenaming((FunctionSymbol) exp.ident.getSymbol());
    funName=funName==null?exp.ident.string:funName;
    Identifier ident0=new Identifier(funName);
    return new ApplicationExp(ident0, exps0);
  }
	
	public Expression visit(ArrayBuilderExp exp)
  {
		Long v=Utilities.getIntValueFromExpression(exp.exp1);
		if(v==null)
		{
			SMT.illegalUsage(exp.exp1, "The expression is not an int");
			return null;
		}			
		long l=v.longValue();
//		if(l>Integer.MAX_VALUE)
//		{
//			SMT.tooLargeType("An array with more then Integer.MAX_VALUE elements is too large!", l, exp.getType());
//		}
		IntType domainType=new IntType(AST.number(0), AST.number(l-1));
		Type imageType = exp.type == null ? null : (Type)exp.type.accept(this);
		Expression e=(Expression) exp.exp2.accept(this);
		return new MapBuilderExp(domainType, imageType, e);
  }
	
	public Expression visit(AssertExp exp)
	{
		SMT.notImplemented(exp, "Assert Expressions are not yet implemented");
		return null;
	}
	
	/***************************************************************************
	 * Determine if the given QuantifiedVariable or the QuantifiedVariableCore
	 * it contains contain any conditions
	 * @param qv The QuantifiedVariable that shall be analysed.
	 * @return True if the given QuantifiedVariable contains any conditions.
	 **************************************************************************/
  private boolean quantifiedVariableHasConditions(QuantifiedVariable qv)
  {
  	if(qv.exp!=null)
  		return true;
  	for(QuantifiedVariableCore qvcore:qv.qvcore)
  	{
  		if(qvcore.wexp!=null)
  			return true;
  		if(qvcore instanceof IdentifierSetQuantifiedVar)
  			return true;
  	}
  	return false;
  }
  
  /*
	 * The way we deal with the choice expressions corresponds to the
	 * the already existing undeterministic handling of choice expression
	 * We are checking if there is a choice such that some negated conclusion
	 * is true. If we can find such a choice we found a choice that
	 * violates the assertion we want to check. If we cannot find such
	 * a choice every choice satisfies the assertion.
	 */
	
	/*
	 * We assume that there is a value that can be chosen.
	 * If this is not the case the translated might give false results.
	 */
  
  /*
   * Remark on the handling of choose expressions
   * 
   * Choose expression are translated to uninterpreted functions that ensure the respective properties.
   * Since the occurring conditions respective the expressions in ChhoseInExp and ChooseInElseExp may contains
   * variables (introduced by quantifiers, lets, functions definitions) these functions must depend on these variables.
   * 
   * We could plug in the values of the variables introduced by lets in the respective expressions. While in several
   * cases this would remove the number of variables (e.g. let x=2 in choose y:Int[0,5] with value<=x). But in other cases
   * it would increase the number of variables (e.g. fun(x:Int[0,5],y:Int[0,5]):Int[0,5]=... let z=x+y in choose y:Int[0,5] with value<=z)
   * Therefore it would be necessary to analyze which lets should be replaced. This would be a possible improvment of this program.
   * Here we will not make such an analysis and therefore the variables which are introduced by lets are also considered for
   * the parameters of the choose functions.
   * 
   */
  
  private Triple<List<ValueSymbol>,QuantifiedVariable,Expression> generateChooseFunction(Expression exp,String funNameSuggestion,boolean considerAsDefined)
  {
  	String chooseName;
  	if(considerAsDefined)
  		chooseName=decPre.giveFunctionNameAndSetAsDefined(funNameSuggestion);
  	else
  		chooseName=decPre.giveNewName(funNameSuggestion);
  	List<ValueSymbol>freeVariables=Utilities.getTypedFreeVariablesIgnoreValues(exp, values,existentiallyQuantifiedVariables);
  	return generateChooseLikeFunction(freeVariables,exp,chooseName);
  }
  
  private Triple<List<ValueSymbol>,QuantifiedVariable,Expression> generateMinMaxFunction(Expression exp,String funNameSuggestion)
  {
  	String chooseName=decPre.giveFunctionNameAndSetAsDefined(funNameSuggestion);
  	List<ValueSymbol>freeVariables=Utilities.getTypedFreeVariablesIgnoreValues(exp, values);
  	return generateChooseLikeFunction(freeVariables,exp,chooseName);
  }
  
  private Triple<List<ValueSymbol>,QuantifiedVariable,Expression> generateChooseLikeFunction(List<ValueSymbol>freeVariables,Expression exp,String chooseName)
  {
  	decPre.setUninterpretedFunctionsRequired(true);
		if(freeVariables.size()>0)
			decPre.setNonNullaryUninterpretedFunctionsRequired(true);
  	List<Expression> args=new ArrayList<AST.Expression>(freeVariables.size());
  	List<Parameter> params=new ArrayList<Parameter>(freeVariables.size());	
  	List<QuantifiedVariableCore> qvcoreList=new ArrayList<AST.QuantifiedVariableCore>(freeVariables.size());  	
  	Set<String> names=new HashSet<String>();
  	for(ValueSymbol val:freeVariables)
  	{
  		String name=variableRenamings.get(val);
  		if(name==null)
  		{
  			name=val.ident.string;
  		} 		
  		args.add(new IdentifierExp(new Identifier(name)));
  		//I deliberately generate two times the syntactic types. I do not want that the same instance of the object
  		//is used twice. This may be unnecessary, nevertheless I do it to prevent possible problems
  		qvcoreList.add(new IdentifierTypeQuantifiedVar(new Identifier(name), Utilities.canonic2syntacticType(val.type)));
  		
  		if(names.contains(name))
  		{
  			name=Utilities.NewIdentString(name, names);
  		}
  		names.add(name);
  		
  		params.add(new Parameter(new Identifier(name), Utilities.canonic2syntacticType(val.type)));
  	}
  	Declaration dec=new FunctionDeclaration(new Multiple(false), new Identifier(chooseName), params, Utilities.canonic2syntacticType(exp.getType()));
  	generatedDeclarations.add(dec);
  	QuantifiedVariable quant=new QuantifiedVariable(qvcoreList, null);
  	Expression application=new ApplicationExp(new Identifier(chooseName), args);
  	return new Triple<List<ValueSymbol>,QuantifiedVariable, Expression>(freeVariables,quant, application);
  }
  
  
  private Expression getChooseTheoremExpression(QuantifiedVariable qv,Expression application,Expression expFromChoose)
  {
  	Expression exp0=new EqualsExp(application, expFromChoose);
  	return getChooseTheoremExpression(qv,exp0);
  }
  
  private Expression getChooseTheoremExpression(QuantifiedVariable qv,Expression theoremExpression)
  {
  	Utilities.Pair<List<Expression>, List<QuantifiedVariableCore>> conds=traverseQuantifiedVariable(qv);
  	Expression exp0=theoremExpression;
  	exp0=(Expression) exp0.accept(this);
  	for(int i=qv.qvcore.length-1;i>=0;i--)
  	{
  		List<QuantifiedVariableCore> corelist=new ArrayList<QuantifiedVariableCore>(1);
			corelist.add(conds.second.get(i));
			QuantifiedVariable quant=new QuantifiedVariable(corelist , null);
			exp0=incorporateExistsCondition(conds.first.get(i), exp0);
			exp0=new ExistsExp(quant, exp0);
  	}
  	return exp0;
  }
  
  private Expression getGuardedChooseTheoremExpression(QuantifiedVariable qv,Expression application,Expression expFromChoose,Expression elseExp)
  {
  	Expression guard=getChooseTheoremExpression(qv,new TrueExp());
  	Expression exp0=getChooseTheoremExpression(qv,application,expFromChoose);
  	if(elseExp==null) 		
    	return new ImpliesExp(guard, exp0);
//  		return new IfThenElseExp(guard, exp0, new TrueExp());
  	else
  		return new IfThenElseExp(guard, exp0, elseExp);
  }
  
  private Expression quantifyChooseExp(Expression exp,QuantifiedVariable qv,List<Expression> conditions)
  {
  	Expression exp0=exp;
  	if(conditions.size()!=0)
  	{
  		ListIterator<Expression> it=conditions.listIterator();
    	Expression cond=it.next();
    	while(it.hasNext())
    	{
    		cond=AST.and(cond, it.next());
    	}
    	cond=new NotExp(cond);
    	cond=(Expression) cond.accept(this);
    	exp0=new OrExp(cond, exp0);
  	}
  	return Utilities.constructNestedForallExpression(qv, exp0);
  }
  
  
  public Expression visit(ChooseExp exp)
  {
  	boolean hasConditions=quantifiedVariableHasConditions(exp.qvar);
  	Triple<List<ValueSymbol>,QuantifiedVariable, Expression> p=generateChooseFunction(exp,"choose",hasConditions);
  	List<ValueSymbol> free=p.first;
  	QuantifiedVariable quant=p.second;
  	Expression application=p.third;
  	if(hasConditions)
  	{
	  	Expression chooseExp;
	  	if(exp.qvar.qvcore.length==1)
	  	{
	  		String variable=exp.qvar.qvcore[0].ident.string;
	  		Identifier ident=new Identifier(variable);
	  		ident.setSymbol(exp.qvar.qvcore[0].ident.getSymbol());
	  		chooseExp=new IdentifierExp(ident);
	  	}
	  	else
	  	{
	  		List<Expression> tupleExpressions=new ArrayList<AST.Expression>(quant.qvcore.length);
	  		for(QuantifiedVariableCore qvc:exp.qvar.qvcore)
	  		{
	  			Identifier ident=new Identifier(qvc.ident.string);
	  			ident.setSymbol(qvc.ident.getSymbol());
	  			tupleExpressions.add(new IdentifierExp(ident));
	  		}
	  		chooseExp=new TupleExp(tupleExpressions);
	  	}
	  	boolean guard=false;
	  	List<Expression> conditionsToConsider=new ArrayList<AST.Expression>();
	  	if(SMT.considerGuardsForChoices())
	  	{	 
	  		for(ValueSymbol sym:free)
		  	{
		  		if(SMT.guardChoices==SMT.UsageModeGuardsForChoices.UseGuards&&decPre.conds.hashCondition(sym))
		  		{
		  			guard=true;
		  			conditionsToConsider.clear();
		  			break;
		  		}
		  		List<Expression> c=decPre.conds.getSimpleCondition(sym);
		  		if(c!=null)
		  			conditionsToConsider.addAll(c);
		  	}
	  	}
	  	Expression theoremExp;
	  	if(SMT.considerGuardsForChoices())
	  		decPre.conds.analyse(exp.qvar);
	  	if(!guard)
	  		theoremExp=getChooseTheoremExpression(exp.qvar, application, chooseExp);
	  	else
	  		theoremExp=getGuardedChooseTheoremExpression(exp.qvar, application, chooseExp, null);  	
	  	if(free.size()>0)
	  		theoremExp=quantifyChooseExp(theoremExp, quant, conditionsToConsider);
	  	if(SMT.considerGuardsForChoices())
	    	decPre.conds.revertVariablesWithConditions();
	  	String chooseTh1Name=decPre.giveNewName("chooseTheorem");
    	Identifier chooseThIdent=new Identifier(chooseTh1Name);
    	Declaration chooseTheorem=new TheoremDefinition(chooseThIdent, theoremExp, true);
			generatedDeclarations.add(chooseTheorem);
  	}
  	return (Expression) application.accept(cloner);
  }  
  
 

  public Expression visit(ChooseInElseExp exp)
  {
  	Triple<List<ValueSymbol>,QuantifiedVariable, Expression> p=generateChooseFunction(exp,"choose",true);
  	List<ValueSymbol> free=p.first;
  	QuantifiedVariable quant=p.second;
  	Expression application=p.third;
//  	Expression exp1=(Expression) exp.exp1.accept(this);
  	Expression exp1=exp.exp1;
  	Expression exp2=(Expression) exp.exp2.accept(this);
//  	exp1=new EqualsExp(exp1, application);
  	exp2=new EqualsExp(exp2, (Expression) application.accept(cloner));
  	boolean hasConditions=quantifiedVariableHasConditions(exp.qvar);
  	Expression theoremExp;
  	if(SMT.considerGuardsForChoices())
  		decPre.conds.analyse(exp.qvar);
  	if(hasConditions)
  	{
  		theoremExp=getGuardedChooseTheoremExpression(exp.qvar, application, exp1, exp2);
  	}
  	else
  	{
  		theoremExp=getChooseTheoremExpression(exp.qvar, application, exp1);
  	}
  	if(free.size()>0)
  		theoremExp=Utilities.constructNestedForallExpression(quant, theoremExp);
  	if(SMT.considerGuardsForChoices())
    	decPre.conds.revertVariablesWithConditions();
		String chooseTh1Name=decPre.giveNewName("chooseTheorem");
  	Identifier chooseThIdent=new Identifier(chooseTh1Name);
  	Declaration chooseTheorem=new TheoremDefinition(chooseThIdent, theoremExp, true);
		generatedDeclarations.add(chooseTheorem);
  	return (Expression) application.accept(cloner);
  }
  

  public Expression visit(ChooseInExp exp)
  {
  	Triple<List<ValueSymbol>,QuantifiedVariable, Expression> p=generateChooseFunction(exp,"choose",true);
  	List<ValueSymbol> free=p.first;
  	QuantifiedVariable quant=p.second;
  	Expression application=p.third;
//  	Expression exp1=(Expression) exp.exp.accept(this);
  	Expression exp1=exp.exp;
  	boolean guard=false;
  	boolean hasConditions=quantifiedVariableHasConditions(exp.qvar);
  	List<Expression> conditionsToConsider=new ArrayList<AST.Expression>();
  	if(hasConditions&&SMT.considerGuardsForChoices())
  	{	  		
  		for(ValueSymbol sym:free)
	  	{
	  		if(SMT.guardChoices==SMT.UsageModeGuardsForChoices.UseGuards&&decPre.conds.hashCondition(sym))
	  		{
	  			guard=true;
	  			conditionsToConsider.clear();
	  			break;
	  		}
	  		List<Expression> c=decPre.conds.getSimpleCondition(sym);
	  		if(c!=null)
	  			conditionsToConsider.addAll(c);
	  	}
  	}
  	Expression theoremExp;
  	if(SMT.considerGuardsForChoices())
  		decPre.conds.analyse(exp.qvar);
  	if(!guard)
  		theoremExp=getChooseTheoremExpression(exp.qvar, application, exp1);
  	else
  		theoremExp=getGuardedChooseTheoremExpression(exp.qvar, application, exp1, null);  	
  	if(free.size()>0)
  		theoremExp=quantifyChooseExp(theoremExp, quant, conditionsToConsider);
  	if(SMT.considerGuardsForChoices())
    	decPre.conds.revertVariablesWithConditions();
  	String chooseTh1Name=decPre.giveNewName("chooseTheorem");
  	Identifier chooseThIdent=new Identifier(chooseTh1Name);
  	Declaration chooseTheorem=new TheoremDefinition(chooseThIdent, theoremExp, true);
		generatedDeclarations.add(chooseTheorem);
  	return (Expression) application.accept(cloner);
  }    
  
  public Expression visit(EqualsExp exp)
  {
  	if((SMT.getTransformToNegationNormalForm()&&SMT.skolemisationInEquiavalences)&&exp.exp1.getType() instanceof BoolType)
  	{
  		Expression n1=new NotExp(exp.exp1);
  		Expression n2=new NotExp(exp.exp2);
  		Expression o1=new OrExp(n1, exp.exp2);
  		Expression o2=new OrExp(exp.exp1, n2);
  		return (Expression) (new AndExp(o1, o2)).accept(this);
  	}
  	else
  	{
    	Expression exp10 = (Expression)exp.exp1.accept(this);
      Expression exp20 = (Expression)exp.exp2.accept(this);
      return new EqualsExp(exp10, exp20);
  	}
  }
  
  public Expression visit(EquivExp exp)
  {
  	if(SMT.getTransformToNegationNormalForm()&&SMT.skolemisationInEquiavalences)
  	{
  		Expression n1=new NotExp(exp.exp1);
  		Expression n2=new NotExp(exp.exp2);
  		Expression o1=new OrExp(n1, exp.exp2);
  		Expression o2=new OrExp(exp.exp1, n2);
  		return (Expression) (new AndExp(o1, o2)).accept(this);
  	}
  	else
  	{
  		if(SMT.considerGuardsForChoices())
  		{
  			decPre.conds.analyseWithSimpleConditions(exp.exp2);
  		}
      Expression exp10 = (Expression)exp.exp1.accept(this);
      if(SMT.considerGuardsForChoices())
      {
      	decPre.conds.revertSimpleConditions();
      	decPre.conds.analyseWithSimpleConditions(exp.exp1);
      }
      Expression exp20 = (Expression)exp.exp2.accept(this);
      if(SMT.considerGuardsForChoices())
      	decPre.conds.revertSimpleConditions();
      //in SMT-LIB we have to use "=" instead of "<=>"
      return new EqualsExp(exp10, exp20);
  	}
  }

  public Expression visit(ExistsExp exp)
  {
  	if(SMT.considerGuardsForChoices())
  	{
  		decPre.conds.analyse(exp.qvar);
  	}
  	Expression resultExp;
  	if(decPre.processingTheorem())
  	{
  		resultExp= existentiallyQuantifyExpression(exp.qvar, exp.exp);
  	}
  	else
  	{
  		int counter=0;
    	for(QuantifiedVariableCore qvc:exp.qvar.qvcore)
    	{
    		counter++;
    		existentiallyQuantifiedVariables.push((ValueSymbol) qvc.ident.getSymbol());
    	}
    	resultExp=existentiallyQuantifyExpression(exp.qvar, exp.exp);
    	for(int i=0;i<counter;i++)
    		existentiallyQuantifiedVariables.pop();
  	}  	
  	if(SMT.considerGuardsForChoices())
  		decPre.conds.revertVariablesWithConditions();
  	return resultExp;
  }
   

  public Expression visit(ForallExp exp)
  {
  	if(SMT.considerGuardsForChoices())
  		decPre.conds.analyse(exp.qvar);
  	Expression resultExp;
  	if(decPre.processingTheorem())
  	{
  		int counter=0;
    	for(QuantifiedVariableCore qvc:exp.qvar.qvcore)
    	{
    		counter++;
    		existentiallyQuantifiedVariables.push((ValueSymbol) qvc.ident.getSymbol());
    	}
    	resultExp=universallyQuantifyExpression(exp.qvar, exp.exp);
    	for(int i=0;i<counter;i++)
    		existentiallyQuantifiedVariables.pop();
  	}
  	else
  	{
  		resultExp= universallyQuantifyExpression(exp.qvar, exp.exp);
  	}   
  	if(SMT.considerGuardsForChoices())
  		decPre.conds.revertVariablesWithConditions();
  	return resultExp;   
  }
  
  
  public Expression visit(HeightExp exp)
  {
  	SMT.notImplemented(exp, "Height Expressions are not yet implemented");
  	return null;
  } 
  
  public Expression visit(IdentifierExp exp)
  {
  	Identifier ident0;
  	String str=variableRenamings.get(exp.ident.getSymbol());
  	if(str!=null)
  	{
  		ident0=new Identifier(str);
  		return new IdentifierExp(ident0);
  	}
  	str=decPre.getValueRenaming((ValueSymbol) exp.ident.getSymbol());
  	str=str==null?exp.ident.string:str;
  	ident0=new Identifier(str);
    return new IdentifierExp(ident0);
  }
  
  
  /**
   * Remark
   * If transformToNegationNormalForm is true ifthenelse expressions should be removed
   * since they correspond to a conjunction of implications.
   * The problem is we cannot simply transform these expressions to
   * conjunctions of implications as the second and third subexpression
   * need not to be boolean expression. We would need the last predicate
   * in which these expressions occurred to use these formula in the premise of
   * the implication. 
   * This would be inconventient so we do not transform ifthenelse as long as no other way to
   * transform these expressions is found.
   * This means that later we cannot skolemize the first expression.
   */
  public Expression visit(IfThenElseExp exp)
  {
    Expression exp10 = (Expression)exp.exp1.accept(this);
    if(SMT.considerGuardsForChoices())
    {
    	decPre.conds.analyseWithSimpleConditions(exp.exp1);
    }    
    Expression exp20 = (Expression)exp.exp2.accept(this);
    if(SMT.considerGuardsForChoices())
    {
    	decPre.conds.revertSimpleConditions();
      decPre.conds.analyseWithSimpleConditions(new NotExp(exp.exp1));
    }    
    Expression exp30 = (Expression)exp.exp3.accept(this);
    if(SMT.considerGuardsForChoices())
    {
    	decPre.conds.revertSimpleConditions();
    }
    return new IfThenElseExp(exp10, exp20, exp30);
  }

  public Expression visit(ImpliesExp exp)
  {
  	if(SMT.getTransformToNegationNormalForm())
  	{
  		Expression n1=new NotExp(exp.exp1);
  		return (Expression) (new OrExp(n1, exp.exp2)).accept(this);
  	}
  	else
  	{
      Expression exp10 = (Expression)exp.exp1.accept(this);
      if(SMT.considerGuardsForChoices())
      {
      	decPre.conds.analyseWithSimpleConditions(exp.exp1);
      }    
      Expression exp20 = (Expression)exp.exp2.accept(this);
      if(SMT.considerGuardsForChoices())
      {
      	decPre.conds.revertSimpleConditions();
      }
      return new ImpliesExp(exp10, exp20);
  	}
  } 
    
  /*
   * SMT-LIB supports let expressions. But SMT-LIB only supports a parallel let so we have
   * to rewrite the sequential RISCAL let.
   */	  
  public Expression visit(LetExp exp)
  {
  	//RISCAL generates verification conditions with such lets
  	//We want to eliminate these lets.
  	if(exp.exp instanceof NotExp)
  	{
  		NotExp ne=(NotExp)exp.exp;
  		if(ne.exp instanceof IdentifierExp)
  		{
  			Symbol vSym=((IdentifierExp)ne.exp).ident.getSymbol();
  			Binder b=exp.binder[0];
  			if(b.ident.getSymbol()==vSym)
  			{
  				NotExp neg=new NotExp(b.exp);
  				return (Expression) neg.accept(this);
  			} 			
//  			for(Binder b:exp.binder)
//    		{
//    			if(b.ident.getSymbol()==vSym)
//    			{
//    				NotExp neg=new NotExp(b.exp);
//    				return (Expression) neg.accept(this);
//    			}
//    		}
//    		return (Expression) exp.exp.accept(this);
  		}
  	}
  	if(SMT.considerGuardsForChoices())
  	{
  		decPre.conds.analyse(exp.binder);
  	}
  	List<Binder> binderList=new ArrayList<AST.Binder>(exp.binder.length);
  	for(Binder b:exp.binder)
  	{
  		//Necessary  because of boolector. Boolector does not accept something like let (( result result))
  		if(b.exp instanceof IdentifierExp)
  		{
  			IdentifierExp iE=(IdentifierExp)b.exp;
  			if(iE.ident.string.equals(b.ident.string))
  			{
  				String aux=variableRenamings.get((ValueSymbol)iE.ident.getSymbol());
  				if(aux!=null)
  				{
  					variableRenamings.put((ValueSymbol) b.ident.getSymbol(), aux);
  				}
  				continue;
  			}
  		}
  		String name;
  		if(visitedQuantifiedVars.contains(b.ident.string)||decPre.forbiddenName(b.ident.string))
  		{
  			name=giveNewVariable(b.ident.string);
				variableRenamings.put((ValueSymbol) b.ident.getSymbol(), name);
  		}
  		else
  		{
  			name=b.ident.string;
  			visitedQuantifiedVars.add(name);
  		}
  		Identifier ident0=new Identifier(name);
  		Expression bindExp=(Expression) b.exp.accept(this);
  		binderList.add(new Binder(ident0, bindExp));
  	}  	
  	Expression exp0 = (Expression)exp.exp.accept(this);
  	ListIterator<Binder> it=binderList.listIterator(binderList.size());
  	while(it.hasPrevious())
  	{
  		List<Binder> binder0 = new ArrayList<Binder>(1);
  		binder0.add(it.previous());
  		exp0=new LetParExp(binder0, exp0);
  	}
  	if(SMT.considerGuardsForChoices())
  		decPre.conds.revertVariablesWithConditions();
  	return exp0;
  }
  
  public Expression visit(LetParExp exp)
  {
  	if(SMT.considerGuardsForChoices())
  	{
  		decPre.conds.analyse(exp.binder);
  	}
  	List<Binder> binderList=new ArrayList<AST.Binder>(exp.binder.length);
  	for(Binder b:exp.binder)
  	{
  	//Necessary  because of boolector. Boolector does not accept something like let (( result result))
  		if(b.exp instanceof IdentifierExp)
  		{
  			IdentifierExp iE=(IdentifierExp)b.exp;
  			if(iE.ident.string.equals(b.ident.string))
  			{
  				String aux=variableRenamings.get((ValueSymbol)iE.ident.getSymbol());
  				if(aux!=null)
  				{
  					variableRenamings.put((ValueSymbol) b.ident.getSymbol(), aux);
  				}
  				continue;
  			}
  		}
  		String name;
  		if(visitedQuantifiedVars.contains(b.ident.string)||decPre.forbiddenName(b.ident.string))
  		{
  			name=giveNewVariable(b.ident.string);
				variableRenamings.put((ValueSymbol) b.ident.getSymbol(), name);
  		}
  		else
  		{
  			name=b.ident.string;
  			visitedQuantifiedVars.add(name);
  		}
  		Identifier ident0=new Identifier(name);
  		Expression bindExp=(Expression) b.exp.accept(this);
  		binderList.add(new Binder(ident0, bindExp));
  	}  	
    Expression exp0 = (Expression)exp.exp.accept(this);
    if(SMT.considerGuardsForChoices())
  		decPre.conds.revertVariablesWithConditions();
    if(binderList.size()==0)
    	return exp0;
    return new LetParExp(binderList, exp0);
  }
  

  
  public Expression visit(MatchExp exp)
  {
  	SMT.notImplemented(exp, "Match Expressions are not yet implemented");
  	return null;
  }
  
  public Expression visit(FactorialExp exp)
  {	
  	IntType it=((IntType)exp.exp.getType());
  	if(it.ivalue2<=0)
  		return AST.number(1);
  	if(it.ivalue1.equals(it.ivalue2)&&it.ivalue1==1)
  		return AST.number(1);
  	Expression exp0 = (Expression)exp.exp.accept(this);
  	if(factorialInfo.factorialOccured==false)
  	{
  		factorialInfo.factorialOccured=true;
  		factorialInfo.minIntLowerBound=it.ivalue1;
  		factorialInfo.maxIntUpperBound=it.ivalue2;
  	}
  	else
  	{
  		factorialInfo.minIntLowerBound=Math.min(factorialInfo.minIntLowerBound, it.ivalue1);
  		factorialInfo.maxIntUpperBound=Math.max(factorialInfo.maxIntUpperBound, it.ivalue2);
  	}
  	List<Expression> exps=new ArrayList<AST.Expression>(1);
  	exps.add(exp0);
  	return new ApplicationExp(new Identifier(factorialInfo.name), exps);	  	
  }
  
  public Expression visit(NotEqualsExp exp)
  {
  	if((SMT.getTransformToNegationNormalForm()&&SMT.skolemisationInEquiavalences)&&exp.exp1.getType() instanceof BoolType)
  	{
  		Expression n1=new NotExp(exp.exp1);
  		Expression n2=new NotExp(exp.exp1);
  		Expression a1=new AndExp(exp.exp1, n2);
  		Expression a2=new AndExp(n1, exp.exp2);
  		return (Expression) (new OrExp(a1, a2)).accept(this);
  	}
  	else
  	{
      Expression exp10 = (Expression)exp.exp1.accept(this);
      Expression exp20 = (Expression)exp.exp2.accept(this);
      return new NotEqualsExp(exp10, exp20);
  	}
  }
  
  public Expression visit(NotExp exp)
  {
  	if(SMT.getTransformToNegationNormalForm())
  	{
  		Expression negatedExp=exp.exp;
    	if(negatedExp instanceof AndExp)
    	{
    		Expression n1=new NotExp(((AndExp)negatedExp).exp1);
    		Expression n2=new NotExp(((AndExp)negatedExp).exp2);
    		return (Expression) (new OrExp(n1, n2)).accept(this);
    	}
    	else if(negatedExp instanceof OrExp)
    	{
    		Expression n1=new NotExp(((OrExp)negatedExp).exp1);
    		Expression n2=new NotExp(((OrExp)negatedExp).exp2);
    		return (Expression) (new AndExp(n1, n2)).accept(this);
    	}
    	else if(negatedExp instanceof NotExp)
    	{
    		return (Expression) (((NotExp)negatedExp).exp).accept(this);
    	}
    	else if(negatedExp instanceof ImpliesExp)
    	{
    		Expression n=new NotExp(((ImpliesExp)negatedExp).exp2);
    		Expression e=((ImpliesExp)negatedExp).exp1;
    		return (Expression) (new AndExp(e, n)).accept(this);
    	}
    	else if(negatedExp instanceof EquivExp)
    	{
    		Expression e1=((EquivExp)negatedExp).exp1;
    		Expression e2=((EquivExp)negatedExp).exp2;
    		if(SMT.skolemisationInEquiavalences)
    		{
      		Expression n1=new NotExp(e1);
      		Expression n2=new NotExp(e2);
      		Expression a1=new AndExp(e1, n2);
      		Expression a2=new AndExp(n1, e2);
      		return (Expression) (new OrExp(a1, a2)).accept(this);
    		}
    		e1=(Expression) e1.accept(this);
  			e2=(Expression) e2.accept(this);
  			return new NotEqualsExp(e1, e2);
    		
    	}
    	else if(negatedExp instanceof EqualsExp)
    	{
    		if(SMT.skolemisationInEquiavalences&&((EqualsExp)negatedExp).exp1.getType() instanceof BoolType)
    		{
    			Expression e1=((EqualsExp)negatedExp).exp1;
      		Expression e2=((EqualsExp)negatedExp).exp2;
      		Expression n1=new NotExp(e1);
      		Expression n2=new NotExp(e2);
      		Expression a1=new AndExp(e1, n2);
      		Expression a2=new AndExp(n1, e2);
      		return (Expression) (new OrExp(a1, a2)).accept(this);
    		}
    		Expression e1=(Expression) ((EqualsExp)negatedExp).exp1.accept(this);
    		Expression e2=(Expression) ((EqualsExp)negatedExp).exp2.accept(this);
    		return new NotEqualsExp(e1, e2);   		
    	}
    	else if(negatedExp instanceof NotEqualsExp)
    	{
    		NotEqualsExp ne=(NotEqualsExp)negatedExp;
    		if(SMT.skolemisationInEquiavalences&&ne.exp1.getType() instanceof BoolType)
    		{
      		Expression n1=new NotExp(ne.exp1);
      		Expression n2=new NotExp(ne.exp2);
      		Expression o1=new OrExp(n1, ne.exp2);
      		Expression o2=new OrExp(ne.exp1, n2);
      		return (Expression) (new AndExp(o1, o2)).accept(this);
    		}
    		Expression e1=(Expression) ((NotEqualsExp)negatedExp).exp1.accept(this);
    		Expression e2=(Expression) ((NotEqualsExp)negatedExp).exp2.accept(this);
    		return new EqualsExp(e1, e2);
    	}
    	else if(negatedExp instanceof ForallExp)
    	{
    		ForallExp f=(ForallExp)negatedExp;
    		Expression n=new NotExp(f.exp);
    		return (Expression) (new ExistsExp(f.qvar, n)).accept(this);
    	}
    	else if(negatedExp instanceof ExistsExp)
    	{
    		ExistsExp e=(ExistsExp)negatedExp;
    		Expression n=new NotExp(e.exp);
    		return (Expression) (new ForallExp(e.qvar, n)).accept(this);
    	}  
    	
    	
    	else if(negatedExp instanceof TrueExp)
    	{
    		return new FalseExp();
    	}
    	else if(negatedExp instanceof FalseExp)
    	{
    		return new TrueExp();
    	}
    	
    	
    	else if(negatedExp instanceof LessExp)
    	{
    		LessExp lexp=(LessExp)negatedExp;
    		return (Expression) (new GreaterEqualExp(lexp.exp1, lexp.exp2)).accept(this);
    	}
    	else if(negatedExp instanceof LessEqualExp)
    	{
    		LessEqualExp lexp=(LessEqualExp)negatedExp;
    		return (Expression) (new GreaterExp(lexp.exp1, lexp.exp2)).accept(this);
    	}
    	else if(negatedExp instanceof GreaterExp)
    	{
    		GreaterExp lexp=(GreaterExp)negatedExp;
    		return (Expression) (new LessEqualExp(lexp.exp1, lexp.exp2)).accept(this);
    	}
    	else if(negatedExp instanceof GreaterEqualExp)
    	{
    		GreaterEqualExp lexp=(GreaterEqualExp)negatedExp;
    		return (Expression) (new LessExp(lexp.exp1, lexp.exp2)).accept(this);
    	}
    	else if(negatedExp instanceof LetExp)
    	{
    		LetExp e=(LetExp)negatedExp;
    		Expression n=new NotExp(e.exp);
    		List<Binder> lst=new ArrayList<AST.Binder>(Arrays.asList(e.binder));
    		return (Expression) (new LetExp(lst, n)).accept(this);
    	}
    	else if(negatedExp instanceof LetParExp)
    	{
    		LetParExp e=(LetParExp)negatedExp;
    		Expression n=new NotExp(e.exp);
    		List<Binder> lst=new ArrayList<AST.Binder>(Arrays.asList(e.binder));
    		return (Expression) (new LetParExp(lst, n)).accept(this);
    	}
  	} 	
    Expression exp0 = (Expression)exp.exp.accept(this);
    return new NotExp(exp0);
  }
  

  public Expression visit(NumberExp exp)
  {
		Utilities.Pair<List<Expression>, List<QuantifiedVariableCore>> conds=traverseQuantifiedVariable(exp.qvar);
		Expression exp0=AST.number(1);
//		List<QuantifiedVariableCore> qvcOld=orderQV(exp.qvar.qvcore);
		for(int i=conds.second.size()-1;i>=0;i--)
		{
//			There can be conditions for other variables that contain the current variable.
//			if(conds.first.get(i)==null)
//			{
//				long typeSize=((ValueSymbol)qvcOld.get(i).ident.getSymbol()).type.getSize();
//				exp0=new TimesExp(AST.number(typeSize), exp0);
//				continue;
//			}
			List<QuantifiedVariableCore> corelist=new ArrayList<QuantifiedVariableCore>(1);
			corelist.add(conds.second.get(i));
			QuantifiedVariable qv=new QuantifiedVariable(corelist , null);
			exp0=incorporateSumCondition(conds.first.get(i), exp0);
			exp0=new SumExp(qv, exp0);
		}
		return exp0;
  }
  
  public Expression visit(OrExp exp)
  {
  	if(exp.exp1 instanceof TrueExp||exp.exp2 instanceof TrueExp)
  		return new TrueExp();
  	if(exp.exp1 instanceof FalseExp)
  	{
  		return (Expression) exp.exp2.accept(this);
  	}
  	if(exp.exp2 instanceof FalseExp)
  	{
  		return (Expression) exp.exp1.accept(this);
  	} 	
  	if(SMT.considerGuardsForChoices())
  	{
  		decPre.conds.analyseWithSimpleConditions(new NotExp(exp.exp2));
  	}
	  Expression exp10 = (Expression)exp.exp1.accept(this);
	  if(SMT.considerGuardsForChoices())
	  {
	  	decPre.conds.revertSimpleConditions();
	  	decPre.conds.analyseWithSimpleConditions(new NotExp(exp.exp1));
	  }
	  Expression exp20 = (Expression)exp.exp2.accept(this);
	  if(SMT.considerGuardsForChoices())
	  	decPre.conds.revertSimpleConditions();
	  return new OrExp(exp10, exp20);
  }
  
  
  //PlusExpMult: a+(a+1)+...+(b-1)+b  
  /**
   * Here we use:
   * sum_{i=a}^{b} i = (a+b)*(b-a+1)/2
   */
  public Expression visit(PlusExpMult exp)
  {
  	Expression exp10 = (Expression)exp.exp1.accept(this);
    Expression exp20 = (Expression)exp.exp2.accept(this);
    Expression result=new Expression.PlusExp(exp10, exp20);
    result=new TimesExp(result, new PlusExp(AST.number(1), new Expression.MinusExp(exp20, exp10)));
  	result=new Expression.DividesExp(result, AST.number(2));
  	result=new IfThenElseExp(new LessEqualExp(exp10, exp20), result, AST.number(0));
  	return result;
  }
  public Expression visit(PowerExp exp)
  {
  	Long pow=Utilities.getIntValueFromExpression(exp.exp2);
  	if(pow!=null)
  	{
  		Expression powExp=(Expression) exp.exp1.accept(this);
  		for(long i=2;i<=pow;i++)
  		{
  			//We do not want that the factors of the resulting expression are the same instances.
  			powExp=new TimesExp(powExp, (Expression) exp.exp1.accept(this));
  		}
  		return powExp;
  	}	
  	String variable=giveNewVariable("power_var");
  	IntType powerType=(IntType)exp.exp2.getType();
  	long upperbound=powerType.ivalue2;
  	if(upperbound==0)
  		return AST.number(1);
  	Type type=new IntType(AST.number(1),AST.number(upperbound));
  	IdentifierTypeQuantifiedVar qvcore=new IdentifierTypeQuantifiedVar(new Identifier(variable), type);
  	List<QuantifiedVariableCore> qvcorelst=new ArrayList<AST.QuantifiedVariableCore>(1);
  	qvcorelst.add(qvcore);
  	QuantifiedVariable qv=new QuantifiedVariable(qvcorelst, null);
  	Expression factor=new IfThenElseExp(new LessEqualExp(new IdentifierExp(new Identifier(variable)), (Expression) exp.exp2.accept(this)), 
  			(Expression) exp.exp1.accept(this), AST.number(1));
  	return new ProductExp(qv, factor);  	
  }
  

  public Expression visit(PrintExp exp)
  {
  	return (Expression) exp.exp.accept(this);
//  	SMT.notImplemented(exp, "Print Expressions are not yet implemented");
//  	return null;
  }
  public Expression visit(PrintInExp exp)
  {
  	return (Expression) exp.exp.accept(this);
//  	SMT.notImplemented(exp, "PrintInExpressions are not yet implemented");
//  	return null;
  }
  public Expression visit(PrintTraceInExp exp)
  {
    return (Expression) exp.exp.accept(this);
//    SMT.notImplemented(exp, "PrintInExpressions are not yet implemented");
//    return null;
  }

  //why is the parameter called command?
  /**
   * We assume that function applications do not yield errors (we assume that the respective conditions
   * are already checked). Thus we replace check expressions by true.
   */
  public Expression visit(CheckExp command)
  {
  	return new TrueExp();
//  	SMT.notImplemented(command, "Check Expressions are not yet implemented");
//  	return null; 
  }
  
  public Expression visit(RecApplicationExp exp)
  {
  	SMT.notImplemented(exp, "Recursive Application Expressions are not yet implemented");
  	return null;  
  }
  
  public Expression visit(RecIdentifierExp exp)
  {
  	SMT.notImplemented(exp, "Recursive Identifier Expressions are not yet implemented");
  	return null;  
  }
  public Expression visit(RecordExp exp)
  {
  	List<Expression> exps=Arrays.stream(exp.eidents).map(ei -> (Expression) ei.exp.accept(this)).collect(Collectors.toList());
  	return new TupleExp(exps);
  }
  public Expression visit(RecordSelectionExp exp)
  {
  	RecordType rT=(RecordType)exp.exp.getType();
  	List<String> lst=Arrays.stream(rT.symbols).map(sym -> sym.ident.string).collect(Collectors.toList());
  	int index=lst.indexOf(exp.ident.string)+1;//tuple indices start with 1
  	Decimal dec=new Decimal(Integer.toString(index));
  	return new TupleSelectionExp((Expression)exp.exp.accept(this), dec);
  }
  public Expression visit(RecordUpdateExp exp)
  {
  	RecordType rT=(RecordType)exp.exp1.getType();
  	List<String> lst=Arrays.stream(rT.symbols).map(sym -> sym.ident.string).collect(Collectors.toList());
  	int index=lst.indexOf(exp.ident.string)+1;//tuple indices start with 1
  	Decimal dec=new Decimal(Integer.toString(index));
  	return new TupleUpdateExp((Expression) exp.exp1.accept(this), dec, (Expression) exp.exp2.accept(this));
  }
   
  public Expression visit(SetBuilderExp exp)
  {
  	QuantifiedVariable qVar=processQuantifiedVariable(exp.qvar);
  	Expression e=(Expression) exp.exp.accept(this);
  	return new SetBuilderExp(e, qVar);  	  	
  }
  
	private Expression incorporateSumCondition(Expression condition,Expression exp)
	{
		if(condition==null)
			return exp;
		return new IfThenElseExp((Expression) condition.accept(this), exp, AST.number(0));
	}
  
  public Expression visit(SumExp exp)
  {
  	if(SMT.considerGuardsForChoices())
  		decPre.conds.analyse(exp.qvar);
  	Utilities.Pair<List<Expression>, List<QuantifiedVariableCore>> conds=traverseQuantifiedVariable(exp.qvar);
  	Expression exp0=(Expression) exp.exp.accept(this);
  	for(int i=conds.second.size()-1;i>=0;i--)
  	{
  		List<QuantifiedVariableCore> corelist=new ArrayList<QuantifiedVariableCore>(1);
			corelist.add(conds.second.get(i));
			QuantifiedVariable qv=new QuantifiedVariable(corelist , null);
			exp0=incorporateSumCondition(conds.first.get(i), exp0);
			exp0=new SumExp(qv, exp0);
  	}
  	if(SMT.considerGuardsForChoices())
  		decPre.conds.revertVariablesWithConditions();
  	return exp0;
  }
  
	private Expression incorporateProductCondition(Expression condition,Expression exp)
	{
		if(condition==null)
			return exp;
		return new IfThenElseExp((Expression) condition.accept(this), exp, AST.number(1));
	}
  
  public Expression visit(ProductExp exp)
  {
  	if(SMT.considerGuardsForChoices())
  		decPre.conds.analyse(exp.qvar);
  	Utilities.Pair<List<Expression>, List<QuantifiedVariableCore>> conds=traverseQuantifiedVariable(exp.qvar);
  	Expression exp0=(Expression) exp.exp.accept(this);
  	for(int i=conds.second.size()-1;i>=0;i--)
  	{
  		List<QuantifiedVariableCore> corelist=new ArrayList<QuantifiedVariableCore>(1);
			corelist.add(conds.second.get(i));
			QuantifiedVariable qv=new QuantifiedVariable(corelist , null);
			exp0=incorporateProductCondition(conds.first.get(i), exp0);
			exp0=new ProductExp(qv, exp0);
  	}
  	if(SMT.considerGuardsForChoices())
  		decPre.conds.revertVariablesWithConditions();
  	return exp0;
  }
  
  public Expression visit(MinExp exp)
  {
  	if(!SMT.representMinMaxByUninterpretedFunctions)
  	{
  		QuantifiedVariable qVar=processQuantifiedVariable(exp.qvar);
  		Expression e=(Expression) exp.exp.accept(this);
    	return new MinExp(qVar, e);
  	}
  	Triple<List<ValueSymbol>,QuantifiedVariable, Expression> p=generateMinMaxFunction(exp,"min");
  	List<ValueSymbol> free=p.first;
  	QuantifiedVariable quant=p.second;
  	Expression application=p.third;
  	Expression exp1=exp.exp;
  	Expression exp0=new LessEqualExp(application,exp.exp);
  	exp0=universallyQuantifyExpression(exp.qvar, exp0);
  	application=(Expression) application.accept(cloner);
  	exp0=Utilities.constructNestedForallExpression(quant, exp0);
  	String minThName1=decPre.giveNewName("minTheorem1");
  	Identifier minTh1ident=new Identifier(minThName1);
  	Declaration th1=new TheoremDefinition(minTh1ident, exp0, true);
  	generatedDeclarations.add(th1);
  	
  	boolean guard=false;
  	List<Expression> conditionsToConsider=new ArrayList<AST.Expression>();
  	boolean hasConditions=quantifiedVariableHasConditions(exp.qvar);
  	if(hasConditions&&SMT.considerGuardsForChoices())
  	{	  		
  		for(ValueSymbol sym:free)
	  	{
	  		if(SMT.guardChoices==SMT.UsageModeGuardsForChoices.UseGuards&&decPre.conds.hashCondition(sym))
	  		{
	  			guard=true;
	  			conditionsToConsider.clear();
	  			break;
	  		}
	  		List<Expression> c=decPre.conds.getSimpleCondition(sym);
	  		if(c!=null)
	  			conditionsToConsider.addAll(c);
	  	}
  	}
  	quant=(QuantifiedVariable) quant.accept(cloner);
  	Expression theoremExp;
  	if(!guard)
  		theoremExp=getChooseTheoremExpression(exp.qvar, application, exp1);
  	else
  		theoremExp=getGuardedChooseTheoremExpression(exp.qvar, application, exp1, null);  	
  	if(free.size()>0)
  		theoremExp=quantifyChooseExp(theoremExp, quant, conditionsToConsider);
  	String chooseTh1Name=decPre.giveNewName("minTheorem2");
  	Identifier minTh2ident=new Identifier(chooseTh1Name);
  	Declaration th2=new TheoremDefinition(minTh2ident, theoremExp, true);
  	generatedDeclarations.add(th2);
  	return (Expression) application.accept(this);
  }
  
  public Expression visit(MaxExp exp)
  {
  	if(!SMT.representMinMaxByUninterpretedFunctions)
  	{
  		QuantifiedVariable qVar=processQuantifiedVariable(exp.qvar);
  		Expression e=(Expression) exp.exp.accept(this);
    	return new MaxExp(qVar, e);
  	}
  	Triple<List<ValueSymbol>,QuantifiedVariable, Expression> p=generateMinMaxFunction(exp,"max");
  	List<ValueSymbol> free=p.first;
  	QuantifiedVariable quant=p.second;
  	Expression application=p.third;
  	Expression exp1=exp.exp;
  	Expression exp0=new GreaterEqualExp(application,exp.exp);
  	exp0=universallyQuantifyExpression(exp.qvar, exp0);
  	application=(Expression) application.accept(cloner);
  	exp0=Utilities.constructNestedForallExpression(quant, exp0);
  	String minThName1=decPre.giveNewName("maxTheorem1");
  	Identifier minTh1ident=new Identifier(minThName1);
  	Declaration th1=new TheoremDefinition(minTh1ident, exp0, true);
  	generatedDeclarations.add(th1);
  	
  	boolean guard=false;
  	List<Expression> conditionsToConsider=new ArrayList<AST.Expression>();
  	boolean hasConditions=quantifiedVariableHasConditions(exp.qvar);
  	if(hasConditions&&SMT.considerGuardsForChoices())
  	{	  		
  		for(ValueSymbol sym:free)
	  	{
	  		if(SMT.guardChoices==SMT.UsageModeGuardsForChoices.UseGuards&&decPre.conds.hashCondition(sym))
	  		{
	  			guard=true;
	  			conditionsToConsider.clear();
	  			break;
	  		}
	  		List<Expression> c=decPre.conds.getSimpleCondition(sym);
	  		if(c!=null)
	  			conditionsToConsider.addAll(c);
	  	}
  	}
  	quant=(QuantifiedVariable) quant.accept(cloner);
  	Expression theoremExp;
  	if(!guard)
  		theoremExp=getChooseTheoremExpression(exp.qvar, application, exp1);
  	else
  		theoremExp=getGuardedChooseTheoremExpression(exp.qvar, application, exp1, null);  	
  	if(free.size()>0)
  		theoremExp=quantifyChooseExp(theoremExp, quant, conditionsToConsider);
  	String maxTh1Name=decPre.giveNewName("maxTheorem2");
  	Identifier maxTh2ident=new Identifier(maxTh1Name);
  	Declaration th2=new TheoremDefinition(maxTh2ident, theoremExp, true);
  	generatedDeclarations.add(th2);
  	return (Expression) application.accept(this);
  }
  

  //TimesExpMult: a*(a+1)*...(b-1)*b
  public Expression visit(TimesExpMult exp)
  {
  	Expression exp10 = (Expression)exp.exp1.accept(this);
    Expression exp20 = (Expression)exp.exp2.accept(this);
    IntType iT1=(IntType)exp.exp1.getType();
    IntType iT2=(IntType)exp.exp2.getType();
  	long ival1=Math.min(iT1.ivalue1, iT2.ivalue1);
  	long ival2=Math.min(iT1.ivalue2, iT2.ivalue2);
  	IntType newIT=new IntType(ival1, ival2);
  	String varname=giveNewVariable("i");
  	IdentifierTypeQuantifiedVar qvcore=new IdentifierTypeQuantifiedVar(new Identifier(varname), newIT);
  	List<QuantifiedVariableCore> qvcorelst=new ArrayList<AST.QuantifiedVariableCore>(1);
  	qvcorelst.add(qvcore);
  	QuantifiedVariable qv=new QuantifiedVariable(qvcorelst, null);
  	Expression condition=AST.and(new GreaterEqualExp(new IdentifierExp(new Identifier(varname)), exp10), 
  			new LessEqualExp(new IdentifierExp(new Identifier(varname)), exp20));
  	Expression factor=new IfThenElseExp(condition, new IdentifierExp(new Identifier(varname)), AST.number(1));
  	return new ProductExp(qv, factor);
  }

  public Expression visit(TupleExp exp)
  {
  	List<Expression>lst=Arrays.stream(exp.exps).map(e -> (Expression)e.accept(this)).collect(Collectors.toList());
  	return new TupleExp(lst);

  }
  public Expression visit(TupleSelectionExp exp)
  {
  	Expression exp1=(Expression) exp.exp.accept(this);
  	Decimal dec=(Decimal) exp.index.accept(this);
  	return new TupleSelectionExp(exp1, dec);
  }
  public Expression visit(TupleUpdateExp exp)
  {
  	Expression tupleExp=(Expression) exp.exp1.accept(this);
  	Expression newValue=(Expression) exp.exp2.accept(this);
  	Decimal dec=(Decimal) exp.index.accept(this);
  	return new TupleUpdateExp(tupleExp, dec, newValue);

  }

  public QuantifiedVariableCore visit(IdentifierTypeQuantifiedVar qvcore)
  {
  	String str=qvcore.ident.string;
  	Identifier ident0;
  	if(visitedQuantifiedVars.contains(str)||decPre.forbiddenName(str))
		{
			String newName=giveNewVariable(str);
			ident0 = new Identifier(newName);
			variableRenamings.put((ValueSymbol) qvcore.ident.getSymbol(), newName);
		}
  	else	  	
  	{
  		visitedQuantifiedVars.add(str);
  		ident0 = (Identifier)qvcore.ident.accept(this);
  	}	  		
  	Type type0=(Type) ((IdentifierTypeQuantifiedVar)qvcore).type.accept(this);
  	if (type0.isCanonical()) type0 = Utilities.canonic2syntacticType(type0);
    return new IdentifierTypeQuantifiedVar(ident0, type0);
  }
  
  /**
   * Transform an IdentifierSetQuantifiedVar to an IdentifierTypeQuantifiedVar.
   * The caller is responsible for asserting that the quantified variable is
   * actually in the set.
   */
  public QuantifiedVariableCore visit(IdentifierSetQuantifiedVar qvcore)
  {
  	String str=qvcore.ident.string;
  	Identifier ident0;
  	if(visitedQuantifiedVars.contains(str)||decPre.forbiddenName(str))
		{
			String newName=giveNewVariable(str);
			ident0 = new Identifier(newName);
			variableRenamings.put((ValueSymbol) qvcore.ident.getSymbol(), newName);
		}
  	else	  
  	{
  		visitedQuantifiedVars.add(str);
  		ident0 = (Identifier)qvcore.ident.accept(this);
  	}
  	SetType aux=(SetType)Utilities.getBaseType(qvcore.exp.getType());
  	Type type0=Utilities.canonic2syntacticType(aux.type);
  	return new IdentifierTypeQuantifiedVar(ident0, type0);
  }	 
  
  //Not really necessary
//  public Type visit(ArrayType type)
//  {
//  	Long v=Utilities.getIntValueFromExpression(type.exp);
//		if(v==null)
//		{
//			SMT.unexpectedErrorDedected(type.exp, "The expression is not an int");
//			return null;
//		}			
//		long l=v.longValue();
//		IntType domainType=new IntType(AST.number(0), AST.number(l-1));
//		Type imageType=(Type) type.type.accept(this);
//		return new MapType(domainType, imageType);
//  }
  
  public Type visit(RecordType type)
  {
  	List<Type> types;
    if (type.isCanonical())
    {
    	ValueSymbol[] vals=type.symbols;
    	types=Arrays.stream(vals).map(v -> Utilities.canonic2syntacticType(v.type)).collect(Collectors.toList());
    }
    else
    {
    	types=Arrays.stream(type.param).map(p -> (Type) p.type.accept(this)).collect(Collectors.toList());
    }
    return new TupleType(types);
  }
  
  public Type visit(IdentifierType type)
  {
  	String typeName=decPre.getTypeRenaming((TypeSymbol) type.ident.getSymbol());
  	typeName=typeName==null?type.ident.string:typeName;
  	Identifier ident0=new Identifier(typeName);
    return new IdentifierType(ident0);
  }
     
  public Expression visit(NotInSetExp exp)
  {
    Expression exp10 = (Expression)exp.exp1.accept(this);
    Expression exp20 = (Expression)exp.exp2.accept(this);
    return new NotExp(new InSetExp(exp10,exp20));
  }
  
  public Expression visit(NotSubsetExp exp)
  {
    Expression exp10 = (Expression)exp.exp1.accept(this);
    Expression exp20 = (Expression)exp.exp2.accept(this);
    return new NotExp(new SubsetExp(exp10,exp20));
  }
}
//---------------------------------------------------------------------------
// end of file
//---------------------------------------------------------------------------
