// ---------------------------------------------------------------------------
// CloneIds.java
// Clone expressions replacing identifiers/symbols.
// $Id: CloneIds.java,v 1.2 2021/05/12 13:19:49 schreine Exp $
//
// Author: Wolfgang Schreiner <Wolfgang.Schreiner@risc.jku.at>
// Copyright (C) 2017-, Research Institute for Symbolic Computation (RISC)
// Johannes Kepler University, Linz, Austria, http://www.risc.jku.at
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program.  If not, see <http://www.gnu.org/licenses/>.
// ----------------------------------------------------------------------------
package riscal.tasks;

import riscal.syntax.*;
import riscal.syntax.AST.*;
import riscal.types.Environment.Symbol;
import riscal.types.Environment.Symbol.ValueSymbol;

import java.util.*;

public class CloneIds extends ASTCloner
{
  /****************************************************************************
   * Bind parameters by arguments in expression (which is cloned)
   * @param params the parameters.
   * @param args the arguments.
   * @param exp the expresison.
   * @return the cloned expression with the variables bound.
   ***************************************************************************/
  public static Expression bind(List<Identifier> params, List<Expression> args, Expression exp)
  {
    int n = params.size();
    Map<ValueSymbol,ValueSymbol> map = new HashMap<ValueSymbol,ValueSymbol>();
    List<Binder> binders = new ArrayList<Binder>(n);
    for (int i=0; i<n; i++)
    {
      Identifier ident0 = params.get(i);
      Identifier ident1 = new Identifier(ident0.string);
      ValueSymbol symbol0 = (ValueSymbol)ident0.getSymbol();
      ValueSymbol symbol1 = new ValueSymbol(ident1, symbol0.type, symbol0.kind);
      map.put(symbol0, symbol1);
      Binder b = new Binder(ident1, args.get(i));
      binders.add(b);
    }
    Expression exp0 = clone(exp, map);
    if (n > 0) exp0 = new Expression.LetParExp(binders, exp0);
    exp0.setType(exp.getType());
    return exp0;
  }
   
  /****************************************************************************
   * Clone expression replacing identifiers.
   * @param exp the expression.
   * @param map the identifier (symbol) mapping.
   * @return the cloned expression)
   ***************************************************************************/
  public static Expression clone(Expression exp, Map<ValueSymbol,ValueSymbol> map)
  {
    CloneIds clone = new CloneIds(map);
    return (Expression)exp.accept(clone);
  }
  
  private Map<ValueSymbol,ValueSymbol> map;
  private CloneIds(Map<ValueSymbol,ValueSymbol> map)
  {
    // preserve types of expressions
    super(true);
    this.map = map;
  }
  
  public Identifier visit(Identifier id)
  {
    // identifier is only cloned, if symbol is to be replaced
    Symbol symbol = id.getSymbol();
    if (!(symbol instanceof ValueSymbol)) return id;
    ValueSymbol symbol0 = (ValueSymbol)symbol;
    ValueSymbol symbol1 = map.get(symbol0);
    if (symbol1 == null) return id;
    Identifier id1 = new Identifier(id.string);
    id1.setSymbol(symbol1);
    return id1;
  }
}
// ---------------------------------------------------------------------------
// end of file
// ---------------------------------------------------------------------------