// ---------------------------------------------------------------------------
// Hash.java
// Hash code computation
// $Id: Hash.java,v 1.4 2023/07/03 07:12:26 schreine Exp schreine $
//
// Author: Wolfgang Schreiner <Wolfgang.Schreiner@risc.jku.at>
// Copyright (C) 2016-, Research Institute for Symbolic Computation (RISC)
// Johannes Kepler University, Linz, Austria, http://www.risc.jku.at
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program.  If not, see <http://www.gnu.org/licenses/>.
// ----------------------------------------------------------------------------
package riscal.util;

public final class Hash
{
  /****************************************************************************
   * Compute a hash code from an array of non-null objects (Java version)
   * @param v the array.
   * @return the hash code.
   ***************************************************************************/
  public static int hashCode0(Object[] v) 
  { 
    // gives many collisions for arrays with small integer values
    return java.util.Arrays.hashCode(v); 
  }
  
  /*****************9***********************************************************
   * Compute a hash code from a non-null array of non-null objects (Java derived)
   * @param v the array.
   * @return the hash code.
   ***************************************************************************/
  public static int hashCode(Object[] v) 
  { 
    // omit null checks, not required for our applications
    // if (v == null) return 0;
    int result = 1;
    for (int i = 0; i < v.length; i++)
    {
      int elt = /* v[i] == null ? 0 : */ v[i].hashCode();
      // https://stackoverflow.com/questions/299304/why-does-javas-hashcode-in-string-use-31-as-a-multiplier
      // https://stackoverflow.com/questions/1835976/what-is-a-sensible-prime-for-hashcode-calculation/
      // result = 31 * result + elt;
      result = 92821 * result + elt;
    }
    return result;
  }
  
  /****************************************************************************
   * Compute a hash code from a non-null array of non-null objects (15 bit nonces)
   * @param v the array.
   * @return the hash code.
   ***************************************************************************/
  public static int hashCode15(Object[] v) 
  { 
    // some odd 15 bit primes
    final int P1 = 32749;
    final int P2 = 29243;
    final int P3 = 19997;
    final int P4 = 31321;
    final int P5 = 23827;
    final int n = v.length;
    int result = 0;
    int i = 0;
    while (true)
    {
      if (i == n) return result;
      result += P1*v[i].hashCode(); i++;
      if (i == n) return result;
      result += P2*v[i].hashCode(); i++;
      if (i == n) return result;
      result += P3*v[i].hashCode(); i++;
      if (i == n) return result;
      result += P4*v[i].hashCode(); i++;
      if (i == n) return result;
      result += P5*v[i].hashCode(); i++;
    }
  }
  
  /****************************************************************************
   * Compute a hash code from a non-null array of non-null objects (31 bit nonces)
   * @param v the array.
   * @return the hash code.
   ***************************************************************************/
  public static int hashCode31(Object[] v) 
  { 
    // some odd 31 bit primes
    final int P1 = 1288409459;
    final int P2 = 785938529;
    final int P3 = 2147483647;
    final int P4 = 352532317;
    final int P5 = 1955449841;
    final int n = v.length;
    int result = 0;
    int i = 0;
    while (true)
    {
      if (i == n) return result;
      result += P1*v[i].hashCode(); i++;
      if (i == n) return result;
      result += P2*v[i].hashCode(); i++;
      if (i == n) return result;
      result += P3*v[i].hashCode(); i++;
      if (i == n) return result;
      result += P4*v[i].hashCode(); i++;
      if (i == n) return result;
      result += P5*v[i].hashCode(); i++;
    }
  }
}
// ----------------------------------------------------------------------------
// end of file
// ----------------------------------------------------------------------------